# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from collections import namedtuple


Check = namedtuple('check', 'name status remark')
PASS = 'Pass'
FAIL = 'Fail'


def _check_collection_size(request, collection_name):
    """Return a Check that a mongodb collection is not empty."""
    status = FAIL
    try:
        count = request.db[collection_name].count()
        if count > 0:
            remark = '{0} {1} found'.format(count, collection_name)
            status = PASS
        else:
            remark = 'There are no {0}. Is ingest running?'
            remark = remark.format(collection_name)
    except:
        remark = 'The mongodb {0} collection is not available.'
        remark = remark.format(collection_name)
    finally:
        check = Check('{0} collection'.format(collection_name), status, remark)
    return check


def check_charms_collection(request):
    """Return a Check of the mongodb charms collection."""
    return _check_collection_size(request, 'charms')


def check_bundles_collection(request):
    """Return a Check of the mongodb bundles collection."""
    return _check_collection_size(request, 'bundles')


def check_elasticsearch(request):
    """Return a Check of the elasticsearch charms index."""
    status = FAIL
    try:
        results = request.index_client.search(terms='')
        if results['result_total'] > 0:
            remark = '{0} matches found'.format(results['result_total'])
            status = PASS
        else:
            remark = 'There are no matches. Is ingest running?'
    except:
        remark = 'The elasticsearch charm index is not available.'
    finally:
        check = Check('charms index', status, remark)
    return check


def check_api2(request):
    """Return a Check of the API charms interesting end point."""
    status = FAIL
    try:
        from charmworld.views.api import API2
        api = API2(request)
        interesting = api.charms(path=['interesting'])['result']
        if (interesting['new']
                and interesting['popular']
                and interesting['featured']):
            status = PASS
            remark = 'Interesting has new, popular, and featured.'
        else:
            remark = 'Interesting is missing new, popular, or featured.'
    except:
        remark = 'The API2 is not available.'
    finally:
        check = Check('API2 interesting', status, remark)
    return check


def check_api3(request):
    """Return a Check of the API charms interesting end point."""
    status = FAIL
    try:
        from charmworld.views.api import API3
        api = API3(request)
        interesting = api.search(path=['interesting'])['result']
        if (interesting['new']
                and interesting['popular']
                and interesting['featured']):
            status = PASS
            remark = 'Interesting has new, popular, and featured.'
        else:
            remark = 'Interesting is missing new, popular, or featured.'
    except:
        remark = 'The API3 is not available.'
    finally:
        check = Check('API3 interesting', status, remark)
    return check


def check_collections(request):
    """Return a Check of the collections that should exist in the Mongo DB."""
    status = FAIL
    required = set((
        'basket-queue', 'baskets', 'bundles', 'charm-queue', 'charms'))
    try:
        existing = set(request.db.collection_names())
        missing = required.difference(existing)
        if missing:
            missing = ', '.join(sorted(missing))
            remark = 'Required collections do not (yet) exist: {0}'
            remark = remark.format(missing)
        else:
            status = PASS
            remark = 'All required collections exist.'
    except:
        remark = "Cannot query the MongoDB's collection names."
    finally:
        check = Check('MongoDB collections', status, remark)
    return check


def check_ingest_queues(request):
    """Check the size of the ingest queues."""
    status = FAIL
    try:
        charm_queue_size = request.db['charm-queue'].count()
        basket_queue_size = request.db['basket-queue'].count()
        status = PASS
        remark = 'queued charms: {0}, queued baskets: {1}.'
        remark = remark.format(charm_queue_size, basket_queue_size)
    except:
        remark = 'Cannot query the ingest queues.'
    finally:
        check = Check('Ingest queue sizes', status, remark)
    return check
