# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
import datetime
import logging

import pymongo
import requests

from charmworld.utils import (
    configure_logging,
    get_ini,
    Timer,
)
from config import MONGO_URL
from utils import lock
from utils import LockHeld


def make_github_request(endpoint_name, repo_name=None):
    if endpoint_name == 'repos':
        url = 'https://api.github.com/users/charms/repos'
    elif endpoint_name == 'pulls':
        if repo_name is None:
            raise ValueError('repo_name cannot be None for pull requests')
        url = 'https://api.github.com/repos/charms/%s/pulls' % repo_name
    r = requests.get(url)
    return r.json()


def get_charm_repos(log=None):
    with Timer() as timer:
        repos = make_github_request('repos')
    if log:
        log.debug("github charms query time %0.2f", timer.duration())
    return [r['name'] for r in repos]


def get_pull_requests(repos, log=None):
    pull_requests = []
    with Timer() as timer:
        for r in repos:
            pull_requests += make_github_request('pulls', r)
    if log:
        log.debug("github charms query time %0.2f", timer.duration())
    return [p for p in pull_requests if p['merged_at'] is None]


def create_entries_from_pull_requests(pull_requests):
    entries = []
    for p in pull_requests:
        date_created = datetime.datetime.strptime(
            p['created_at'], '%Y-%m-%dT%H:%M:%SZ')
        date_modified = datetime.datetime.strptime(
            p['updated_at'], '%Y-%m-%dT%H:%M:%SZ')
        entry = {
            'type': 'pull request',
            'date_modified': date_modified,
            'date_created': date_created,
            'summary': 'Pull request on %s' % p['base']['repo']['name'],
            'item': p['issue_url'],
            'status': p['state'],
        }
        entries.append(entry)
    return entries


def update_pull_request_collection(db, entries):
    db.drop_collection('github_pull_requests')
    if (entries != []):
        # Needs check as bulk insert of an empty list is prohibited
        db.github_pull_requests.insert(entries)


def main():
    configure_logging()
    log = logging.getLogger("charm.github")
    conn = pymongo.Connection(MONGO_URL, safe=True)
    db = conn['juju']
    settings = get_ini()
    charm_repos = get_charm_repos(log)
    pull_requests = get_pull_requests(charm_repos, log)
    entries = create_entries_from_pull_requests(pull_requests)
    try:
        with lock('github',
                  int(settings['script_lease_time']) * 60, db, log):
            update_pull_request_collection(db, entries)
    except LockHeld, error:
        log.warn(str(error))
    log.debug('Completed job with %s entries.' % len(entries))


if __name__ == '__main__':
    main()
