# Copyright 2012, 2013 Marco Ceppi, Canonical Ltd.  This software is
# licensed under the GNU Affero General Public License version 3 (see
# the file LICENSE).

"""
Charm review queue

 Contributed by Juan Negronil


 Changes
 - convert to job.
 - use dateutil date parsing instead of string comparsion.
 - merge proposals show target with series/branch.

"""
import datetime
import itertools
import logging

from launchpadlib.launchpad import Launchpad
import pymongo

from charmworld.utils import (
    configure_logging,
    get_ini,
    Timer,
)
from config import MONGO_URL
from utils import lock
from utils import LockHeld


def get_tasks_and_proposals(log):
    '''Gets review items (bugs and proposals) from launchpad.'''
    # TODO: Eventually these single requests should be broken into two each:
    # one to get anything created within the window, and one to get anything
    # older than the window that is still open. Right now this is probably the
    # more performant way to go about it, but that's going to change as the
    # data outside of our window gets larger.
    try:
        lp_credentials = get_ini()['lp_credentials_file']
    except KeyError:
        lp_credentials = None
    with Timer() as timer:
        if lp_credentials:
            lp = Launchpad.login_with('charm-tools',
                                      'production',
                                      version='devel',
                                      credentials_file=lp_credentials)
        else:
            lp = Launchpad.login_anonymously('charm-tools',
                                             'production',
                                             version='devel')

        # Entity roots for items of interests
        charm = lp.distributions['charms']
        charmers = lp.people['charmers']
        charm_contributors = lp.people['charm-contributors']

    log.debug("lp login time %0.2f", timer.duration())

    branch_filter = "Show only Bugs with linked Branches"
    with Timer() as timer:
        bugs = charm.searchTasks(
            tags=['new-formula', 'new-charm'],
            tags_combinator="Any",
            linked_branches=branch_filter)
        charmers_bugs = charmers.searchTasks(linked_branches=branch_filter)
        proposals = charmers.getRequestedReviews()
        contributor_proposals = charm_contributors.getRequestedReviews()

        juju = lp.projects['juju']
        doc_proposals = [mp for mp in juju.getMergeProposals()
                         if mp.target_branch_link.endswith("~juju/juju/docs")]

    log.debug("lp query time %0.2f", timer.duration())
    return (itertools.chain(bugs, charmers_bugs),
            itertools.chain(proposals, contributor_proposals, doc_proposals))


def calculate_proposal_latency(proposal, now):
    if proposal.date_reviewed:
        latency = (
            proposal.date_reviewed - proposal.date_review_requested)
    else:
        # `now` has been calculated in UTC, and LP returns UTC TZ
        # datetimes. We can safely remove the TZ, letting us get the
        latency = (
            now - proposal.date_review_requested.replace(tzinfo=None))
    return latency.total_seconds()


def calculate_task_latency(task, now):
    link_entries = [e for e in task.bug.activity.entries if
                    e['whatchanged'] == 'branch linked']
    if len(link_entries) == 1:
        start_date = link_entries[0]['datechanged']
        start_date = datetime.datetime.strptime(
            start_date.split('.')[0], '%Y-%m-%dT%H:%M:%S')
    elif len(link_entries) > 1:
        #Filter down to links associated with the currently linked branches
        linked_branches = task.bug.linked_branches
        branch_names = [link.branch.display_name for link in linked_branches]
        current_entries = [e for e in link_entries if
                           e['newvalue'] in branch_names]
        if current_entries != []:
            start_date = current_entries[0]['datechanged']
            start_date = datetime.datetime.strptime(
                start_date.split('.')[0], '%Y-%m-%dT%H:%M:%S')
        else:
            # At some point, the name of the branch changed, so we can't find
            # the event when it was added. This happens rarely, so we just
            # abandon stats for this task.
            return
    elif link_entries == []:
        # We have no entries, probably b/c the lp step failed to login with
        # authentication.
        return

    if task.date_fix_committed:
        end_date = task.date_fix_committed.replace(tzinfo=None)
    else:
        end_date = now
    latency = end_date - start_date
    return latency.total_seconds()


def update_review_queue(db, tasks, proposals, log):
    '''Updates the review queue information.'''
    now = datetime.datetime.utcnow()
    window = now - datetime.timedelta(days=180)  # 6 month window for latency
    bug_statuses = [
        'New', 'Confirmed', 'Triaged', 'In Progress', 'Fix Committed']
    r_queue = []
    latencies = []
    for task in tasks:
        if (task.date_created.replace(tzinfo=None) >= window or
            task.status in bug_statuses):
            latency = calculate_task_latency(task, now)
            if latency:
                latencies.append(latency)

        if task.status in bug_statuses:
            entry = {
                'type': 'task',
                'date_modified': task.bug.date_last_updated,
                'date_created': task.date_created,
                'summary': task.title.split('"')[1].strip(),
                'item': task.web_link,
                'status': task.status,
            }
            r_queue.append(entry)

    proposal_statuses = [
        'Needs review', 'Needs fixing', 'Needs information']
    for proposal in proposals:
        if (proposal.date_created.replace(tzinfo=None) >= window or
            proposal.queue_status in proposal_statuses):
            latency = calculate_proposal_latency(proposal, now)
            latencies.append(latency)
        if proposal.queue_status in proposal_statuses:
            parts = proposal.target_branch_link.rsplit('/', 3)
            if parts[-1] == "trunk":
                parts.pop(-1)
            target = "/".join(parts[2:])
            origin = "/".join(proposal.source_branch_link.rsplit('/', 3)[1:])

            num_comments = proposal.all_comments.total_size
            if num_comments != 0:
                last_modified = proposal.all_comments[
                    num_comments - 1].date_created
            else:
                last_modified = proposal.date_created

            entry = {
                'type': 'proposal',
                'date_modified': last_modified,
                'date_created': proposal.date_created,
                'summary': "Merge for %s from %s" % (target, origin),
                'item': proposal.web_link,
                'status': proposal.queue_status,
            }
            r_queue.append(entry)

    # Drop old collection
    log.info(
        "Dropping old review queue (%d items)", db['review_queue'].count())
    db.drop_collection('review_queue')

    # Bulk insert
    log.info("Updating review queue (%d items)" % len(r_queue))
    if (r_queue != []):
        # Needs check as bulk insert of an empty list is prohibited
        db['review_queue'].insert(r_queue)

    # Get stats:
    stats = latency_stats(latencies)
    if stats:
        stats['date'] = now
        db.review_queue_latency.insert(stats)


def latency_stats(queue):
    if queue != []:
        min_latency = min(queue)
        mean_latency = sum(queue) / len(queue)
        max_latency = max(queue)
        return {'min': min_latency, 'max': max_latency, 'mean': mean_latency}


def main():
    configure_logging()
    log = logging.getLogger("charm.review")
    c = pymongo.Connection(MONGO_URL, safe=True)
    db = c['juju']
    settings = get_ini()
    tasks, proposals = get_tasks_and_proposals(log)
    try:
        with lock('ingest-review',
                  int(settings['script_lease_time']) * 60, db, log):
            update_review_queue(db, tasks, proposals, log)
    except LockHeld, error:
        log.warn(str(error))

if __name__ == '__main__':
    main()
