# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
import json

from mock import patch

from charmworld.jobs.github import (
    create_entries_from_pull_requests,
    get_charm_repos,
    get_pull_requests,
    make_github_request,
)
from charmworld.testing import MongoTestBase


class FakeResponse:
    def __init__(self, json={}):
        self._json = json

    def json(self):
        return self._json


class FakeGet:
    def __init__(self, json=None):
        self.json = json

    def __call__(self, url):
        self.url = url
        return FakeResponse(self.json)


class GithubTest(MongoTestBase):

    def test_make_github_request_repos(self):
        with patch('requests.get', FakeGet()) as mock:
            make_github_request('repos')
        self.assertEqual(mock.url, 'https://api.github.com/users/charms/repos')

    def test_make_github_request_pull_requests(self):
        with patch('requests.get', FakeGet()) as mock:
            make_github_request('pulls', 'foo')
        self.assertEqual(
            mock.url, 'https://api.github.com/repos/charms/foo/pulls')

    def test_make_github_request_pull_requests_requires_repo(self):
        with patch('requests.get', FakeGet()):
            self.assertRaises(ValueError, make_github_request, 'pulls')

    def test_get_charm_repos(self):
        test_json = json.load(file('charmworld/testing/data/repos.json'))
        with patch('requests.get', FakeGet(json=test_json)) as mock:
            repos = get_charm_repos()
        self.assertEqual(
            mock.url, 'https://api.github.com/users/charms/repos')
        self.assertEqual(1, len(repos))

    def test_get_pull_requests(self):
        test_json = json.load(file('charmworld/testing/data/pulls.json'))
        with patch('requests.get', FakeGet(json=test_json)) as mock:
            pulls = get_pull_requests(['foo'])
        self.assertEqual(
            mock.url, 'https://api.github.com/repos/charms/foo/pulls')
        self.assertEqual(2, len(pulls))

    def test_get_pull_requests_filters_merged(self):
        test_json = json.load(file('charmworld/testing/data/pulls.json'))
        test_json[0]['merged_at'] = '2012-10-01T11:12:13Z'
        with patch('requests.get', FakeGet(json=test_json)):
            pulls = get_pull_requests(['foo'])
        self.assertEqual(1, len(pulls))

    def test_create_entries_from_pull_requests(self):
        test_json = json.load(file('charmworld/testing/data/pulls.json'))
        with patch('requests.get', FakeGet(json=test_json)):
            pulls = get_pull_requests(['foo'])
        entries = create_entries_from_pull_requests(pulls)
        self.assertEqual(2, len(entries))
        entry = entries[0]
        self.assertEqual('Pull request on bar', entry['summary'])
        self.assertEqual('https://github.com/foo/bar/pull/1', entry['item'])

    def test_create_entries_from_pull_requests_empty(self):
        entries = create_entries_from_pull_requests([])
        self.assertEqual([], entries)
