# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
import datetime
import logging

import mock

from charmworld.jobs.review import (
    calculate_proposal_latency,
    calculate_task_latency,
    update_review_queue,
)
from charmworld.testing import MongoTestBase


NOW = datetime.datetime.utcnow()
YESTERDAY = NOW - datetime.timedelta(1)
TWO_DAYS = NOW - datetime.timedelta(2)


def get_mock_task(confirmed=False, entries=[], linked_branches=[]):
    if confirmed:
        status = "Confirmed"
        date_fix_committed = YESTERDAY
    else:
        date_fix_committed = None
        status = "New"
    return mock.Mock(
        status=status,
        date_fix_committed=date_fix_committed,
        date_created=TWO_DAYS,
        title=u'Bug #0 in a collection: "A bug task"',
        web_link="http://example.com",
        bug=mock.Mock(
            date_last_updated=NOW,
            linked_branches=linked_branches,
            activity=mock.Mock(entries=entries)
        ),
    )


def get_mock_proposal(reviewed=False):
    base = 'https://api.launchpad.net/devel/'
    if reviewed:
        date_reviewed = YESTERDAY
        queue_status = "Approved"
    else:
        date_reviewed = None
        queue_status = "Needs review"
    return mock.Mock(
        target_branch_link=base + '~charmers/charms/precise/foo/trunk',
        source_branch_link=base + '~bar/charms/precise/foo/fnord',
        all_comments=mock.Mock(total_size=0),
        date_review_requested=TWO_DAYS,
        date_created=TWO_DAYS,
        web_link='http://example.com',
        date_reviewed=date_reviewed,
        queue_status=queue_status,
    )


class ReviewTest(MongoTestBase):

    def test_review_handles_empty_queue(self):
        self.db.review_queue.insert([{'foo': 'bar'}])
        log = logging.getLogger('foo')
        update_review_queue(self.db, [], [], log)
        self.assertEqual(0, self.db.review_queue.count())

    def test_review_handles_bugs(self):
        log = logging.getLogger('foo')
        update_review_queue(self.db, [get_mock_task()], [], log)
        self.assertEqual(1, self.db.review_queue.count())
        item = self.db.review_queue.find_one()
        self.assertEqual(NOW.date(), item['date_modified'].date())
        self.assertEqual(TWO_DAYS.date(), item['date_created'].date())
        self.assertEqual('A bug task', item['summary'])
        self.assertEqual('http://example.com', item['item'])
        self.assertEqual('New', item['status'])

    def test_review_handles_proposals(self):
        log = logging.getLogger('foo')
        update_review_queue(self.db, [], [get_mock_proposal()], log)
        self.assertEqual(1, self.db.review_queue.count())
        item = self.db.review_queue.find_one()
        self.assertEqual(TWO_DAYS.date(), item['date_modified'].date())
        self.assertEqual(TWO_DAYS.date(), item['date_created'].date())
        self.assertEqual(
            'Merge for foo from precise/foo/fnord', item['summary'])
        self.assertEqual('http://example.com', item['item'])
        self.assertEqual('Needs review', item['status'])

    def test_calculate_task_latency_with_two_valid_links(self):
        linked_branches = [
            mock.Mock(branch=mock.Mock(display_name='lp:~foo/bar/bjorn')),
            mock.Mock(branch=mock.Mock(display_name='lp:~foo/bar/baz'))
        ]
        two_days = datetime.datetime.strftime(
            TWO_DAYS, '%Y-%m-%dT%H:%M:%S')
        yesterday = datetime.datetime.strftime(
            YESTERDAY, '%Y-%m-%dT%H:%M:%S')
        entries = [
            {'whatchanged': 'branch linked',
             'newvalue': 'lp:~foo/bar/baz',
             'datechanged': two_days},
            {'whatchanged': 'branch linked',
             'newvalue': 'lp:~foo/bar/bjorn',
             'datechanged': yesterday}]
        latency = calculate_task_latency(
            get_mock_task(
                True, entries=entries, linked_branches=linked_branches), NOW)
        self.assertEqual(60 * 60 * 24, int(latency))

    def test_calculate_task_latency_bug_with_one_valid_link(self):
        linked_branches = [
            mock.Mock(branch=mock.Mock(display_name='lp:~foo/bar/bjorn'))
        ]
        two_days = datetime.datetime.strftime(
            TWO_DAYS, '%Y-%m-%dT%H:%M:%S')
        yesterday = datetime.datetime.strftime(
            YESTERDAY, '%Y-%m-%dT%H:%M:%S')
        entries = [
            {'whatchanged': 'branch linked',
             'newvalue': 'lp:~foo/bar/baz',
             'datechanged': yesterday},
            {'whatchanged': 'branch linked',
             'newvalue': 'lp:~foo/bar/bjorn',
             'datechanged': two_days}]
        latency = calculate_task_latency(
            get_mock_task(
                True, entries=entries, linked_branches=linked_branches), NOW)
        self.assertEqual(60 * 60 * 24, int(latency))

    def test_calculate_task_latency_bug_with_one_link(self):
        date_changed = datetime.datetime.strftime(
            TWO_DAYS, '%Y-%m-%dT%H:%M:%S')
        entries = [{'whatchanged': 'branch linked',
                    'newvalue': 'lp:~foo/bar/baz',
                    'datechanged': date_changed}]
        latency = calculate_task_latency(
            get_mock_task(True, entries=entries), NOW)
        self.assertEqual(60 * 60 * 24, int(latency))

    def test_calculate_proposal_latency(self):
        latency = calculate_proposal_latency(get_mock_proposal(True), NOW)
        self.assertEqual(60 * 60 * 24, int(latency))

    def test_review_job_saves_latency(self):
        log = logging.getLogger('foo')
        self.assertEqual(0, self.db.review_queue_latency.count())
        update_review_queue(self.db, [], [get_mock_proposal(True)], log)
        self.assertEqual(1, self.db.review_queue_latency.count())
        item = self.db.review_queue_latency.find_one()
        self.assertEqual(NOW.date(), item['date'].date())
        self.assertEqual(1, datetime.timedelta(seconds=item['max']).days)
        self.assertEqual(1, datetime.timedelta(seconds=item['mean']).days)
        self.assertEqual(1, datetime.timedelta(seconds=item['min']).days)

    def test_review_discards_closed_bugs(self):
        tasks = [get_mock_task(), get_mock_task()]
        tasks[0].status = 'Fix released'
        log = logging.getLogger('foo')
        update_review_queue(self.db, tasks, [], log)
        self.assertEqual(1, self.db.review_queue.count())

    def test_review_discards_approved_proposals(self):
        proposals = [get_mock_proposal(), get_mock_proposal()]
        proposals[0].queue_status = 'Approved'
        log = logging.getLogger('foo')
        update_review_queue(self.db, [], proposals, log)
        self.assertEqual(1, self.db.review_queue.count())
