# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from logging import getLogger

from charmworld.jobs.ingest import (
    add_files,
    IngestError,
    process_charm,
    scan_charm,
    store_charm_files,
)
from charmworld.models import (
    CharmFileSet,
    getfs,
)
from charmworld.testing import factory
from charmworld.testing import JobTestBase
from charmworld.utils import quote_key


class TestScanCharm(JobTestBase):

    def setUp(self):
        super(TestScanCharm, self).setUp()
        self.fs = getfs(self.db)
        self.log = getLogger('charm.scan')
        ignore, self.charm = factory.makeCharm(self.db, with_real_files=True)
        branch_dir, revision, path = self.use_context(factory.charm_branch())
        self.charm['branch_dir'] = branch_dir
        self.charm['store_data'] = {'digest': revision}
        self.charm['files'] = dict([
            (quote_key(cfile.filename), dict(cfile)) for cfile in
            store_charm_files(self.fs, branch_dir, self.charm, self.log)
        ])

    def test_config_metadata_with_dotted_keys(self):
        del self.charm['config']
        scan_charm(self.charm, self.fs, self.log)
        # The key 'dotted.name' is replaced with 'dotted%2Ename', and
        # 'foo.bar' is replaced with 'foo%2Ebar'.
        self.assertFalse('dotted.name' in self.charm)
        self.assertEqual('foo', self.charm['dotted%2Ename'])

    def test_options_format(self):
        del self.charm['config']
        scan_charm(self.charm, self.fs, self.log)
        expected = [
            {
                'name': 'source',
                'default': 'lp:charmworld',
                'type': 'string',
                'description': 'The bzr branch to pull the charmworld source'
                               ' from.'
            },
            {
                'name': 'revno',
                'default': -1,
                'type': 'int',
                'description': 'The revno of the bzr branch to use. -1 for'
                               ' current tip.'
            },
            {
                'name': 'execute-ingest-every',
                'default': 15,
                'type': 'int',
                'description': 'The amount of time (in minutes) that the'
                               ' ingest tasks for charmworld are executed'
            },
            {
                'name': 'error-email',
                'default': '',
                'type': 'str',
                'description': 'Address to mail errors to.'
            }
        ]
        self.assertItemsEqual(expected, self.charm['config']['options'])

    def test_short_interface_specification(self):
        # Charms can specify the interfaces they require or provide
        # in two forms: The long form requires for each relation
        # a dictionary that has a least the key 'interface'; the short
        # form specifies only the interface name for as a relation's
        # value. (See
        # https://juju.ubuntu.com/docs/charm.html#the-metadata-file
        # for more details.)
        #
        # normalize_interfaces() transforms the short form
        # into the long form for further processing and for storage.
        del self.charm['provides']
        del self.charm['requires']
        scan_charm(self.charm, self.fs, self.log)
        # The sample metadata file has entries for the short and long
        # form of the interface specification in the "provides" and
        # in the "requires" section. The short entries are converted
        # into the long form.
        expected = {
            'database': {'interface': 'mongodb'},
            'sql_db': {'interface': 'pgsql'},
        }
        self.assertEqual(expected, self.charm['requires'])
        expected = {
            'ssl_website': {'interface': 'https'},
            'website': {'interface': 'http'},
        }
        self.assertEqual(expected, self.charm['provides'])

    def test_short_interface_specification_reingest(self):
        # Reingestion of changes will update the charm in the db.
        self.charm['provides'] = {'api': {'interface': 'rest'}}
        self.charm['requires'] = {'database': {'interface': 'mysql'}}
        scan_charm(self.charm, self.fs, self.log)
        # The sample charm's metadata uses this information.
        p_expected = {
            'ssl_website': {'interface': 'https'},
            'website': {'interface': 'http'},
        }
        r_expected = {
            'database': {'interface': 'mongodb'},
            'sql_db': {'interface': 'pgsql'},
        }
        self.assertEqual(r_expected, self.charm['requires'])
        self.assertEqual(p_expected, self.charm['provides'])

    def test_invalid_metadata_raises_IngestError(self):
        metadata = 'invalid_syntax: ['
        metadata_file = factory.make_charm_file(
            self.db, self.charm, '/path/to/metadata.yaml', content=metadata)
        self.charm['files'][quote_key('metadata.yaml')] = {
            'fileid': metadata_file._id
        }
        with self.assertRaises(IngestError) as ctx:
            scan_charm(self.charm, self.fs, self.log)
        info = str(ctx.exception)
        self.assertTrue(info.startswith(
            'Invalid charm metadata ~charmers/charms/precise/sample'))

    def test_invalid_config_yaml_raises_IngestError(self):
        config = 'invalid_syntax: ['
        config_file = factory.make_charm_file(
            self.db, self.charm, '/path/to/config.yaml', content=config)
        self.charm['files'][quote_key('config.yaml')] = {
            'fileid': config_file._id
        }
        with self.assertRaises(IngestError) as ctx:
            scan_charm(self.charm, self.fs, self.log)
        info = str(ctx.exception)
        self.assertTrue(info.startswith(
            'Invalid charm config yaml. ~charmers/charms/precise/sample'))

    def test_short_url_label_generation(self):
        # charm['short_url'] starts with the owner name for
        # unpromulgated charms.
        charm = {
            'owner': 'foo',
            'series': 'precise',
            'name': 'bar',
            'promulgated': False,
        }
        new_charm = process_charm(charm)
        self.assertEqual('/~foo/precise/bar', new_charm['short_url'])
        # The label of unpromulgated charms starts with the owner name.
        self.assertEqual('~foo:precise/bar', new_charm['label'])
        # Promulgated charms do not have the owner name in the short URL
        # and in the label.
        charm = {
            'owner': 'charmers',
            'series': 'precise',
            'name': 'bar',
            'promulgated': True,
        }
        new_charm = process_charm(charm)
        self.assertEqual('/charms/precise/bar', new_charm['short_url'])
        self.assertEqual('precise/bar', new_charm['label'])
        # This is also true if the charm is not owned by charmes.
        charm = {
            'owner': 'foo',
            'series': 'precise',
            'name': 'bar',
            'promulgated': True,
        }
        new_charm = process_charm(charm)
        self.assertEqual('/charms/precise/bar', new_charm['short_url'])
        self.assertEqual('precise/bar', new_charm['label'])

    def test_charm_name_inconsistency(self):
        # As reported in bug 1202665, the LP source package name of a
        # charm and the name specified in metadata,yaml may be different.
        # This does no affect the retrieval of charm related files.
        with factory.charm_branch() as (branch_dir, revision, path):
            charm_data = {
                'owner': 'charmers',
                'series': 'precise',
                # The sample metadata file says "name: charmworld"; we
                # need a different name for this test.
                'name': 'differs-from-metadata',
                'bname': 'trunk',
                'store_data': {'digest': revision},
                'promulgated': True,
            }
            add_files(self.fs, branch_dir, charm_data, self.log)
        scan_charm(charm_data, self.fs, self.log)
        file_id = CharmFileSet.gen_fileid(charm_data, 'config.yaml')
        # A file with this ID exists; no exception is raised.
        CharmFileSet.get_by_id(self.fs, file_id)
        self.assertEqual('differs-from-metadata', charm_data['name'])
