# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

from contextlib import contextmanager

from mock import (
    create_autospec,
    patch,
)
from pyelasticsearch import ElasticSearch

from charmworld.testing import logger_buffer
from charmworld.jobs.config import (
    CHARM_QUEUE,
    BASKET_QUEUE,
    settings,
)
from charmworld.jobs.worker import (
    QueueWorker,
    main,
)
from charmworld.jobs.ingest import (
    DBIngestJob,
    UpdateCharmJob,
)
from charmworld.jobs.tests.test_ingest import charm_update_environment
from charmworld.jobs.utils import get_queue
from charmworld.search import ElasticSearchClient
from charmworld.testing import (
    factory,
    MongoTestBase,
)


class SuccessfulRun(DBIngestJob):

    def run(self, charm_data):
        charm_data['ingested'] = True


class FailureRun(DBIngestJob):

    def run(self, charm_data):
        if charm_data['charm'] == 'foo':
            raise Exception('Exception forced by job.')
        else:
            charm_data['ingested'] = True


class GatheringRun(DBIngestJob):

    def setup(self, db=None):
        super(GatheringRun, self).setup(db)
        self.charms = []

    def run(self, charm_data):
        self.charms.append(charm_data)


@contextmanager
def main_environment(charm, index_client, queue):
    """An environment in which worker.main() can be run."""
    with patch('charmworld.jobs.worker.configure_logging'):
        with charm_update_environment(charm, index_client):
            yield


def FauxWorkerFactory(test_case):
    class FauxWorker:
        run_called = False

        def __init__(self, *args):
            pass

        @classmethod
        def run(cls, run_forever):
            test_case.assertFalse(run_forever)
            cls.run_called = True

    return FauxWorker


class WorkerTest(MongoTestBase):

    def setUp(self):
        super(WorkerTest, self).setUp()
        self.charm_queue = get_queue(self.db, CHARM_QUEUE)
        self.basket_queue = get_queue(self.db, BASKET_QUEUE)
        self.charm_queue.put({'charm': 'foo'})
        self.charm_queue.put({'charm': 'bar'})

    def tearDown(self):
        self.charm_queue.clear()
        super(WorkerTest, self).tearDown()

    def test_basic_running(self):
        # Barring errors, the worker will go through its entire set of
        # ingest jobs for a charm.
        ingest_jobs = (
            SuccessfulRun(),
            GatheringRun(),
        )
        queue_jobs = {
            self.charm_queue: ingest_jobs
        }
        worker = QueueWorker(self.db, queue_jobs)
        worker.run()

        self.assertEqual(0, self.charm_queue.size())
        expected = [
            {'ingested': True, 'charm': 'foo'},
            {'ingested': True, 'charm': 'bar'},
        ]
        charms = ingest_jobs[1].charms
        self.assertEqual(expected, charms)

    def test_failure_while_running(self):
        ingest_jobs = (FailureRun(), GatheringRun())
        queue_jobs = {
            self.charm_queue: ingest_jobs
        }
        worker = QueueWorker(self.db, queue_jobs)
        worker.run()

        self.assertEqual(0, self.charm_queue.size())
        expected = [
            {'ingested': True, 'charm': 'bar'},
        ]
        charms = ingest_jobs[1].charms
        self.assertEqual(expected, charms)

    def test_stop_when_next_item_none(self):

        class LockedJobQueue:
            """Queue representing the state when a locked jobs exists."""

            @staticmethod
            def next():
                """A locked job cannot be processed."""
                return None

            @staticmethod
            def size():
                """If a queue has a locked job, its length > 0"""
                return 1

        ingest_jobs = (
            SuccessfulRun(),
            GatheringRun(),
        )
        worker = QueueWorker(self.db, {LockedJobQueue: ingest_jobs})
        worker.run()

    def test_ingest_job_contract(self):
        # All ingest jobs accept the right inputs.
        ingest_jobs = [create_autospec(UpdateCharmJob)()]
        worker = QueueWorker(self.db, {self.charm_queue: ingest_jobs})
        worker.run()
        self.assertEqual(0, self.charm_queue.size())

    def test_interval_is_int(self):
        worker = QueueWorker(self.db, {self.charm_queue: []})
        self.assertIs(int, type(worker.interval),
                      'QueueWorker.interval is a %s, not an int.' %
                      type(worker.interval))

    def run_main_with_exit(self):
        with patch('charmworld.jobs.worker.configure_logging'):
            with self.assertRaises(SystemExit) as e:
                main(['ingest-queued'], db=self.db)
        return e.exception.code

    def test_elasticsearch_failure(self):
        payload = factory.get_payload_json()
        factory.add_store_data(payload)
        self.charm_queue.clear()
        self.charm_queue.put(payload)
        charm = factory.get_charm_json(payload=payload)
        index_client = ElasticSearchClient(
            ElasticSearch('http://localhost:70'), 'foo')
        with patch.dict(settings,
                        values={'es_urls': 'http://localhost:70'}):
            with main_environment(charm, index_client, self.charm_queue):
                with logger_buffer('charm.worker') as handler:
                    exit_code = self.run_main_with_exit()
        self.assertEqual(40, handler.buffer[1].levelno)
        self.assertIn('Could not connect', handler.buffer[1].msg)
        self.assertEqual(1, exit_code)
        self.assertEqual(0, self.charm_queue.size())

    def test_job_for_bundles_is_run_if_bundle_queue_has_entries(self):
        # Basket queue entries are processed by run_job.
        queue = get_queue(self.db, BASKET_QUEUE)
        queue_entry = {'bundle': 'foo'}
        queue.put(queue_entry)
        with patch('charmworld.jobs.worker.run_job') as run_job:
            self.run_main_with_exit()
            queue_entries_run = [x[0][1] for x in run_job.call_args_list]
            self.assertIn(queue_entry, queue_entries_run)

    def test_worker_does_not_run_forever_by_default(self):
        FauxWorker = FauxWorkerFactory(self)
        with patch('charmworld.jobs.worker.QueueWorker', new=FauxWorker):
            self.run_main_with_exit()
        self.assertTrue(FauxWorker.run_called)

    def test_worker_runs_forever_if_command_line_says_to(self):
        FauxWorker = FauxWorkerFactory(self)
        with patch('charmworld.jobs.worker.QueueWorker', new=FauxWorker):
            self.run_main_with_exit()
        self.assertTrue(FauxWorker.run_called)
