# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from pyramid.decorator import reify
from pyramid.request import Request
from pyramid.security import has_permission
from pyramid.security import unauthenticated_userid
import velruse

from charmworld.models import getdb
from charmworld.models import UserMgr
from charmworld.search import ElasticSearchClient
from charmworld.ubuntusso import UBUNTU_SSO


class RequestWithUserAttribute(Request):
    """A Request factory that auto adds logged in user to request.

    All views can then access the current user via request.user. This is also
    available in the template contexts. It's the instantiated instance of a
    User() object.

    """
    @reify
    def user(self):
        userid = unauthenticated_userid(self)
        if userid is not None:
            # This should return None if the user doesn't exist
            # in the database.
            user = UserMgr.find_one(self.db, userid)
            return user

    @reify
    def login_url(self):
        """Openid url into request for ease of placing in templates."""
        return velruse.login_url(self, UBUNTU_SSO)

    def has_permission(self, perm):
        """Helper provided in the request context for checking permissions.

        This is most handy for use in templates where request is provided in
        context.

        """
        return has_permission(perm, self.context, self)

    @reify
    def db(self):
        """When required connect to the database for the request.

        We set the connection into the settings registry if this is the first
        call. getconnection will cheat and use that connection for subsequent
        calls.

        """
        settings = self.registry.settings
        db = getdb(settings['connection'], settings['mongo.database'])
        return db

    @reify
    def index_client(self):
        """Provide an index client, instantiating as needed."""
        return ElasticSearchClient.from_settings(self.registry.settings)
