"""Helpers to proof charms and bundles."""


class ProofError(Exception):

    def __init__(self, debug_info, msg):
        self.debug_info = debug_info
        self.msg = msg


class CharmProof(object):

    @staticmethod
    def check_config(charm, test_key, test_value):
        type_map = {
            'boolean': (bool,),
            'int': (int, ),
            'string': (str, unicode),
        }

        if not charm.options:
            raise ProofError([
                'Looking at charm id: ' + charm._id],
                'The charm has no options.'
            )

        if test_key not in charm.options:
            raise ProofError([
                'Looking at charm id: ' + charm._id,
                'Found config keys: ' + str(charm.options.keys())],
                'The charm has no option for: ' + test_key
            )

        # # The config key is a valid one for this charm. Check that the value
        # # is of the right type.
        specified_type = charm.options[test_key]['type']
        valid_types = type_map[specified_type]
        msg = "%s is not of type %s."
        if type(test_value) not in valid_types:
            raise ProofError(
                ['Looking at charm id: ' + charm._id],
                msg % (test_key, specified_type)
            )


# XXX: The code below should be ported over to the deployer or proof libary
# when we reconcile using the deployer code more. It needs to gain the whole
# idea of Proof that we've built and these helpers are part of it. Included
# here inline for sanity sake. The deployer does not currently perform these
# checks.
def verify_points_exist_and_same_type(check_requires, check_provides,
                                      requires, provides):
    # Give two service relation sets and specific interfaces to look for,
    # make sure they both exist and are of the same basic type "http".
    exist = False
    same_type = False
    require_type = None
    provide_type = None

    if check_requires in requires and check_provides in provides:
        exist = True

    if check_requires and check_requires in requires:
        require_type = requires[check_requires]['interface']
    if check_provides and check_provides in provides:
        provide_type = provides[check_provides]['interface']

    if require_type and provide_type and require_type == provide_type:
        same_type = True

    if not exist or not same_type:
        # Then this definition is invalid.
        msg = ('The requested relation %s to %s is incompatible between '
               'services.')
        raise ProofError(
            [('Endpoints are the same type on both services? ' +
              str(same_type)), ],
            msg % (check_requires, check_provides))
    else:
        return True


def get_list_of_shared_interfaces(service_one, service_two):
    # Generate a set of compatible interface types between services.
    common_interfaces = set()
    one_requires = service_one.requires
    one_provides = service_one.provides

    two_requires = service_two.requires
    two_provides = service_two.provides

    # Check the list of common interfaces between one_provides and two
    # requires.
    one_provides_interfaces = set()
    two_requires_interfaces = set()

    for name, data in one_provides.iteritems():
        one_provides_interfaces.add(data['interface'])
    for name, data in two_requires.iteritems():
        two_requires_interfaces.add(data['interface'])

    common_interfaces.update(
        one_provides_interfaces & two_requires_interfaces)

    # And check the common interfaces between one_requires and
    # two_provides.
    one_requires_interfaces = set()
    two_provides_interfaces = set()

    for name, data in one_requires.iteritems():
        one_requires_interfaces.add(data['interface'])
    for name, data in two_provides.iteritems():
        two_provides_interfaces.add(data['interface'])

    common_interfaces.update(
        one_requires_interfaces & two_provides_interfaces)

    return common_interfaces
