from mock import Mock

from charmworld.lib.cdeployer import parse_deployer
from charmworld.lib.proof import (
    CharmProof,
    get_list_of_shared_interfaces,
    ProofError,
    verify_points_exist_and_same_type,
)

from charmworld.models import Charm
from charmworld.testing import TestCase


class TestDeployerParser(TestCase):

    sample_deployer_file = """
        {
            "wiki": {
                "services": {
                    "wiki": {
                        "charm": "mediawiki",
                        "num_units": 2,
                        "branch": "lp:charms/precise/mediawiki",
                        "constraints": "mem=2"
                    }
                },
                "series": "precise" }}"""

    def test_parses_valid_yaml_file(self):
        """A deployer file should be parseable as yaml."""
        deployer_file = self.sample_deployer_file
        deployer = parse_deployer(deployer_file)
        self.assertEqual(deployer['wiki']['services'].keys(), ['wiki'])

    def test_failing_parsing_yaml_throws_proof_error(self):
        """If the yaml is not parse the yaml thows a ProofError"""
        deployer_file = '{'
        with self.assertRaises(ProofError) as exc:
            parse_deployer(deployer_file)

        self.assertEqual(
            'Could not parse the yaml provided.',
            exc.exception.msg
        )
        self.assertEqual(
            deployer_file,
            exc.exception.debug_info[0]
        )


class TestCharmProof(TestCase):
    """Run API tests against the proof API.

    Note: This is not run directly but mixed into the tests for API3
    """

    def make_fake_charm_config(self, options):
        charm = Charm({
            '_id': '/precise/id',
            'config': options
        })

        return charm

    def test_check_config_passes_when_correct(self):
        """Bundles might attempt to set a boolean config value string/int"""
        charm = Mock()
        charm.options = {
            'name': {
                'type': 'boolean',
                'default': True,
            }
        }
        charm._id = 'precise/test'

        res = CharmProof.check_config(charm, 'name', False)
        # There's no result and no exception thrown.
        self.assertIsNone(res)

    def test_check_config_fails_when_no_options_defined(self):
        """Bundles might attempt to set config values that are not valid"""
        charm = Mock()
        charm.options = {}
        charm._id = 'precise/test'

        with self.assertRaises(ProofError) as exc:
            CharmProof.check_config(charm, 'name', 'test')

        self.assertEqual(
            'The charm has no options.',
            exc.exception.msg
        )
        self.assertEqual(
            ['Looking at charm id: precise/test'],
            exc.exception.debug_info
        )

    def test_check_config_bool(self):
        """Bundles might attempt to set a boolean config value string/int"""
        charm = Mock()
        charm.options = {
            'name': {
                'type': 'boolean',
                'default': True,
            }
        }
        charm._id = 'precise/test'

        with self.assertRaises(ProofError) as exc:
            CharmProof.check_config(charm, 'name', 'test')

        self.assertEqual(
            'name is not of type boolean.',
            exc.exception.msg
        )
        self.assertEqual(
            ['Looking at charm id: precise/test'],
            exc.exception.debug_info
        )

    def test_check_config_fails_when_option_does_not_exist(self):
        """Bundles might attempt to set config values that do not exist."""
        charm = Mock()
        charm.options = {'foo': {}}
        charm._id = 'precise/test'

        with self.assertRaises(ProofError) as exc:
            CharmProof.check_config(charm, 'name', 'test')

        self.assertEqual(
            'The charm has no option for: name',
            exc.exception.msg
        )
        self.assertEqual(
            [
                'Looking at charm id: precise/test',
                'Found config keys: [\'foo\']'
            ],
            exc.exception.debug_info
        )


class TestBundleRelations(TestCase):

    def test_list_of_shared_interfaces(self):
        charm_one = Mock()
        charm_one.requires = {
            'db': {
                'interface': 'mysql',
            }
        }
        charm_one.provides = {
            'website': {
                'interface': 'http',
            }
        }

        charm_two = Mock()
        charm_two.requires = {
            'cache': {
                'interface': 'memcache',
            },
            'proxy': {
                'interface': 'http'
            }
        }
        charm_two.provides = {
            'db': {
                'interface': 'mysql',
            }
        }

        res = get_list_of_shared_interfaces(charm_one, charm_two)

        self.assertIn('http', res)
        self.assertIn('mysql', res)
        self.assertNotIn('website', res)
        self.assertNotIn('db', res)

    def test_no_shared_interfaces(self):
        charm_one = Mock()
        charm_one.requires = {
            'db': {
                'interface': 'mysql',
            }
        }
        charm_one.provides = {}

        charm_two = Mock()
        charm_two.requires = {}
        charm_two.provides = {
            'website': {
                'interface': 'http',
            }
        }

        res = get_list_of_shared_interfaces(charm_one, charm_two)
        self.assertEqual(0, len(res))

    def test_invalid_fully_qualified_endpoints(self):
        # relation = ['wiki:db', 'haproxy:cache']
        check_requires = 'db'
        check_provides = 'cache'

        requires = {
            'db': {
                'interface': 'mysql'
            }
        }
        provides = {
            'cache': {
                'interface': 'http'
            }
        }

        with self.assertRaises(ProofError) as exc:
            verify_points_exist_and_same_type(
                check_requires,
                check_provides,
                requires,
                provides)

        msg = ('The requested relation db to cache is incompatible between '
               'services.')
        self.assertEqual(
            msg,
            exc.exception.msg
        )
        self.assertEqual(
            'Endpoints are the same type on both services? False',
            exc.exception.debug_info[0]
        )

    def test_valid_fully_qualified_endpoints(self):
        # relation = ['wiki:website', 'haproxy:cache']
        check_requires = 'website'
        check_provides = 'cache'

        requires = {
            'website': {
                'interface': 'http'
            }
        }
        provides = {
            'cache': {
                'interface': 'http'
            }
        }

        res = verify_points_exist_and_same_type(
            check_requires,
            check_provides,
            requires,
            provides)

        # There's no ProofError raised. Just an empty function return value
        self.assertTrue(res)
