# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Migrate data for the mongodb instance.

Loads migrations from the latest available in the versions directory.

"""
import argparse
from collections import namedtuple
from datetime import datetime
import imp
from os import listdir
from os.path import abspath
from os.path import dirname
from os.path import join
import re

from pymongo.errors import OperationFailure

from charmworld.models import (
    CharmSource,
    getconnection,
    getdb,
)
from charmworld.qa_questions import iter_categories
from charmworld.search import ElasticSearchClient
from charmworld.utils import get_ini
from charmworld.utils import configure_logging


SCRIPT_TEMPLATE = """
# {description}

def upgrade(db):
    \"\"\"Complete this function with work to be done for the migration/update.

    db is the pymongo db instance for our datastore. Charms are in db.charms
    for instance.
    \"\"\"


"""


def str_to_filename(s):
    """Replaces spaces, quotes, and double underscores to underscores."""
    s = s.replace(' ', '_')\
         .replace('"', '_')\
         .replace("'", '_')\
         .replace(".", "_")
    while '__' in s:
        s = s.replace('__', '_')
    return s


class MigrationBeforeExodus(Exception):
    pass


class MissingExodusCollection(Exception):
    pass


class MissingExodusIndex(Exception):
    pass


class MultipleExodus(Exception):
    pass


class DataStore(object):
    """Communicate with the data store to determine version status."""

    def __init__(self, db):
        """Talk to the data store

        :param db: The mongo db connection.

        """
        self.db = db

    @property
    def current_version(self):
        """Return the current version of the data store."""
        found = self.db.migration_version.find_one({
            '_id': 'version'
        })

        if found:
            return found['version']
        else:
            return None

    def update_version(self, version):
        """Update the version number in the data store."""
        self.db.migration_version.save({
            '_id': 'version',
            'version': version,
            'date': datetime.utcnow(),
        })

    def version_datastore(self):
        """Init the data to track the current version in the datastore.

        The data store starts out with version 0.

        Raises exception if the store is already versioned.
        """
        if self.current_version is not None:
            raise Exception('Data store is already versioned: {0}'.format(
                self.current_version))
        self.db.migration_version.insert({
            '_id': 'version',
            'version': 0,
            'date': datetime.utcnow(),
        })


VersionInfo = namedtuple('VersionInfo', ['num', 'filename', 'is_exodus'])


class Versions(object):

    def __init__(self, path):
        """Collect version scripts and store them in self.versions

        """
        FILENAME_WITH_VERSION = re.compile(r'^(\d{3,}).*')
        self.versions_dir = path
        # Create temporary list of files, allowing skipped version numbers.
        files = listdir(path)
        self.versions = {}

        for name in files:
            if not name.endswith('.py'):
                # Skip .pyc and the like
                continue

            match = FILENAME_WITH_VERSION.match(name)
            if match:
                num = int(match.group(1))
                self.versions[num] = name
            else:
                pass  # Must be a helper file or something, let's ignore it.

    def initialize(self, datastore):
        datastore.version_datastore()
        for category_dict in iter_categories():
            datastore.db.qa.insert(category_dict)

    def ensure_initialized(self, datastore, init=False):
        if datastore.current_version is not None:
            return
        if init:
            self.initialize(datastore)
        else:
            raise Exception('Data store is not versioned')

    def iter_versions(self, current):
        return (i for i in sorted(self.versions.items()) if i[0] > current)

    def version_path(self, description, version):
        filename = "{0}_{1}.py".format(
            str(version).zfill(3),
            str_to_filename(description)
        )
        path = join(self.versions_dir, filename)
        return path, filename

    def create_new_version_file(self, description, content=None):
        """Create Python files for new version"""
        version = self.latest + 1

        if not description:
            raise ValueError('Please provide a short migration description.')
        if content is None:
            content = SCRIPT_TEMPLATE.format(description=description)
        path, filename = self.version_path(description, version)
        with open(path, 'w') as new_migration:
            new_migration.write(content)

        # Update our current store of version data.
        self.versions[version] = filename

        return filename

    @property
    def latest(self):
        """:returns: Latest version in Collection"""
        return max(self.versions.keys()) if len(self.versions) > 0 else 0

    def get_exodus(self, filename):
        module = self.import_version(filename)
        return getattr(module, 'exodus_update', None)

    def list_pending(self, current):
        return [VersionInfo(n, f, self.get_exodus(f) is not None) for n, f in
                self.iter_versions(current)]

    def _check_exodus(self, version_info):
        exodus_seen = False
        migration_seen = False
        for num, filename, is_exodus in version_info:
            if is_exodus:
                if migration_seen:
                    raise MigrationBeforeExodus(
                        'There are migrations scheduled before an exodus.')
                if exodus_seen:
                    raise MultipleExodus(
                        'There are multiple exoduses scheduled.')
                exodus_seen = True
            else:
                migration_seen = True

    def import_version(self, module_name):
        return imp.load_source(
            module_name.strip('.py'),
            join(self.versions_dir, module_name))

    def run_migration(self, db, index_client, module_name):
        module = self.import_version(module_name)
        getattr(module, 'upgrade')(db, index_client)

    def get_pending_source(self, source, number):
        db = source.collection.database
        pending_suffix = '_pending_%03d' % number
        pending_name = source.index_client.index_name + pending_suffix
        pending_index = source.index_client.get_aliasable_client(
            pending_name)
        pending_collection = db[source.collection.name + pending_suffix]
        return CharmSource(db, pending_index, pending_collection)

    def prepare_exodus(self, source, init=False):
        db = source.collection.database
        datastore = DataStore(db)
        self.ensure_initialized(datastore, init)
        current_version = datastore.current_version
        pending_migrations = self.list_pending(current_version)
        self._check_exodus(pending_migrations)
        if len(pending_migrations) == 0 or not pending_migrations[0].is_exodus:
            return None
        number, filename, is_exodus = pending_migrations[0]
        pending = self.get_pending_source(source, number)
        temp_collection = db[pending.collection.name + 'temp']
        with source.index_client.replacement() as temp_index:
            db.create_collection(temp_collection.name)
            temp = CharmSource(db, temp_index, temp_collection)
            self.get_exodus(filename)(source, temp, current_version)
            aliased = pending.index_client.get_aliased()
            temp_collection.rename(pending.collection.name, dropTarget=True)
            pending.index_client.update_aliased(temp_index.index_name, aliased)
        return pending

    def finalize_exodus(self, source, version_number):
        pending = self.get_pending_source(source, version_number)
        aliased = source.index_client.get_aliased()
        pending_aliased = pending.index_client.get_aliased()
        if len(pending_aliased) == 0:
            raise MissingExodusIndex(
                'Exodus index "%s" does not exist.' %
                pending.index_client.index_name)
        if len(pending_aliased) != 1:
            raise AssertionError('More than index aliased to pending.')
        try:
            pending.collection.rename(source.collection.name,
                                      dropTarget=True)
        except OperationFailure as e:
            if e.message.endswith('source namespace does not exist'):
                raise MissingExodusCollection(
                    'Exodus collection "%s" does not exist.' %
                    pending.collection.name)
            raise
        source.index_client.update_aliased(
            pending_aliased[0], aliased)
        pending.index_client.update_aliased(None, [pending_aliased[0]])
        for alias_name in aliased:
            source.index_client.get_related_client(alias_name).delete_index()

    def upgrade(self, datastore, index_client, init):
        """Run `upgrade` methods for required version files.

        :param datastore: An instance of DataStore

        """
        self.ensure_initialized(datastore.db, init)
        pending_versions = self.list_pending(datastore.current_version)
        self._check_exodus(pending_versions)

        if len(pending_versions) == 0:
            return None

        for num, module_name, is_exodus in pending_versions:
            if is_exodus:
                source = CharmSource(datastore.db, index_client)
                self.finalize_exodus(source, pending_versions[0][0])
            else:
                self.run_migration(datastore.db, index_client, module_name)
            datastore.update_version(num)
        return num


def parse_args():
    desc = "Mongo migration tool."
    parser = argparse.ArgumentParser(description=desc)
    subparsers = parser.add_subparsers(help='sub-command help')

    parser_current = subparsers.add_parser('current')
    parser_current.set_defaults(func=Commands.current)

    parser_latest = subparsers.add_parser('latest')
    parser_latest.set_defaults(func=Commands.latest)

    parser_new = subparsers.add_parser('new')
    parser_new.add_argument(
        '-d', '--description', action='store',
        required=True,
        help='The description of the new migration.')
    parser_new.set_defaults(func=Commands.new)

    parser_upgrade = subparsers.add_parser('upgrade')
    parser_upgrade.set_defaults(func=Commands.upgrade)
    parser_upgrade.add_argument(
        '-i', '--init', action='store_true',
        default=False,
        help='Auto init the database if not already init.')

    parser_prepare_upgrade = subparsers.add_parser('prepare-upgrade')
    parser_prepare_upgrade.set_defaults(func=Commands.prepare_upgrade)
    parser_prepare_upgrade.add_argument(
        '-i', '--init', action='store_true',
        default=False,
        help='Auto init the database if not already init.')

    args = parser.parse_args()
    return args


class Commands(object):
    """Container for the various commands the cli can execute"""

    @staticmethod
    def get_datastore(ini):
        connection = getconnection(ini)
        db = getdb(connection, ini.get('mongo.database'))
        ds = DataStore(db)
        return ds

    @classmethod
    def current(cls, ini, args):
        """Fetch the current migration version of the data store."""
        ds = cls.get_datastore(ini)
        version = ds.current_version

        if version is None:
            print 'Data store is not currently versioned.'
        else:
            print version

    @staticmethod
    def latest(ini, args):
        """Determine the latest migration version available."""
        migrations = Versions(ini['migrations'])
        print migrations.latest

    @staticmethod
    def new(ini, args):
        """Generate a new migration file to be completed."""
        migrations = Versions(ini['migrations'])
        filename = migrations.create_new_version_file(args.description)
        print "Created new migration: {0}".format(filename)

    @classmethod
    def prepare_upgrade(cls, ini, args):
        source = CharmSource.from_settings(ini)
        migrations = Versions(ini['migrations'])
        migrations.prepare_exodus(source, args.init)

    @classmethod
    def upgrade(cls, ini, args):
        """Upgrade the data store to the latest available migration."""
        ds = cls.get_datastore(ini)
        index_client = ElasticSearchClient.from_settings(ini)
        migrations = Versions(ini['migrations'])
        new_version = migrations.upgrade(ds, index_client, args.init)

        if new_version is None:
            print 'There are no new migrations to apply'
        else:
            print "Updated the datastore to version: {0}".format(new_version)


def get_migration_path():
    return join(abspath(dirname(__file__)), 'versions')


def main():
    """Target for the console entry point."""
    configure_logging()
    args = parse_args()
    ini = get_ini()

    # Add the migration path to the ini.
    ini['migrations'] = get_migration_path()

    args.func(ini, args)


if __name__ == '__main__':
    main()
