# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import textwrap
import yaml

from charmworld.models import (
    Bundle,
    CharmSource,
    store_bundles,
)
from charmworld.search import (
    BUNDLE,
    IndexMissing,
)
from charmworld.testing import (
    factory,
    MongoTestBase,
)
from charmworld.migrations import migrate
from charmworld.migrations.migrate import Versions

import mock


class MigrationTestBase(MongoTestBase):

    def setUp(self):
        super(MigrationTestBase, self).setUp()
        self.versions = Versions(migrate.get_migration_path())
        self.use_context(mock.patch.object(migrate, 'configure_logging'))


class TestMigration019(MigrationTestBase):

    def setUp(self):
        super(TestMigration019, self).setUp()
        self.use_index_client()

    def test_new_db_and_index_have_all_old_data(self):
        source = CharmSource.from_request(self)
        charm_1 = factory.get_charm_json(
            name='one',
            provides={
                'website': {
                    'interface': 'http',
                },
            },
            requires={
                'database': {
                    'interface': 'mysql',
                }
            })
        source.save(charm_1)
        charm_2 = factory.get_charm_json(
            name='two',
            provides={
                'database': {
                    'interface': 'mysql',
                }
            })
        source.save(charm_2)

        exodus_update = self.versions.get_exodus(
            '019_fixed_mapping_for_interfaces.py')
        target = self.versions.get_pending_source(source, 19)
        target.index_client.create_index()
        target.index_client.put_mapping()
        self.addCleanup(target.index_client.delete_index)
        exodus_update(source, target, 19)
        self.assertEqual(2, target.collection.find().count())
        self.assertEqual(2, target.index_client.search('mysql')['matches'])
        self.assertEqual(1, target.index_client.search('http')['matches'])


class TestMigration020(MigrationTestBase):

    def setUp(self):
        super(TestMigration020, self).setUp()
        self.use_index_client()

    def test_review_entries_updated(self):
        review_entries = [
            factory.make_review_entry(),
            factory.make_review_entry()]
        review_entries[0]['item'] = 'code.launchpad.net'
        review_entries[1]['item'] = 'bugs.launchpad.net'
        self.db.review_queue.insert(review_entries)
        self.versions.run_migration(
            self.db, self.index_client,
            '020_adds_type_field_to_existing_review_items.py')
        results = list(self.db.review_queue.find())
        self.assertEqual(2, len(results))
        for entry in results:
            if entry['item'] == 'code.launchpad.net':
                self.assertEqual('proposal', entry['type'])
            elif entry['item'] == 'bugs.launchpad.net':
                self.assertEqual('task', entry['type'])
            else:
                self.fail()

    def test_fails_on_unknown(self):
        review_entries = [factory.make_review_entry()]
        review_entries[0]['item'] = 'example.com'
        self.db.review_queue.insert(review_entries)
        self.assertRaises(
            Exception,
            self.versions.run_migration,
            self.db, self.index_client,
            '020_adds_type_field_to_existing_review_items.py')


class TestExodusRestrictions(MigrationTestBase):

    def test_exodus_restrictions(self):
        """Ensure that deploy-from-scratch does not violate exodus rules."""
        self.versions._check_exodus(self.versions.list_pending(0))
