# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from contextlib import contextmanager
import logging
import os
import shutil
from StringIO import StringIO
from tempfile import mkdtemp
import unittest
from urllib2 import URLError

from requests import Response
from httmock import HTTMock
import mock
from mongoqueue import MongoQueue
from pyramid import testing
from webtest import TestApp

from charmworld import main
from charmworld.models import getconnection
from charmworld.models import getdb
from charmworld.models import getfs
from charmworld.routes import build_routes
from charmworld.search import (
    ElasticSearchClient,
    IndexMissing,
)
from charmworld.testing.factory import makeSSOResponse
from charmworld.utils import get_ini


INI = get_ini()
MONGO_URL = os.environ.get('TEST_MONGODB')
MONGO_DATABASE = INI.get('mongo.database')
# Update the ini values with the database fired up by the nosemongo plugin.
INI['mongo.url'] = MONGO_URL

REAL_FILE_CHARM_DIR = os.path.join(os.path.split(__file__)[0], 'data')


@contextmanager
def logger_buffer(log_name):
    logger = logging.getLogger(log_name)
    handler = logging.handlers.BufferingHandler(20)
    logger.addHandler(handler)
    try:
        yield handler
    finally:
        logger.removeHandler(handler)


class TestCase(unittest.TestCase):

    def get_handler(self, log_name):
        return self.use_context(logger_buffer(log_name))

    def use_context(self, context):
        def cleanup():
            context.__exit__(None, None, None)

        result = context.__enter__()
        # Only register the cleaup after __enter__ is successfully called.
        self.addCleanup(cleanup)
        return result

    def use_index_client(self, put_mapping=True, aliased=False):
        self.index_client = self.use_context(
            temp_index_client(put_mapping=put_mapping, aliased=aliased))
        return self.index_client

    @contextmanager
    def assertNoException(self):
        with self.assertRaises(AssertionError):
            with self.assertRaises(Exception):
                yield


class MongoTestBase(TestCase):

    @property
    def connection(self):
        if getattr(self, '_connection', None) is None:
            self._connection = getconnection(INI, safe=True)
        return self._connection

    def setUp(self):
        self.db = getdb(self.connection, MONGO_DATABASE)
        self.fs = getfs(self.db)

    def tearDown(self):
        # Drop the Mongo database.
        self.connection.drop_database(MONGO_DATABASE)


class WebTestBase(MongoTestBase):

    def setUp(self):
        super(WebTestBase, self).setUp()
        app = main({}, **INI)
        self.app = TestApp(app)

    def tearDown(self):
        self.app = None
        super(WebTestBase, self).tearDown()


class ViewTestBase(MongoTestBase):

    def setUp(self):
        super(ViewTestBase, self).setUp()
        self.config = testing.setUp()
        self.index_client = None

    def getRequest(self, params=None, post=None):
        request = testing.DummyRequest(params=params, post=post)
        request.db = self.db
        request.index_client = self.index_client
        return request

    def tearDown(self):
        testing.tearDown()
        super(ViewTestBase, self).tearDown()

    def enable_routes(self):
        build_routes(self.config)
        self.config.scan('charmworld.views')


class RemoteURLBase(MongoTestBase):

    url_prefix = "https://jenkins.qa.ubuntu.com/job/"
    url_debug = False

    def setUp(self):
        super(RemoteURLBase, self).setUp()
        self.url_result_map = {"": ""}

    def lookup_url_result(self, url):
        if self.url_debug:
            print "url", url
        short_url = url[len(self.url_prefix):]
        if short_url in self.url_result_map:
            return StringIO(self.url_result_map[short_url])
        raise URLError(url)

    @contextmanager
    def url_mock(self):
        def httmock(url, request):
            try:
                result = self.lookup_url_result(request.url)
                return result.getvalue()
            except URLError:
                return response404()
        with mock.patch("urllib2.urlopen") as urlopen:
            urlopen.side_effect = self.lookup_url_result

            with HTTMock(httmock):
                yield


def response404(url=None, req=None):
    # Work around https://github.com/patrys/httmock/issues/3 by
    # making a Response that always evaluates to True
    class TrueResponse(Response):

        def __nonzero__(self):
            return True

    response = TrueResponse()
    response.status_code = 404
    return response


class JobTestBase(MongoTestBase):

    def patch_ingest_job(self, job, methods):
        # Mocks multiple attr on a job.
        old_methods = []
        for m in methods:
            old_methods.append(getattr(job, m))
            setattr(job, m, mock.Mock())

        def cleanup():
            for i, m in enumerate(methods):
                setattr(job, m, old_methods[i])
        self.addCleanup(cleanup)

    def get_mongo_queues(self, data=None):
        in_queue = MongoQueue(self.db.in_queue, "in-queue")
        out_queue = MongoQueue(self.db.out_queue, "out-queue")

        if data:
            in_queue.put(data)

        def cleanup():
            in_queue.clear()
            out_queue.clear()
        self.addCleanup(cleanup)
        return (in_queue, out_queue)


@contextmanager
def login_request(app, groups=None):
    """Context manager to help functionally test a logged in user account.

    :param app: The application constructed for testing. Usually self.app when
                using WebTestBase.
    :return response: The post-logged in WebTest response object.
    """
    sso_resp = makeSSOResponse(groups=groups)
    resp = app.get(
        '/auth_callback',
        params=sso_resp,
        status=302)
    resp = resp.follow()
    yield resp


def get_bad_mock(exception_type=Exception, *args):
    """Convenience helper to generate a mock that will raise a particular
    exception with necessary messages.
    """
    if not len(args):
        args = ["Exception forced by mock."]
    return mock.Mock(side_effect=exception_type(*args))


@contextmanager
def temp_index_client(name='temp_index', put_mapping=True, aliased=False):
    client = ElasticSearchClient.from_settings(INI, name)
    try:
        client.delete_index()
    except IndexMissing:
        pass
    if aliased:
        alias_client = client.get_aliasable_client()
        alias_client.create_index()
        client.update_aliased(alias_client.index_name, [])
    else:
        client.create_index()
    try:
        if put_mapping:
            client.put_mapping()
        yield client
    finally:
        client.delete_index()


def force_index_client(index_client):
    """Use a specific index name.

    This returns a context manager in whose context
    ElasticSearchClient.from_settings will use the supplied client's index
    name.
    """
    old_from_settings = ElasticSearchClient.from_settings

    @classmethod
    def from_settings(cls, settings):
        client = old_from_settings(settings, index_client.index_name)
        return client
    return mock.patch.object(ElasticSearchClient, 'from_settings',
                             from_settings)


@contextmanager
def temp_dir():
    path = mkdtemp()
    try:
        yield path
    finally:
        shutil.rmtree(path)


def load_data_file(filename):
    """Load files from the testing/data paths."""
    data_path = os.path.join(
        os.path.dirname(__file__),
        'data',
        filename,
    )
    return open(data_path).read()
