# Copyright 2012, 2013 Marco Ceppi, Canonical Ltd.  This software is
# licensed under the GNU Affero General Public License version 3 (see
# the file LICENSE).

"""Unit tests for our application models."""
from bzrlib.branch import Branch
from datetime import (
    date,
    datetime,
)
import hashlib
import logging
import md5
import os
from os.path import dirname
from os.path import join
from os.path import split
import stat
from textwrap import dedent

from mock import patch
from pyelasticsearch.exceptions import (
    ElasticHttpError,
)
import pymongo
import yaml

from charmworld.models import (
    acquire_session_secret,
    Bundle,
    BundledCharmDescription,
    Charm,
    CharmFile,
    CharmFileSet,
    CharmSource,
    construct_charm_id,
    FeaturedSource,
    find_charms,
    _find_charms,
    make_bundle_doc,
    options_to_storage,
    QAData,
    QADataSource,
    resolve_charm_from_description,
    slurp_files,
    storage_to_options,
    store_bundles,
    sync_index,
    UserMgr,
)
from charmworld.search import (
    BUNDLE,
    CHARM,
)
from charmworld.testing import (
    factory,
    force_index_client,
    INI,
    MongoTestBase,
    temp_dir,
    TestCase,
)
from charmworld.testing.factory import (
    bundle_branch,
    charm_branch,
)
from charmworld.utils import (
    quote_key,
    quote_yaml,
    read_locked,
)


class TestConstructCharmID(TestCase):

    def test_construct_charm_id(self):
        payload = factory.get_payload_json(owner='foo', series='bar',
                                           name='baz', promulgated=False)
        factory.add_store_data(payload, revision=25, promulgated=False)
        charm_id = construct_charm_id(payload)
        self.assertEqual('~foo/bar/baz/25', charm_id)

    def test_construct_charm_id_promulgated(self):
        # Promulgation has no effect on this ID.
        payload = factory.get_payload_json(owner='foo', series='bar',
                                           name='baz', promulgated=True)
        factory.add_store_data(payload, revision=25, promulgated=True)
        charm_id = construct_charm_id(payload)
        self.assertEqual('~foo/bar/baz/25', charm_id)

    def test_construct_charm_id_error(self):
        payload = factory.get_payload_json(owner='foo', series='bar',
                                           name='baz')
        factory.add_store_data(payload, charm_error=True, revision=25)
        charm_id = construct_charm_id(payload)
        self.assertEqual('~foo/bar/baz/0', charm_id)

    def test_construct_charm_id_no_revision(self):
        payload = factory.get_payload_json(owner='foo', series='bar',
                                           name='baz')
        charm_id = construct_charm_id(payload, use_revision=False)
        self.assertEqual('~foo/bar/baz', charm_id)


class CharmTestCase(TestCase):
    """Unit tests for the Charm model that wraps charm representations."""

    def test_init_with_invalid_charm(self):
        charm_data = {}
        charm = Charm(charm_data)
        self.assertIs(charm_data, charm._raw_representation)
        self.assertEqual(charm._COMMON_REPRESENTATION, charm._representation)

    def test_init_with_valid_charm(self):
        charm_data = factory.get_charm_json(
            name='name', owner='owner', series='series', promulgated=True)
        charm = Charm(charm_data)
        self.assertIs('name', charm._representation['name'])
        self.assertIs('owner', charm._representation['owner'])
        self.assertIs('series', charm._representation['series'])
        self.assertIs(True, charm._representation['promulgated'])

    def test_eq(self):
        charm_1 = Charm(factory.get_charm_json(
            owner='owner', series='series', name='name'))
        charm_2 = Charm(factory.get_charm_json(
            owner='other', series='series', name='name'))
        charm_3 = Charm(factory.get_charm_json(
            owner=charm_1.owner, series=charm_1.series, name=charm_1.name))
        self.assertNotEqual(charm_1, charm_2)
        self.assertEqual(charm_1, charm_3)

    def test_address(self):
        # The address property is a string.
        charm_data = factory.get_charm_json(
            owner='owner', series='series', name='name')
        charm = Charm(charm_data)
        self.assertEqual(
            'cs:~owner/series/name', charm.address)

    def test_branch_spec(self):
        # The branch_spec property is a string.
        charm_data = factory.get_charm_json(
            owner='owner', series='series', name='name')
        charm = Charm(charm_data)
        self.assertEqual(
            '~owner/charms/series/name/trunk', charm.branch_spec)

    def test_bzr_branch(self):
        # The bzr_branch property is a string.
        charm_data = factory.get_charm_json(
            owner='owner', series='series', name='name')
        charm = Charm(charm_data)
        self.assertEqual(
            'lp:~owner/charms/series/name/trunk', charm.bzr_branch)

    def test_owner(self):
        # The owner property is a string.
        charm_data = factory.get_charm_json(owner='owner')
        charm = Charm(charm_data)
        self.assertEqual('owner', charm.owner)

    def test_series(self):
        # The series property is a string.
        charm_data = factory.get_charm_json(series='series')
        charm = Charm(charm_data)
        self.assertEqual('series', charm.series)

    def test_name(self):
        # The name property is a string.
        charm_data = factory.get_charm_json(name='name')
        charm = Charm(charm_data)
        self.assertEqual('name', charm.name)

    def test_bname(self):
        # The bname property is a string.
        charm_data = factory.get_charm_json(bname='trunk')
        charm = Charm(charm_data)
        self.assertEqual('trunk', charm.bname)

    def test_commit(self):
        # The commit property is a revid string.
        charm_data = factory.get_charm_json()
        charm_data['commit'] = 'joe@example.com-20120723223205-babble'
        charm = Charm(charm_data)
        self.assertEqual(
            'joe@example.com-20120723223205-babble', charm.commit)
        # The default is an empty string that happens when a branch is empty.
        charm = Charm({})
        self.assertEqual('', charm.commit)

    def test_distro_series(self):
        # The distro_series property is a list of series names.
        charm_data = factory.get_charm_json()
        charm_data['distro_series'] = distro_series = ['series']
        charm = Charm(charm_data)
        self.assertIs(distro_series, charm.distro_series)
        # The default is an empty list.
        charm = Charm({})
        self.assertEqual([], charm.distro_series)

    def test_promulgated(self):
        # The promulgated property is a bool.
        charm_data = factory.get_charm_json(promulgated=True)
        charm = Charm(charm_data)
        self.assertIs(True, charm.promulgated)
        # The default is False.
        charm = Charm({})
        self.assertIs(False, charm.promulgated)

    def test_id(self):
        # The _id property is string string based on the branch_spec.
        charm_data = factory.get_charm_json(
            owner='owner', series='series', name='name')
        charm_data['_id'] = '~owner/charms/series/name/trunk'
        charm = Charm(charm_data)
        self.assertEqual('~owner/charms/series/name/trunk', charm._id)
        # The _id property is actually arbitrary.
        charm = Charm({'_id': 'boing'})
        self.assertEqual('boing', charm._id)

    def test_date_created(self):
        # The date_created property is an ISO timestamp.
        date_created = factory.random_date()
        charm_data = factory.get_charm_json(date_created=date_created)
        charm = Charm(charm_data)
        self.assertEqual(date_created.isoformat(), charm.date_created)
        # The default is None, which happens when the branch cannot be found.
        charm = Charm({})
        self.assertEquals(None, charm.date_created)

    def test_error(self):
        # The error property is a string.
        charm_data = {'error': 'Paris is burning'}
        charm = Charm(charm_data)
        self.assertEqual('Paris is burning', charm.error)
        # The default is an empty string.
        charm = Charm({})
        self.assertEquals('', charm.error)

    def test_branch_dir(self):
        # The branch_dir property is a string.
        charm_data = {'branch_dir': '/var/charms/series/owner/name/bname'}
        charm = Charm(charm_data)
        self.assertEqual(
            '/var/charms/series/owner/name/bname', charm.branch_dir)
        # The default is an empty string.
        charm = Charm({})
        self.assertEquals('', charm.branch_dir)

    def test_files(self):
        # The files property is a dict.
        files = {
            'install': {
                'filename': 'install',
                'subdir': 'hooks',
            },
        }
        charm_data = factory.get_charm_json(files=files)
        charm = Charm(charm_data)
        self.assertIs(files, charm.files)
        # The default is an empty string.
        charm = Charm({})
        self.assertEquals({}, charm.files)

    def test_downloads(self):
        """The downloads property is an int."""
        charm_data = {'downloads': 30}
        charm = Charm(charm_data)
        self.assertEqual(30, charm.downloads)
        # The default is zero.
        charm = Charm({})
        self.assertEquals(0, charm.downloads)

    def test_downloads_in_past_30_days(self):
        # The downloads_in_past_30_days property is an int.
        charm_data = {'downloads_in_past_30_days': 30}
        charm = Charm(charm_data)
        self.assertEqual(30, charm.downloads_in_past_30_days)
        # The default is zero.
        charm = Charm({})
        self.assertEquals(0, charm.downloads_in_past_30_days)

    def test_downloads_in_past_7_days(self):
        # The downloads_in_past_7_days property is an int.
        charm_data = {'downloads_in_past_7_days': 7}
        charm = Charm(charm_data)
        self.assertEqual(7, charm.downloads_in_past_7_days)
        # The default is zero.
        charm = Charm({})
        self.assertEquals(0, charm.downloads_in_past_7_days)

    def test_downloads_in_past_half_year(self):
        # The downloads_in_past_half_year property is an int.
        charm_data = {'downloads_in_past_half_year': 180}
        charm = Charm(charm_data)
        self.assertEqual(180, charm.downloads_in_past_half_year)
        # The default is zero.
        charm = Charm({})
        self.assertEquals(0, charm.downloads_in_past_half_year)

    def test_store_url(self):
        # The store_url property is a string.
        charm_data = {'store_url': 'cs:~owner/series/name-1'}
        charm = Charm(charm_data)
        self.assertEqual('cs:~owner/series/name-1', charm.store_url)
        # The default is an empty sting..
        charm = Charm({})
        self.assertEquals('', charm.store_url)

    def test_store_revision(self):
        # The store_revision property is an int.
        charm_data = {'store_data': {'revision': 17}}
        charm = Charm(charm_data)
        self.assertEqual(17, charm.store_revision)
        # The default is 1 if revision is missing from store_data.
        charm = Charm({'store_data': {}})
        self.assertEquals(1, charm.store_revision)
        # The default is 1 if there is no store_data.
        charm = Charm({})
        self.assertEquals(1, charm.store_revision)

    def test_store_data(self):
        # The store_data property dict.
        charm_data = factory.get_charm_json()
        store_data = {
            'address': 'cs:~owner/series/name-1',
            'store_checked': 'Thu Dec  6 17:42:05 2012',
            'digest': 'jdoe@example.com-20120723223205-babble',
            'revision': 1,
            'sha256': '107c230dc3481f8b8de9decfff6439a23caae2ea670bddb413ddf1',
        }
        charm_data['store_data'] = store_data
        charm = Charm(charm_data)
        self.assertIs(store_data, charm.store_data)
        # The default is an empty dict.
        charm = Charm({})
        self.assertEqual({}, charm.store_data)

    def test_proof(self):
        # The proof property is a dict.
        proof = {
            'w': [' no README file']
        }
        charm_data = factory.get_charm_json()
        charm_data['proof'] = proof
        charm = Charm(charm_data)
        self.assertIs(proof, charm.proof)
        # The default is an empty dict.
        charm = Charm({})
        self.assertEqual({}, charm.proof)

    def test_tests(self):
        # The tests property is a dict.
        tests = {
            'ec2': 'PASS'
        }
        charm_data = factory.get_charm_json()
        charm_data['tests'] = tests
        charm = Charm(charm_data)
        self.assertIs(tests, charm.tests)
        # The default is an empty dict.
        charm = Charm({})
        self.assertEqual({}, charm.tests)

    def test_test_results(self):
        # The test_results property is a dict.
        test_results = {
            'openstack': 987654321
        }
        charm_data = factory.get_charm_json()
        charm_data['test_results'] = test_results
        charm = Charm(charm_data)
        self.assertIs(test_results, charm.test_results)
        # The default is an empty dict.
        charm = Charm({})
        self.assertEqual({}, charm.test_results)

    def test_changes(self):
        # The changes property is a list of revision info dicts.
        changes = [factory.makeChange()]
        charm_data = factory.get_charm_json(changes=changes)
        charm = Charm(charm_data)
        self.assertIs(changes, charm.changes)
        # The default is an empty list.
        charm = Charm({})
        self.assertEqual([], charm.changes)

    def test_first_change(self):
        # The first_change property is a revision info dict.
        change = factory.makeChange()
        charm_data = factory.get_charm_json()
        charm_data['first_change'] = change
        charm = Charm(charm_data)
        self.assertIs(change, charm.first_change)
        # The default is some dummy data.
        charm = Charm({})
        expected = {
            'authors': [],
            'committer': '',
            'created': None,
            'message': '',
            'revno': 0,
        }
        self.assertEqual(expected, charm.first_change)

    def test_last_change(self):
        # The last_change property is a revision info dict.
        change = factory.makeChange()
        charm_data = factory.get_charm_json()
        charm_data['last_change'] = change
        charm = Charm(charm_data)
        self.assertIs(change, charm.last_change)
        # The default is some dummy data.
        charm = Charm({})
        expected = {
            'authors': [],
            'committer': '',
            'created': None,
            'message': '',
            'revno': 0,
        }
        self.assertEqual(expected, charm.last_change)

    def test_revision(self):
        # The revision property is an int.
        charm_data = factory.get_charm_json(revision=5)
        charm = Charm(charm_data)
        self.assertEqual(5, charm.revision)
        # The default is zero.
        charm = Charm({})
        self.assertEquals(0, charm.revision)

    def test_summary(self):
        # The summary property is string.
        summary = 'text'
        charm_data = factory.get_charm_json()
        charm_data['summary'] = summary
        charm = Charm(charm_data)
        self.assertIs(summary, charm.summary)
        # The default is an empty string.
        charm = Charm({})
        self.assertEqual('', charm.summary)

    def test_description(self):
        # The description property is string.
        description = 'text'
        charm_data = factory.get_charm_json()
        charm_data['description'] = description
        charm = Charm(charm_data)
        self.assertIs(description, charm.description)
        # The default is an empty string.
        charm = Charm({})
        self.assertEqual('', charm.description)

    def test_categories(self):
        # The categories property is a list of category names.
        categories = ['cache-proxy', 'app-servers']
        charm_data = factory.get_charm_json(categories=categories)
        charm = Charm(charm_data)
        self.assertEqual(categories, charm.categories)
        # The default is an empty list.
        charm = Charm({})
        self.assertEqual([], charm.categories)

    def test_categories_are_official(self):
        # Unofficial categories are filtered out of the list.
        categories = ['application']
        charm_data = factory.get_charm_json(categories=categories)
        charm = Charm(charm_data)
        self.assertEqual(
            ['applications', 'app-servers', 'cache-proxy', 'databases',
             'file-servers', 'misc'],
            charm.OFFICIAL_CATEGORIES)
        self.assertEqual([], charm.categories)

    def test_maintainer(self):
        # The maintainer property is a list of category names.
        maintainer = ['Jane <jane@example.com>', 'Ian <ian@example.com>']
        charm_data = factory.get_charm_json(maintainer=maintainer)
        charm = Charm(charm_data)
        self.assertIs(maintainer, charm.maintainer)
        # The default is an empty string.
        charm = Charm({})
        self.assertEqual('', charm.maintainer)

    def test_subordinate(self):
        # The subordinate property is a bool.
        charm_data = factory.get_charm_json(subordinate=True)
        charm = Charm(charm_data)
        self.assertIs(True, charm.subordinate)
        # The default is False.
        charm = Charm({})
        self.assertIs(False, charm.subordinate)

    @staticmethod
    def make_provides():
        return {
            "website": {
                "interface": "https",
            }
        }

    def test_provides(self):
        # The provides property is a dict of charm options.
        provides = self.make_provides()
        charm_data = factory.get_charm_json(provides=provides)
        charm = Charm(charm_data)
        self.assertIs(provides, charm.provides)
        # The default is an empty dict.
        charm = Charm({})
        self.assertEqual({}, charm.provides)

    def test_i_provides(self):
        # The i_provides property is a list of interface names.
        provides = self.make_provides()
        charm_data = factory.get_charm_json(provides=provides)
        charm = Charm(charm_data)
        self.assertIs(charm_data['i_provides'], charm.i_provides)
        # The default is an empty list.
        charm = Charm({})
        self.assertEqual([], charm.i_provides)

    @staticmethod
    def make_requires():
        return {
            'database': {
                'interface': 'mongodb',
            }
        }

    def test_requires(self):
        # The requires property is a dict of charm options.
        requires = self.make_requires()
        charm_data = factory.get_charm_json(requires=requires)
        charm = Charm(charm_data)
        self.assertIs(requires, charm.requires)
        # The default is an empty dict.
        charm = Charm({})
        self.assertEqual({}, charm.requires)

    def test_i_requires(self):
        # The i_provides property is a list of interface names.
        requires = self.make_requires()
        charm_data = factory.get_charm_json(requires=requires)
        charm = Charm(charm_data)
        self.assertIs(charm_data['i_requires'], charm.i_requires)
        # The default is an empty list.
        charm = Charm({})
        self.assertEqual([], charm.i_requires)

    def test_config(self):
        # The config property is a dict of charm options.
        config = {'options': {'key': {'default': 'value', 'type': 'string'}}}
        charm_data = factory.get_charm_json(options=config['options'])
        charm = Charm(charm_data)
        self.assertEqual(config, charm.config)
        # The default is a dict with a single key named 'options' set to
        # an empty dict.
        charm = Charm({})
        self.assertEqual({'options': {}}, charm.config)

    def test_config_raw(self):
        # The config_raw property is string of YAML.
        config_raw = 'option: 0'
        charm_data = factory.get_charm_json()
        charm_data['config_raw'] = config_raw
        charm = Charm(charm_data)
        self.assertIs(config_raw, charm.config_raw)
        # The default is an empty string.
        charm = Charm({})
        self.assertEqual('', charm.config_raw)

    def test_options(self):
        # The options property is a dict from the charm's config.
        options = {'key': {'default': 'value', 'type': 'string'}}
        charm_data = factory.get_charm_json(options=options)
        charm = Charm(charm_data)
        self.assertEqual(options, charm.options)
        # The default is a dict.
        charm = Charm({})
        self.assertEqual({}, charm.options)

    def test_options_when_config_is_none(self):
        # The options property is a dict even if the config is None.
        charm = Charm({})
        charm._representation['config'] = None
        self.assertEqual({}, charm.options)

    def test_hooks(self):
        # The hooks property is a list of life cycle and relation names.
        hooks = ['install', 'start', 'stop']
        charm_data = factory.get_charm_json()
        charm_data['hooks'] = hooks
        charm = Charm(charm_data)
        self.assertIs(hooks, charm.hooks)
        # The default is an empty list.
        charm = Charm({})
        self.assertEqual([], charm.hooks)

    def test_label(self):
        # The label property is string.
        label = 'user:series/name'
        charm_data = factory.get_charm_json()
        charm_data['label'] = label
        charm = Charm(charm_data)
        self.assertIs(label, charm.label)
        # The default is an empty string.
        charm = Charm({})
        self.assertEqual('', charm.label)

    def test_short_url(self):
        # The short_url property is string.
        short_url = 'user:series/name'
        charm_data = factory.get_charm_json()
        charm_data['short_url'] = short_url
        charm = Charm(charm_data)
        self.assertIs(short_url, charm.short_url)
        # The default is an empty string.
        charm = Charm({})
        self.assertEqual('', charm.short_url)

    def test_changes_since(self):
        changes = [factory.makeChange(created=date(2013, 6, num))
                   for num in range(1, 31)]
        charm = Charm(factory.get_charm_json(changes=changes))
        self.assertEqual([], charm.changes_since(datetime(2013, 7, 1)))
        self.assertEqual(30, len(charm.changes_since(datetime(2013, 5, 1))))
        self.assertEqual(5, len(charm.changes_since(datetime(2013, 6, 26))))

    def test_hash(self):
        charm_data = factory.get_charm_json()
        self.assertFalse('hash' in charm_data)
        charm = Charm(charm_data)
        self.assertIs(None, charm.hash)
        charm_data['hash'] = 'whatever'
        charm = Charm(charm_data)
        self.assertEqual('whatever', charm.hash)


class TestUser(MongoTestBase):
    """Unit tests for our User object wrapping of the mongo dict."""

    def test_model_from_sso(self):
        """Given an SSO dump we can get back a decent model object."""
        payload = factory.makeSSOResponse()
        user = UserMgr.from_sso(payload)

        self.assertEqual(payload['openid.claimed_id'], user.userid)
        self.assertEqual(payload['openid.sreg.nickname'], user.nickname)
        self.assertEqual(payload['openid.sreg.fullname'], user.fullname)
        self.assertEqual(payload['openid.sreg.email'], user.email)
        self.assertEqual(payload['openid.sreg.timezone'], user.timezone)
        self.assertEqual(payload['openid.lp.is_member'], ','.join(user.groups))

    def test_user_limited_sharing(self):
        """The user might limit information shared."""
        claimed_id = u'https://login.ubuntu.com/+id/MSbTCYM'

        user = UserMgr.from_sso({
            'openid.claimed_id': claimed_id,
            # When a user chooses not to share teams, an empty is_member is
            # returned.
            'openid.lp.is_member': '',
        })

        self.assertEqual(claimed_id, user.userid)
        self.assertEqual('Not Provided', user.nickname)
        self.assertEqual('Not Provided', user.fullname)
        self.assertEqual(None, user.email)
        self.assertEqual('GMT', user.timezone)
        self.assertEqual([], user.groups)

    def test_can_dict_user(self):
        """We need to be able to turn User into a dict for mongo."""
        payload = factory.makeSSOResponse()
        user = UserMgr.from_sso(payload)

        dict_user = dict(user)
        self.assertEqual(
            payload['openid.claimed_id'],
            dict_user['userid'])
        self.assertEqual(
            payload['openid.sreg.nickname'],
            dict_user['nickname'])

    def test_user_stored_in_mongo(self):
        """The user should store in mongo correctly on save."""
        payload = factory.makeSSOResponse()
        user = UserMgr.from_sso(payload)
        user.save(self.db)

        found = self.db.users.find_one({'userid': user.userid})
        self.assertEqual(user.userid, found['userid'])

    def test_user_from_mongo(self):
        """Pulling the user from mongo via UserMgr should get a User out."""
        payload = factory.makeSSOResponse()
        user = UserMgr.from_sso(payload)
        user.save(self.db)

        found = UserMgr.find_one(self.db, user.userid)
        self.assertEqual(user.userid, found.userid)

    def test_user_updated(self):
        """Saving will either insert or save as required."""
        payload = factory.makeSSOResponse()
        user = UserMgr.from_sso(payload)
        # Does the insert
        user.save(self.db)

        found = UserMgr.find_one(self.db, user.userid)
        self.assertEqual(user.userid, found.userid)
        self.assertEqual(user.fullname, found.fullname)

        found.fullname = 'Changed in test'
        found.save(self.db)

        found2 = UserMgr.find_one(self.db, found.userid)
        self.assertEqual(found.userid, found2.userid)
        self.assertEqual(found.fullname, found2.fullname)

    def test_update_from_sso(self):
        """A new set of SSO data updates the user found."""
        payload = factory.makeSSOResponse()
        user = UserMgr.from_sso(payload)
        # Does the insert
        user.save(self.db)

        new_groups = ['launchpad']
        payload['openid.lp.is_member'] = new_groups[0]
        user.update_from_sso(payload)
        self.assertEqual(new_groups, user.groups)

    def test_auth_principalfinder(self):
        class Request:
            user = UserMgr.from_sso(
                factory.makeSSOResponse(groups=['charmers', 'canonical']))
        expected_principals = [
            'user https://login.ubuntu.com/+id/MSbTCYM',
            'group charmers',
            'group canonical']
        principals = UserMgr.auth_principalfinder('', Request)
        self.assertEqual(expected_principals, principals)


class TestQADataSource(MongoTestBase):

    def test_load_question_set(self):
        data_source = QADataSource.from_db(self.db)
        self.assertEqual({}, data_source.load_question_set())
        category = factory.makeQAQuestionGroup()
        self.db.qa.insert(category)
        question_set = data_source.load_question_set()
        self.assertEqual(['reliable'], question_set.keys())
        self.assertEqual(category['questions'],
                         question_set['reliable']['questions'])

    def test_save_qa_data(self):
        charm_data = factory.makeCharm(self.db)[1]
        charm = Charm(charm_data)
        qa_data_source = QADataSource.from_db(self.db)
        qa_data_source.save_qa_data(charm, {'a': 'b'})
        charm_id = construct_charm_id(charm_data)
        qa_data = self.db.charm_qa_answers.find_one(charm_id)
        self.assertEqual({
            '_id': charm_id,
            'name': charm.name,
            'owner': charm.owner,
            'series': charm.series,
            'revision': charm.store_data['revision'],
            'answers': {'a': 'b'},
        }, qa_data)

    def test_get_qa_data(self):
        charm_data = factory.makeCharm(self.db)[1]
        charm_id = construct_charm_id(charm_data)
        charm = Charm(charm_data)
        self.db.charm_qa_answers.save({
            '_id': charm_id,
            'name': charm.name,
            'owner': charm.owner,
            'series': charm.series,
            'revision': charm.store_data['revision'],
            'answers': {'a': 'b'},
        })
        qa_data_source = QADataSource.from_db(self.db)
        self.assertEqual(
            {'a': 'b'},
            qa_data_source.get_qa_data(charm).charm_data,
        )

    def test_stores_multiple_versions(self):
        charm_data = factory.get_charm_json()
        qa_data_source = QADataSource.from_db(self.db)
        charm = Charm(charm_data)
        charm.store_data['revision'] = 1
        qa_data_source.save_qa_data(charm, {'new': 'data'})
        charm.store_data['revision'] = 0
        qa_data_source.save_qa_data(charm, {'old': 'data'})
        self.assertEqual(2, qa_data_source.answers.find().count())

    def test_get_qa_data_matches_unversioned_values(self):
        charm_data = factory.get_charm_json()
        qa_data_source = QADataSource.from_db(self.db)
        charm = Charm(charm_data)
        qa_data_source.save_qa_data(charm, {'qa': 'data'})
        wrong_name = Charm(factory.get_charm_json(series=charm.series,
                           owner=charm.owner))
        wrong_name_qa = qa_data_source.get_qa_data(wrong_name)
        self.assertIs(None, wrong_name_qa.charm_data)
        wrong_series = Charm(factory.get_charm_json(name=charm.name,
                             owner=charm.owner, series='wrong'))
        wrong_series_qa = qa_data_source.get_qa_data(wrong_series)
        self.assertIs(None, wrong_series_qa.charm_data)
        wrong_owner = Charm(factory.get_charm_json(name=charm.name,
                            series=charm.series, owner='wrong'))
        wrong_owner_qa = qa_data_source.get_qa_data(wrong_owner)
        self.assertIs(None, wrong_owner_qa.charm_data)
        right_unversioned = Charm(factory.get_charm_json(name=charm.name,
                                  series=charm.series, owner=charm.owner))
        right_unversioned_qa = qa_data_source.get_qa_data(right_unversioned)
        self.assertEqual({'qa': 'data'}, right_unversioned_qa.charm_data)

    def test_get_qa_data_retrieves_newest(self):
        # Retrieves the newest QA available.
        charm_data = factory.makeCharm(self.db)[1]
        charm = Charm(charm_data)
        qa_data_source = QADataSource.from_db(self.db)
        charm.store_data['revision'] = 0
        qa_data_source.save_qa_data(charm, {'old': 'data'})
        charm.store_data['revision'] = 1
        qa_data_source.save_qa_data(charm, {'new': 'data'})
        charm.store_data['revision'] = 2
        self.assertEqual({'new': 'data'},
                         qa_data_source.get_qa_data(charm).charm_data)

    def test_get_qa_data_omits_newer(self):
        # Omits QA that is newer than the charm.
        charm_data = factory.makeCharm(self.db)[1]
        charm = Charm(charm_data)
        qa_data_source = QADataSource.from_db(self.db)
        charm.store_data['revision'] = 0
        qa_data_source.save_qa_data(charm, {'old': 'data'})
        charm.store_data['revision'] = 1
        qa_data_source.save_qa_data(charm, {'new': 'data'})
        charm.store_data['revision'] = 0
        self.assertEqual({'old': 'data'},
                         qa_data_source.get_qa_data(charm).charm_data)


class TestQAData(MongoTestBase):
    """Unit tests for our User object wrapping of the mongo dict."""

    def setUp(self):
        super(TestQAData, self).setUp()
        category = factory.makeQAQuestionGroup()
        self.db.qa.insert(category)
        self.qa_data_source = QADataSource.from_db(self.db)

    def test_category_property(self):
        qadata = QAData(self.qa_data_source)
        self.assertEqual(['reliable'], qadata.categories)

    def test_none_overall_if_no_data(self):
        """When no charm qa data we get None to indicate N/A."""
        qadata = QAData(self.qa_data_source)
        self.assertIsNone(qadata.score)

    def test_overall_score(self):
        scores = factory.makeQAScore()
        qadata = QAData(self.qa_data_source, charm_data=scores)
        self.assertEqual(4, qadata.score)

    def test_none_category_score_if_missing_data(self):
        """When no scores are available for the cateogry we get None

        None is our indicator of a score that's N/A

        """
        qadata = QAData(self.qa_data_source)
        self.assertIsNone(qadata.category_score('reliable'))

    def test_category_score(self):
        scores = factory.makeQAScore()
        qadata = QAData(self.qa_data_source, charm_data=scores)

        self.assertEqual(4, qadata.category_score('reliable'))

    def test_category_max_score(self):
        """Determine the total available points for the category."""
        qadata = QAData(self.qa_data_source)
        self.assertEqual(9, qadata.category_max_score('reliable'))


class TestCharmFileSet(MongoTestBase):
    # Escape the base_fileid since all code will escape this. We're comparing
    # against this string so it needs to also be escaped.
    base_fileid = 'charmers/precise/charmworld%2Etest/trunk/'
    charm_data = {
        'series': 'precise',
        'owner': 'charmers',
        'name': 'charmworld.test',
        'bname': 'trunk',
    }
    charm_path = join(
        dirname(dirname(__file__)),
        'testing/data/sample_charm'
    )

    def setUp(self):
        super(TestCharmFileSet, self).setUp()
        self.logger = logging.getLogger(self.__class__.__name__)
        self.log_handler = self.get_handler(self.logger.name)

    def save_files(self, charm_path=None, extra_revisions=[]):
        with charm_branch(charm_path, extra_revisions) as (
                branch_dir, revision, path):
            charm_data = self.charm_data.copy()
            charm_data['branch_dir'] = branch_dir
            charm_data['store_data'] = {'digest': revision}
            return CharmFileSet.save_files(
                self.fs,
                charm_data,
                branch_dir,
                self.logger)

    def save_revision_files(self, charm_path=None, extra_revisions=[]):
        with charm_branch(charm_path, extra_revisions) as (
                branch_dir, revision, path):
            charm_data = self.charm_data.copy()
            charm_data['branch_dir'] = branch_dir
            charm_data['store_data'] = {'digest': revision}
            CharmFileSet.save_revision_files(
                self.fs, charm_data, branch_dir, self.logger)
            return charm_data

    def test_fileid_generation(self):
        """A charm must generate an unique fileid."""
        self.assertEqual(
            'charmers/precise/charmworld%2Etest/trunk/readme',
            CharmFileSet.gen_fileid(self.charm_data, 'readme')
        )

    saved_sample_charm_files = [
        'config.yaml',
        'metadata.yaml',
        'README.md',
        'icon.svg',
        'revision',
        'hooks/config-changed',
        'hooks/database-relation-broken',
        'hooks/database-relation-changed',
        'hooks/install',
        'hooks/linkstart',
        'hooks/restart',
        'hooks/start',
        'hooks/stop',
        'hooks/upgrade-charm',
        'hooks/website-relation-changed',
    ]

    def check_bzr_files(self, expected_files, saved_files):
        # There should be the following files stored into the gridfs
        # filestore.
        for name in expected_files:
            fileid = self.base_fileid + quote_key(name)
            self.assertTrue(
                self.fs.exists(fileid),
                'Should find file:' + fileid)
            existing_content = self.fs.get(fileid).read()
            self.assertEqual(expected_files[name], existing_content)
        saved_ids = set(f._id for f in saved_files)
        expected_ids = set(
            self.base_fileid + quote_key(name) for name in expected_files)
        self.assertEqual(expected_ids, saved_ids)

    def expected_files(self):
        result = {}
        sample_data_path = split(split(__file__)[0])[0]
        sample_data_path = join(sample_data_path, 'testing/data/sample_charm')
        for name in self.saved_sample_charm_files:
            result[name] = open(join(sample_data_path, name)).read()
        return result

    def test_load_bzr_files(self):
        """Processing a charm dir loads the file contents."""
        saved = self.save_files()
        self.check_bzr_files(self.expected_files(), saved)

    def test_load_bzr_files_from_old_revision(self):
        """Processing a charm dir loads the file contents."""
        def add_revision_two(cwd, tree):
            path = join(cwd, 'branch/trunk/hooks/new_hook')
            with open(path, 'w') as f:
                f.write('A new hook')
            tree.add(['hooks/new_hook'])
            tree.commit('revision 2')

        saved = self.save_files(extra_revisions=[add_revision_two])
        self.check_bzr_files(self.expected_files(), saved)

    def test_load_bzr_files_from_new_revision(self):
        """Processing a charm dir loads the file contents."""
        def add_revision_two(cwd, tree):
            path = join(cwd, 'branch/trunk/hooks/new_hook')
            with open(path, 'w') as f:
                f.write('A new hook')
            tree.add(['hooks/new_hook'])
            return tree.commit('revision 2')

        saved = self.save_files(extra_revisions=[add_revision_two])
        expected = self.expected_files()
        expected['hooks/new_hook'] = 'A new hook'
        self.check_bzr_files(expected, saved)

    def test_save_files_with_non_existent_bzr_revision(self):
        """Attempts to retrieve a non-existent branch revision result in
        an empty return value of save_files and a log entry."""
        with charm_branch() as (branch_dir, revision, path):
            charm_data = self.charm_data.copy()
            charm_data['branch_dir'] = branch_dir
            charm_data['store_data'] = {'digest': 'bad-revision-id'}
            result = CharmFileSet.save_files(
                self.fs,
                charm_data,
                branch_dir,
                self.logger)
        self.assertEqual([], result)
        messages = [rec.msg for rec in self.log_handler.buffer]
        self.assertEqual(['Revision bad-revision-id not found'], messages)

    def test_case_kept_for_filenames(self):
        """README.md is lower cased in the key, but upper in filename."""
        self.save_files()

        # Verify that the filename of the README is kept.
        found = self.fs.get_last_version('README.md')
        # fileid is escaped because it's used as a mongo key.
        expected = self.base_fileid + quote_key('README.md')
        self.assertEqual(expected, found._id)

    def test_get_by_id(self):
        """Load a CharmFile instance from the datastore."""
        self.save_files()

        readme_id = self.base_fileid + quote_key('README.md')

        cf = CharmFileSet.get_by_id(self.fs, readme_id)
        self.assertEqual('README.md', cf.filename)
        self.assertEqual(readme_id, cf.fileid)

        # We need to be able to dict the metadata to store in the charm record
        # itself.
        gridout = cf._gridout
        metadata = dict(cf)
        self.assertEqual('README.md', metadata['filename'])
        self.assertEqual(readme_id, metadata['fileid'])
        self.assertEqual(u'charmers/precise/charmworld.test/trunk',
                         metadata['charmid'])
        self.assertEqual(gridout.length, metadata['length'])
        self.assertEqual(gridout.md5, metadata['md5'])
        self.assertEqual('', metadata['subdir'])

    def test_get_by_id_hooks_file(self):
        """The hooks file is a subdir down and we want to track that."""
        self.save_files()

        install_id = self.base_fileid + 'hooks/install'

        cf = CharmFileSet.get_by_id(self.fs, install_id)
        self.assertEqual('install', cf.filename)
        self.assertEqual('hooks', cf.subdir)
        self.assertEqual(install_id, cf.fileid)

        # We need to be able to dict the metadata to store in the charm record
        # itself.
        metadata = dict(cf)
        self.assertEqual('hooks', metadata['subdir'])

    def test_save_files_handles_missing_dir(self):
        path = self.use_context(temp_dir())
        files = self.save_files(path)
        self.assertEqual([], files)

    def test_save_file_ignores_symlink_without_target(self):
        # Symlinks that point to nowhere are not saved.
        path = self.use_context(temp_dir())
        link_path = join(path, 'readme')
        target_name = 'does-not-exist'
        os.symlink(target_name, link_path)
        files = self.save_files(path)
        expected = [
            'File does not exist: readme (resolving to does-not-exist).',
        ]
        messages = [rec.msg for rec in self.log_handler.buffer]
        self.assertEqual(expected, messages)
        self.assertEqual([], files)

    def test_save_file_ignores_symlink_with_outside_target(self):
        # Symlinks that point to a location outside the charm's root
        # directory are also ignored.
        path = self.use_context(temp_dir())
        link_path = join(path, 'readme')
        target_name = '../no-trespassing'
        os.symlink(target_name, link_path)
        files = self.save_files(path)
        expected = [
            "Invalid file path: Path readme points to ../no-trespassing, "
            "which is not inside the charm's root directory"
        ]
        messages = [rec.msg for rec in self.log_handler.buffer]
        self.assertEqual(expected, messages)
        self.assertEqual([], files)

    def test_save_file_ignores_symlink_with_absolute_target_path(self):
        # Symlinks that point to an absolute location are ignored.
        path = self.use_context(temp_dir())
        link_path = join(path, 'readme')
        target_name = '/etc/passwd'
        os.symlink(target_name, link_path)
        files = self.save_files(path)
        expected = [
            "Invalid file path: symlink readme points to the absolute path "
            "/etc/passwd."
        ]
        messages = [rec.msg for rec in self.log_handler.buffer]
        self.assertEqual(expected, messages)
        self.assertEqual([], files)

    def test_save_file_ignores_invalid_icons(self):
        path = self.use_context(temp_dir())
        # Not an SVG.
        with open(join(path, 'icon.svg'), 'w') as f:
            f.write('foo')
        files = self.save_files(path)
        self.assertEqual([], files)
        # No width or height provided.
        with open(join(path, 'icon.svg'), 'w') as f:
            f.write('<?xml version="1.0"?><svg></svg>')
        files = self.save_files(path)
        self.assertEqual([], files)

    def test_adds_viewbox_to_icons(self):
        path = self.use_context(temp_dir())
        with open(join(path, 'icon.svg'), 'w') as f:
            f.write('<?xml version="1.0"?><svg width="10" height="10"></svg>')
        self.save_files(path)
        icon_id = self.base_fileid + quote_key('icon.svg')
        cf = CharmFileSet.get_by_id(self.fs, icon_id)
        # Note that ElementTree generates the processing instruction with
        # single quotes, but elements with double quotes.  To keep from having
        # to escape everything, the strings are quoted accordingly below.
        self.assertEqual(
            "<?xml version='1.0' encoding='UTF-8'?>\n"
            + '<svg height="10" viewBox="0 0 10 10" width="10" />',
            cf.read())

    def test_provides_content_type(self):
        path = self.use_context(temp_dir())
        with open(join(path, 'icon.svg'), 'w') as f:
            f.write('<?xml version="1.0"?><svg width="10" height="10"></svg>')
        with open(join(path, 'readme.txt'), 'w') as f:
            f.write('foo')
        self.save_files(path)
        icon_id = self.base_fileid + quote_key('icon.svg')
        cf = CharmFileSet.get_by_id(self.fs, icon_id)
        self.assertEqual('image/svg+xml', cf.contentType)
        readme_id = self.base_fileid + quote_key('readme.txt')
        cf = CharmFileSet.get_by_id(self.fs, readme_id)
        self.assertEqual('text/plain', cf.contentType)

    def test_directories_match(self):
        path = self.use_context(temp_dir())
        os.mkdir(join(path, 'hooks'))
        os.mkdir(join(path, 'hooks_lib'))
        with open(join(path, 'hooks/a_hook'), 'w') as f:
            f.write('hook')
        with open(join(path, 'hooks_lib/hook_lib'), 'w') as f:
            f.write('not stored in our DB')
        saved = self.save_files(path)
        saved = [charm_file._filename for charm_file in saved]
        self.assertEqual(['a_hook'], saved)

    def check_revision_files(self, expected_files, charm_revision):
        expected_names = set(expected_files)
        found_names = set(charm_revision['files'])
        self.assertEqual(expected_names, found_names)
        for name in expected_names:
            expected_fs_id = hashlib.sha256(expected_files[name]).hexdigest()
            self.assertEqual(expected_fs_id, charm_revision['files'][name])

    def test_save_revision_files(self):
        charm_revision = self.save_revision_files()
        self.check_revision_files(self.expected_files(), charm_revision)

    def add_revision(self, cwd, tree, return_new_revision):
        path = join(cwd, 'branch/trunk/hooks/new_hook')
        with open(path, 'w') as f:
            f.write('A new hook')
            tree.add(['hooks/new_hook'])
        path = join(cwd, 'branch/trunk/README.md')
        with open(path, 'a') as f:
            f.write('appended text')
        if return_new_revision:
            return tree.commit('revision 2')
        else:
            return None

    def test_save_revision_files_from_old_revision(self):
        def add_revision_two(cwd, tree):
            return self.add_revision(cwd, tree, False)
        revision = self.save_revision_files(extra_revisions=[add_revision_two])
        self.check_revision_files(self.expected_files(), revision)

    def test_save_revision_files_from_new_revision(self):
        def add_revision_two(cwd, tree):
            return self.add_revision(cwd, tree, True)
        revision = self.save_revision_files(extra_revisions=[add_revision_two])
        expected = self.expected_files()
        expected['hooks/new_hook'] = 'A new hook'
        expected['README.md'] = expected['README.md'] + 'appended text'
        self.check_revision_files(expected, revision)

    def test_save_revision_files_with_non_existent_bzr_revision(self):
        with charm_branch() as (branch_dir, revision, path):
            charm_data = self.charm_data.copy()
            charm_data['branch_dir'] = branch_dir
            charm_data['store_data'] = {'digest': 'bad-revision-id'}
            CharmFileSet.save_revision_files(
                self.fs,
                charm_data,
                branch_dir,
                self.logger)
        self.assertEqual({}, charm_data['files'])
        messages = [rec.msg for rec in self.log_handler.buffer]
        self.assertEqual(['Revision bad-revision-id not found'], messages)


class TestCharmFile(MongoTestBase):

    def test_filename_parsing(self):
        """The filename is the end of the fileid."""
        cf = CharmFile(
            self.fs, '/test/file/path.txt', '/charmers/precise/redis')
        self.assertEqual('path.txt', cf.filename)

    def test_filename_with_caps(self):
        """Whe can specifiy the filename to keep it original."""

        cf = CharmFile(
            self.fs, '/test/file/path.txt', '/charmers/precise/redis',
            filename='PATH.txt')
        self.assertEqual('PATH.txt', cf.filename)

    def test_storing_file_contents(self):
        contents = 'Contents of the file.'
        fileid = '/test/file/path.txt'
        cf = CharmFile(self.fs, fileid, '/charmers/precise/redis')
        cf.save(contents)

        # Make sure we check for the escaped fileid to see it exists.
        self.assertTrue(self.fs.exists(cf.fileid))

        found = self.fs.get(cf.fileid)
        self.assertEqual(contents, found.read())

    def test_save_with_orphan_chunks(self):
        file_id = 'file-id'
        cf = CharmFile(self.fs, file_id, 'baz')
        cf.save('foo')
        # Create orphan chunk
        self.db.fs.files.remove({'_id': file_id})
        self.assertFalse(self.fs.exists(file_id))
        cf.save('bar')
        self.assertEqual('bar', self.fs.get(file_id).read())

    def test_md5(self):
        cf = CharmFile(self.fs, 'file-id', 'baz')
        self.assertIs(None, cf.md5)
        cf.save('asdf')
        self.assertEqual(cf.md5, md5.new('asdf').hexdigest())

    def test_no_contentType(self):
        result = self.fs.put(
            'foo',
            filename='bar',
        )
        instance = self.fs.get(result)
        sentry = object()
        self.assertIs(sentry, getattr(instance, 'contentType', sentry))
        cf = CharmFile(self.fs, None, None, instance=instance)
        self.assertIs(None, cf.contentType)


class TestAcquireSessionSecret(MongoTestBase):

    def test_acquire_session_secret(self):
        """Multiple calls to acquire_session_secret return the same value."""
        secret = acquire_session_secret(self.db)
        self.assertIs(str, type(secret))
        self.assertEqual(secret, acquire_session_secret(self.db))

    def test_db_backed(self):
        """The secret is stored in the database.

        An existing session-secret is used if present.  Deleting it causes a
        new secret to be generated.
        """
        self.db.app_config.insert({'_id': 'session-secret',
                                   'secret': 'arbitrary-string'})
        first_secret = acquire_session_secret(self.db)
        self.assertEqual('arbitrary-string', first_secret)
        self.db.app_config.remove('session-secret')
        second_secret = acquire_session_secret(self.db)
        self.assertNotEqual(first_secret, second_secret)


class TestFindCharms(MongoTestBase):

    def test_find_charms_default(self):
        # By default, find_charms() does not return charms with errors.
        ignore, valid_charm = factory.makeCharm(self.db, name='valid_charm')
        ignore, charm_with_error = factory.makeCharm(
            self.db, name='charm_with_error', charm_error=True)
        result = find_charms(self.db)
        self.assertEqual(1, len(result))
        self.assertEqual(valid_charm, result[0])

    def test_find_charms_invalid_included(self):
        # If the parameter valid_charms_only is set to False, find_charms()
        # returns invalid charms too.
        ignore, valid_charm = factory.makeCharm(self.db, name='valid_charm')
        ignore, charm_with_error = factory.makeCharm(
            self.db, name='charm_with_error', charm_error=True)
        result = find_charms(self.db, valid_charms_only=False)
        self.assertEqual(2, len(result))

    def test_find_charms_with_more_parameters(self):
        # Any parameter of Collection.find() can be specified.
        ignore, charm_1 = factory.makeCharm(self.db, name='valid_charm_1')
        ignore, charm_2 = factory.makeCharm(self.db, name='valid_charm_2')
        result = find_charms(self.db, {'name': 'does-not-exist'})
        self.assertEqual(0, len(result))
        result = find_charms(self.db, spec={'name': 'valid_charm_1'})
        self.assertEqual(1, len(result))
        self.assertEqual(charm_1, result[0])
        result = find_charms(
            self.db, spec={'name': 'valid_charm_1'}, fields=('bname', 'name'))
        expected = {
            '_id': charm_1['_id'],
            'name': charm_1['name'],
            'bname': charm_1['bname']
        }
        self.assertEqual(expected, result[0])

    def test_dont_hide_with_processing_error(self):
        # Charms with errors that occured during the ingest script
        # are returned by find_charms()
        charm_id, charm_data = factory.makeCharm(self.db)
        charm_data['error'] = {}
        self.db.charms.save(charm_data)
        self.assertIn(charm_data,
                      find_charms(self.db, valid_charms_only=False))

    def test_only_promulgated(self):
        ignore, promulgated_charm = factory.makeCharm(
            self.db, promulgated=True)
        factory.makeCharm(self.db)
        result = find_charms(self.db, only_promulgated=True)
        self.assertEqual(1, len(result))
        self.assertEqual(promulgated_charm, result[0])

    def test_all_revisons_not_default(self):
        factory.makeCharm(self.db, owner='a', series='b', name='c',
                          store_revision=3)
        factory.makeCharm(self.db, owner='a', series='b', name='c',
                          store_revision=1)
        self.assertEqual([3], [charm['store_data']['revision']
                         for charm in find_charms(self.db)])

    def test_all_revisons_true(self):
        factory.makeCharm(self.db, owner='a', series='b', name='c',
                          store_revision=3)
        factory.makeCharm(self.db, owner='a', series='b', name='c',
                          store_revision=1)
        self.assertItemsEqual(
            [3, 1], [charm['store_data']['revision'] for charm in
                     find_charms(self.db, all_revisions=True)])

    def test_sort(self):
        charm_1 = factory.makeCharm(self.db, owner='a', series='b')[0]
        charm_2 = factory.makeCharm(self.db, owner='b', series='a')[0]
        charms = find_charms(self.db, sort=[('owner', pymongo.ASCENDING)])
        self.assertEqual([charm_1, charm_2], [c['_id'] for c in charms])
        charms = find_charms(self.db, sort=[('owner', pymongo.DESCENDING)])
        self.assertEqual([charm_2, charm_1], [c['_id'] for c in charms])

    def test__find_charms_gets_unique_fields(self):
        # __find_charms takes appropriate precautions to ensure that the same
        # results are returned, even when there are charms that might be
        # expected to sort between them.
        factory.makeCharm(self.db, owner='a', series='b', name='c',
                          downloads=3, store_revision=2)[1]
        # different owner, 2 downloads
        factory.makeCharm(self.db, series='b', name='c', downloads=2)
        # different name, 2 downloads
        factory.makeCharm(self.db, owner='a', series='b', downloads=2)
        # different series, 2 downloads
        factory.makeCharm(self.db, owner='a', name='c', downloads=2)
        charm_1 = factory.makeCharm(self.db, owner='a', series='b', name='c',
                                    downloads=1, store_revision=1)[1]
        results = list(_find_charms(self.db.charms,
                                    sort=[('downloads', pymongo.DESCENDING)]))
        self.assertNotIn(charm_1, results)


class TestCharmSource(MongoTestBase):

    def test_sync_index(self):
        self.use_index_client()
        one = factory.makeCharm(self.db)[0]
        two = factory.makeCharm(self.db)[0]
        three = factory.makeCharm(self.db)[0]
        charm_source = CharmSource.from_request(self)
        ids = list(charm_source.sync_index())
        self.assertItemsEqual([one, two, three], ids)
        for charm_id in ids:
            foo, bar = charm_source._get_all(charm_id)
            self.assertEqual(foo, bar)

    def test_sync_index_script(self):
        factory.makeCharm(self.db, owner='b')[0]
        # Create a charm which is index-incompatible with the previous one.
        factory.makeCharm(self.db, owner='a',
                          description={'text': 'whatever'})[0]
        handler = self.get_handler('sync-index')
        with patch('charmworld.models.configure_logging'):
            with patch('charmworld.models.get_ini', lambda: INI):
                with force_index_client(self.use_index_client()):
                    sync_index()
            messages = [r.getMessage() for r in handler.buffer]
            self.assertIn('Unable to index charm.', messages)
        self.assertEqual(1, len(self.index_client.api_search()))

    def test_save_deletes_on_error(self):
        self.use_index_client()
        source = CharmSource.from_request(self)
        source.save({
            '_id': 'a',
            'owner': 'foo',
            'series': 'bar',
            'name': 'baz',
            'description': 'whatever',
        })
        with self.assertRaises(ElasticHttpError):
            source.save({
                '_id': 'a',
                'owner': 'foo',
                'series': 'bar',
                'name': 'baz',
                # A dictionary as the value conflicts with the mapping.
                'description': {'text': 'whatever'},
            })
        self.assertIs(None, self.index_client.get('a'))

    def test_sync_index_uses_latest(self):
        self.use_index_client()
        factory.makeCharm(self.db, owner='a', series='b', name='c',
                          store_revision=1)[0]
        two = factory.makeCharm(self.db, owner='a', series='b', name='c',
                                store_revision=2)[0]
        charm_source = CharmSource.from_request(self)
        ids = list(charm_source.sync_index())
        self.assertItemsEqual([two], ids)
        for charm_id in ids:
            foo, bar = charm_source._get_all(charm_id)
            self.assertEqual(foo, bar)

    def test_find_proof_error_charms(self):
        self.use_index_client()
        proof_with_errors = {'e': ['error 1']}
        proof_with_warnings = {'w': ['warning 1']}
        # The view retrieves promulgated charms with proof errors.
        ignore, charm_1 = factory.makeCharm(
            self.db, owner='joe', promulgated=True, proof=proof_with_errors)
        # The view ignores promulgated charms with proof warning.
        factory.makeCharm(self.db, promulgated=True, proof=proof_with_warnings)
        # The view ignored unpromulgated charms with proof errors.
        factory.makeCharm(self.db, promulgated=False, proof=proof_with_errors)
        charm_source = CharmSource.from_request(self)
        errors = list(charm_source.find_proof_error_charms(False))
        self.assertEqual(1, len(errors))
        self.assertEqual(charm_1['name'], errors[0].name)

    def test_obsolete_charms_ignored_by_find_proof_error_charms(self):
        self.use_index_client()
        proof_with_errors = {'e': ['error 1']}
        # The view retrieves promulgated charms with proof errors.
        ignore, charm_1 = factory.makeCharm(
            self.db, owner='joe', promulgated=True, proof=proof_with_errors,
            series='oneiric')
        charm_source = CharmSource.from_request(self)
        errors = list(charm_source.find_proof_error_charms(False))
        self.assertEqual([], errors)

    def test_store_errors_in_queue(self):
        self.use_index_client()
        ignore, charm = factory.makeCharm(self.db, charm_error=True)
        factory.makeCharm(self.db, charm_error=False)
        charm_source = CharmSource.from_request(self)
        errors = list(charm_source.find_mia_charms())
        self.assertEqual(charm['name'], errors[0].name)
        self.assertEqual(charm['branch_spec'], errors[0].branch_spec)

    def test_obsolete_charms_ignored_by_find_mia_charms(self):
        self.use_index_client()
        ignore, charm = factory.makeCharm(
            self.db, charm_error=True, series='oneiric')
        charm_source = CharmSource.from_request(self)
        errors = list(charm_source.find_mia_charms())
        self.assertEqual([], errors)


class TestOptionsStorage(TestCase):

    yaml = {
        'foo': {
            'default': 'bar',
            'type': 'string',
            'description': 'Hello.'
        },
        'baz': {
            'default': 42,
            'type': 'int',
            'description': 'Hello.'
        },
    }
    storage = [{
        'name': 'baz',
        'default': 42,
        'type': 'int',
        'description': 'Hello.'
    }, {
        'name': 'foo',
        'default': 'bar',
        'type': 'string',
        'description': 'Hello.'
    }]

    def test_options_to_storage(self):
        self.assertItemsEqual(self.storage, options_to_storage(self.yaml))

    def test_options_to_storage_none(self):
        self.assertItemsEqual([], options_to_storage(None))

    def test_storage_to_options(self):
        self.assertEqual(self.yaml, storage_to_options(self.storage))

    def test_storage_to_options_duplicate_key(self):
        with self.assertRaises(ValueError) as exc_context:
            storage_to_options([{'name': 'same'}, {'name': 'same'}])
        self.assertEqual('Option name "same" occurs multiple times.',
                         str(exc_context.exception))

    def test_roundtrip(self):
        self.assertEqual(
            self.yaml, storage_to_options(options_to_storage(self.yaml)))
        self.assertItemsEqual(
            self.storage,
            options_to_storage(storage_to_options(self.storage)))


class TestStoreBundles(MongoTestBase):

    def test_store_bundles(self):
        # The bundles can be stored in the database.

        deployer_config = dedent("""\
            wordpress-stage:
                series: precise
                services:
                    blog:
                        charm: cs:precise/wordpress
                        constraints: mem=2
                        options:
                            tuning: optimized
                            engine: apache
        """)
        parsed = yaml.safe_load(deployer_config)
        owner = 'bac'
        bundle_name = 'wordpress-stage'
        basket_name = 'wordpress-basket'
        basket_rev = 5
        basket_id = "%s/%d" % (basket_name, basket_rev)
        _id = Bundle.construct_id(owner, basket_name, bundle_name, basket_rev)
        store_bundles(
            self.db.bundles, parsed, 'bac', basket_id, None, None, None, None)
        self.assertEqual(
            {
                '_id': _id,
                'basket_name': basket_name,
                'basket_revision': basket_rev,
                'name': bundle_name,
                'owner': owner,
                'data': parsed['wordpress-stage'],
                'first_change': None,
                'last_change': None,
                'changes': None,
                'branch_spec': None,
            },
            self.db.bundles.find_one(_id))

    def test_store_bundles_calls_get_flattened_deployment(self):
        # For every key in the top-level deployment data,
        # get_flattened_deployment is called once.
        deployer_config = dedent("""\
            wordpress-stage:
                series: precise
                services:
                    blog:
                        charm: cs:precise/wordpress
                        constraints: mem=2
                        options:
                            tuning: optimized
                            engine: apache
            wordpress-prod:
                series: precise
                services:
                    blog:
                        charm: cs:precise/wordpress
                        constraints: mem=5
                        options:
                            tuning: optimized
                            engine: apache
        """)
        parsed = yaml.safe_load(deployer_config)
        keys = []

        def get_flattened_deployment(data, key):
            keys.append(key)

        with patch('charmworld.models.get_flattened_deployment',
                   get_flattened_deployment):
            store_bundles(
                self.db.bundles, parsed, 'bac', 'wordpress-basket/5', None,
                None, None, None)
        self.assertItemsEqual(['wordpress-stage', 'wordpress-prod'], keys)

    def test_storing_a_bundle_includes_indexing_it(self):
        class FauxElasticSearchClient:
            index_bundles_called = False

            @classmethod
            def from_settings(cls, *args):
                return cls

            @classmethod
            def index_bundles(cls, *args):
                cls.index_bundles_called = True

        with patch(
                'charmworld.models.ElasticSearchClient',
                FauxElasticSearchClient):
            store_bundles(
                self.db.bundles, {}, 'owner', 'wordpress-basket/5', None,
                None, None, None)

        self.assertTrue(FauxElasticSearchClient.index_bundles_called)


class BundleTestCase(MongoTestBase):
    """Unit tests for the Bundle model that wraps bundle representations."""

    bundle_data = {
        'owner': 'sinzui',
        'basket_name': 'mysql',
        'basket_revision': 2,
        'name': 'tiny',
        'title': 'Tiny bundle',
        'description': 'My Tiny Bundle',
        'data': {
            'services': dict(apache=1),
            'relations': dict(a=1),
            'series': 'series',
        },
        'promulgated': True,
        'branch_deleted': True,
    }

    def test_construct_id_no_revision(self):
        bundle_id = Bundle.construct_id(
            'a', 'b', 'c')
        self.assertEqual('~a/b/c', bundle_id)

    def test_init_with_empty_representation(self):
        bundle_data = {}
        bundle = Bundle(bundle_data)
        self.assertIs(bundle_data, bundle._raw_representation)
        self.assertEqual(bundle._MONGO_FIELDS, bundle._representation)

    def test_init_with_minimal_data(self):
        bundle_data = {
            'owner': 'sinzui',
            'basket_name': 'mysql',
            'basket_revision': 1,
            'changes': None,
            'first_change': None,
            'last_change': None,
            'name': 'tiny',
            'title': 'Tiny bundle',
            'description': 'My Tiny Bundle',
            'data': {
                'services': dict(apache=1),
                'relations': dict(a=1),
                'series': 'series',
            },
            'promulgated': True,
            'branch_deleted': True,
            'is_featured': False,
            'branch_spec': '~sinzui/fake/path',
        }
        bundle = Bundle(bundle_data)
        self.assertIs(bundle_data, bundle._raw_representation)
        self.assertEqual(bundle_data, bundle._representation)

    def test_reject_arbitrary_attributes(self):
        bundle = Bundle({})
        with self.assertRaises(AttributeError):
            bundle.not_a_real_attribute

    def test_is_equal(self):
        # If two bundles have the same permanent_url, they are considered
        # equal, even if their data differ.
        common = {
            'name': 'tiny',
            'owner': 'bac',
            'basket_name': 'apache',
            'basket_revision': 1,
        }
        bundle1 = Bundle(dict(common, title='Tiny Apache'))
        bundle2 = Bundle(dict(common, title='Itty Bitty Apache'))
        self.assertEqual(bundle1, bundle2)

    def test_not_equal(self):
        common = {
            'name': 'tiny',
            'basket_name': 'apache',
            'basket_revision': 4,
        }
        bundle1 = Bundle(dict(common, owner='bac'))
        bundle2 = Bundle(dict(common, owner='abc'))
        self.assertNotEqual(bundle1, bundle2)

    def test_id(self):
        bundle = Bundle(
            {'name': 'tiny',
             'owner': 'bac',
             'basket_name': 'apache',
             'basket_revision': 8,
             })
        self.assertEqual(
            "~bac/apache/8/tiny", bundle.id)

    def test_permanent_url(self):
        bundle = Bundle(
            {'name': 'tiny',
             'owner': 'bac',
             'basket_name': 'apache',
             'basket_revision': 6,
             })
        self.assertEqual(
            "bundle:~bac/apache/6/tiny", bundle.permanent_url)

    def test_basket_name(self):
        bundle = Bundle(
            {'name': 'tiny',
             'owner': 'bac',
             'basket_name': 'apache',
             'basket_revision': 6,
             })
        self.assertEqual(
            "apache", bundle.basket_name)

    def test_basket_name_with_none(self):
        # If a bundle is is not stored because of an ingest error,
        # basket_name will return None.
        bundle = Bundle({})
        self.assertIsNone(bundle.basket_name)

    def test_basket_revision(self):
        bundle = Bundle(
            {'name': 'tiny',
             'owner': 'bac',
             'basket_name': 'apache',
             'basket_revision': 6,
             })
        self.assertEqual(6, bundle.basket_revision)

    def test_basket_revision_no_rev(self):
        bundle = Bundle(
            {'name': 'tiny',
             'owner': 'bac',
             'basket_name': 'apache',
             })
        self.assertIsNone(bundle.basket_revision)

    def test_basket_revision_with_none(self):
        # If a bundle is is not stored because of an ingest error,
        # basket_revision will return None.
        bundle = Bundle({})
        self.assertIsNone(bundle.basket_revision)

    def test_short_url(self):
        bundle = Bundle(
            {'name': 'tiny',
             'owner': 'bac',
             'basket_name': 'apache',
             'basket_revision': 7,
             })
        self.assertEqual(
            "~bac/bundle/apache/tiny", bundle.short_url)

    def test_short_url_without_enough_parts(self):
        # When there is not enough information, the path is an empty string.
        bundle = Bundle({})
        self.assertEqual('', bundle.short_url)

    def test_loading_a_bundle_from_the_database(self):
        # A bundle can be loaded from the database with a query.
        self.db.bundles.save(self.bundle_data)
        bundle = Bundle.from_query({'owner': 'sinzui'}, self.db)
        self.assertIsNotNone(bundle)
        self.assertEqual(bundle.basket_name, 'mysql')
        self.assertEqual(bundle.name, 'tiny')

    def test_loading_a_bundle_from_a_given_collection(self):
        # A bundle can be loaded from a specific collection.
        collection = self.db.create_collection(factory.random_string(10))
        collection.save(self.bundle_data)
        bundle = Bundle.from_query({'owner': 'sinzui'}, collection=collection)
        self.assertIsNotNone(bundle)
        self.assertEqual(bundle.basket_name, 'mysql')
        self.assertEqual(bundle.name, 'tiny')

    def test_loading_a_nonexistent_bundle_returns_none(self):
        # If a query does not match any bundles, None is returned.
        bundle = Bundle.from_query({'owner': 'nunyas'}, self.db)
        self.assertIsNone(bundle)

    def test_multiple_matches_return_highest_revision(self):
        # A bundle can be loaded from the database with a query.
        mod_data = dict(self.bundle_data)
        mod_data['_id'] = 'a'
        self.db.bundles.save(mod_data)
        mod_data['_id'] = 'b'
        mod_data['basket_revision'] = 99
        self.db.bundles.save(mod_data)
        mod_data['_id'] = 'c'
        mod_data['basket_revision'] = 77
        self.db.bundles.save(mod_data)
        bundle = Bundle.from_query({'owner': 'sinzui'}, self.db)
        # If a query matches multiple entries, the one with the latest
        # revision is returned.
        self.assertIsNotNone(bundle)
        self.assertEqual(bundle.basket_name, 'mysql')
        self.assertEqual(bundle.basket_revision, 99)
        self.assertEqual(bundle.name, 'tiny')

    def test_bundles_are_iterable(self):
        # When a bundle is iterated over it generates a series of key, value
        # pairs.  This is convenient for generating a dictionary from a
        # bundle (e.g., dict(bundle)).
        bundle = Bundle({'name': 'bundle-name'})
        self.assertIn(('name', 'bundle-name'), list(bundle))
        # The important properties are also represented in the output.
        bundle_data = dict(bundle)
        self.assertIn('id', bundle_data)
        self.assertIn('permanent_url', bundle_data)

    def test_branch_short_url(self):
        bundle = Bundle(
            {'name': 'tiny',
             'owner': 'bac',
             'basket_name': 'apache',
             'basket_revision': 7,
             })
        self.assertEqual(
            "lp:~bac/charms/bundles/apache/bundle", bundle.branch_short_url)

    def test_branch_url(self):
        bundle = Bundle(
            {'name': 'tiny',
             'owner': 'bac',
             'basket_name': 'apache',
             'basket_revision': 7,
             })
        self.assertEqual(
            "https://code.launchpad.net/~bac/charms/bundles/apache/bundle",
            bundle.branch_url)

    def test_basket_id(self):
        bundle = Bundle(dict(
            owner='bac',
            basket_name='sample_bundle',
            basket_revision=1,
            name='wilma'))
        self.assertEqual(
            '~bac/sample_bundle/1',
            bundle.basket_id)

    def test_uses_provided_series(self):
        """If a series is provided in the data, it is returned."""
        bundle = Bundle(
            {'name': 'tiny',
             'owner': 'bac',
             'basket_name': 'apache',
             'basket_revision': 7,
             'data': {'series': 'edgy'},
             })
        self.assertEqual(
            "edgy", bundle.series)

    def test_default_series(self):
        """If no series is provided in the data, None is returned."""
        bundle = Bundle(
            {'name': 'tiny',
             'owner': 'bac',
             'basket_name': 'apache',
             'basket_revision': 7,
             'data': {},
             })
        self.assertIsNone(bundle.series)


class TestBundledCharmDescription(TestCase):

    def test_charm_resolving_from_bundle_branch(self):
        service_data = {
            'charm': 'mediawiki',
            'branch': 'lp:charms/precise/mediawiki'
        }

        description = BundledCharmDescription('wiki', service_data)

        self.assertEqual(
            '~charmers/charms/precise/mediawiki/trunk',
            description.branch)
        self.assertIsNone(description.revision)
        self.assertTrue(description.assume_latest_revision)

        # And supports a version.
        service_data['branch'] = 'lp:charms/precise/mediawiki@77'
        description = BundledCharmDescription('wiki', service_data)

        self.assertEqual(
            '~charmers/charms/precise/mediawiki/trunk',
            description.branch)
        self.assertEqual(77, description.revision)
        self.assertFalse(description.assume_latest_revision)

    def test_charm_resolving_from_bundle_charm_csurl(self):
        service_data = {
            'charm': 'cs:precise/mediawiki-77'
        }

        description = BundledCharmDescription('wiki', service_data)
        self.assertEqual(
            'cs:precise/mediawiki-77',
            description.store_url)
        self.assertTrue(description.assume_store_url)
        self.assertEqual(77, description.revision)

    def test_charm_resolving_from_charm_store_url(self):
        service_data = {
            'charm_url': 'cs:precise/mediawiki-77'
        }

        description = BundledCharmDescription('wiki', service_data)
        self.assertEqual(
            'cs:precise/mediawiki-77',
            description.store_url)
        self.assertFalse(description.assume_store_url)
        self.assertEqual(77, description.revision)

    def test_charm_resolving_from_bundle_charm_and_series(self):
        service_data = {
            'charm': 'mediawiki'
        }
        series = 'precise'
        description = BundledCharmDescription(
            'wiki', service_data, series=series)
        self.assertEqual(
            'precise',
            description.series)
        self.assertEqual(
            'mediawiki',
            description.charm)
        self.assertTrue(description.assume_latest_revision)
        self.assertTrue(description.assume_promulgated)

    def test_charm_resolving_from_bundle_empty_missing_info(self):
        service_data = {
            'charm': 'mediawiki'
        }
        description = BundledCharmDescription('wiki', service_data)
        self.assertEqual('mediawiki', description.charm)
        self.assertIsNone(description.series)
        self.assertIsNone(description.branch)
        self.assertIsNone(description.store_url)
        self.assertIsNone(description.revision)


class TestMakeBundleDoc(TestCase):

    def test_bundle_doc(self):
        doc = make_bundle_doc(
            {'a': 'b'}, 'foo', 'bar/9', 'baz', None, None, None,
            'fake/branch/spec')
        self.assertEqual({
            'owner': 'foo',
            'basket_name': 'bar',
            'basket_revision': 9,
            'name': 'baz',
            'data': {'a': 'b'},
            '_id': '~foo/bar/9/baz',
            'first_change': None,
            'last_change': None,
            'changes': None,
            'branch_spec': 'fake/branch/spec',
        }, doc)


class TestSlurpFiles(MongoTestBase):

    def test_slurp_files_saves_all_files_by_default(self):
        with charm_branch() as (branch_dir, revision, path):
            branch = Branch.open(branch_dir)
            tree = branch.repository.revision_tree(revision)
            with read_locked(tree):
                result = slurp_files(self.fs, tree)
        expected_file_paths = set()
        for dirpath, dirnames, filenames in os.walk(path):
            for filename in filenames:
                filepath = os.path.join(dirpath, filename)
                if stat.S_ISLNK(os.lstat(filepath).st_mode):
                    continue
                relpath = filepath[len(path) + 1:]
                expected_file_paths.add(relpath)
                content = open(filepath).read()
                expected_hash = hashlib.sha256(content).hexdigest()
                self.assertEqual(expected_hash, result[relpath])
        self.assertEqual(expected_file_paths, set(result))

    def test_slurp_files_with_identical_content_saved_only_once(self):

        def add_revision_two(cwd, tree):
            src_path = join(cwd, 'branch/trunk/common')
            dst_path = join(cwd, 'branch/trunk/also-common')
            with open(dst_path, 'w') as f:
                f.write(open(src_path).read())
            tree.add(['also-common'])
            return tree.commit('revision 2')

        with charm_branch(extra=[add_revision_two]) as (
                branch_dir, revision, path):
            branch = Branch.open(branch_dir)
            tree = branch.repository.revision_tree(revision)
            with read_locked(tree):
                result = slurp_files(self.fs, tree)
        expected_file_paths = set()
        for dirpath, dirnames, filenames in os.walk(path):
            for filename in filenames:
                filepath = os.path.join(dirpath, filename)
                if stat.S_ISLNK(os.lstat(filepath).st_mode):
                    continue
                relpath = filepath[len(path) + 1:]
                expected_file_paths.add(relpath)
        expected_file_paths.add('also-common')
        self.assertEqual(result['common'], result['also-common'])


class TestBundleFiles(MongoTestBase):

    def setUp(self):
        super(TestBundleFiles, self).setUp()
        with bundle_branch() as (branch_dir, revision, self.path):
            self.branch = Branch.open(branch_dir)
            self.tree = self.branch.repository.revision_tree(revision)
            with read_locked(self.tree):
                self.hashes = slurp_files(self.fs, self.tree)
        basket_data = {'_id': '~bac/sample_bundle/1'}
        basket_data['file_hashes'] = quote_yaml(self.hashes)
        self.db.baskets.save(basket_data)
        self.bundle_data = dict(owner='bac',
                                basket_name='sample_bundle',
                                basket_revision=1,
                                name='wilma')
        self.bundle = Bundle(self.bundle_data)

    def test_get_file(self):
        icon_fn = 'icon.svg'
        self.assertIn(icon_fn, self.hashes)
        fh = self.bundle.get_file(self.db, icon_fn, self.fs)
        content = fh.read()
        real_icon_svg_path = join(self.path, icon_fn)
        with open(real_icon_svg_path) as f:
            real_content = f.read()
        # We cannot assert content and real_content are equal due to
        # viewbox manipulation in slurp_files.
        self.assertIn('mediawiki', content)
        self.assertIn('mediawiki', real_content)

    def test_get_file_list(self):
        filelist = sorted(self.bundle.get_file_list(self.db, self.fs))
        expected = sorted(('icon.svg', 'bundles.yaml', 'README.md'))
        self.assertEqual(expected, filelist)


class TestBundleLoadingCharms(MongoTestBase):

    def setUp(self):
        """Create 3 of the same charms at different revisions"""
        super(TestBundleLoadingCharms, self).setUp()
        charm_args = {
            'name': 'wikipedia',
            'series': 'precise',
            'owner': 'charmers',
            'promulgated': True
        }
        charm_args['revision'] = 1
        charm_args['store_revision'] = 1
        factory.makeCharm(self.db, **charm_args)
        charm_args['revision'] = 2
        charm_args['store_revision'] = 2
        factory.makeCharm(self.db, **charm_args)
        charm_args['revision'] = 3
        charm_args['store_revision'] = 3
        factory.makeCharm(self.db, **charm_args)

    def test_resolve_by_store_url(self):
        data = {
            'charm': 'cs:precise/wikipedia-1'
        }
        charm_description = BundledCharmDescription('wiki', data)
        charm = resolve_charm_from_description(self.db, charm_description)
        self.assertIsNotNone(charm)
        self.assertEqual('~charmers/precise/wikipedia/1', charm['_id'])
        self.assertEqual('wikipedia', charm['name'])
        self.assertEqual(1, charm['store_data']['revision'])
        self.assertEqual(1, charm['revision'])

    def test_resolve_by_store_url_sans_version(self):
        data = {
            'charm': 'cs:precise/wikipedia'
        }
        charm_description = BundledCharmDescription('wiki', data)
        charm = resolve_charm_from_description(self.db, charm_description)
        self.assertIsNotNone(charm)
        self.assertEqual('~charmers/precise/wikipedia/3', charm['_id'])
        self.assertEqual('wikipedia', charm['name'])
        self.assertEqual(3, charm['store_data']['revision'])
        self.assertEqual(3, charm['revision'])

    def test_resolve_by_name_and_series(self):
        data = {
            'charm': 'wikipedia'
        }
        charm_description = BundledCharmDescription(
            'wiki', data, series='precise')
        charm = resolve_charm_from_description(self.db, charm_description)

        self.assertIsNotNone(charm)
        self.assertEqual('wikipedia', charm['name'])
        self.assertEqual(3, charm['store_data']['revision'])
        self.assertEqual(3, charm['revision'])

    def test_resolve_by_branch(self):
        data = {
            'branch': 'lp:charms/precise/wikipedia'
        }
        charm_description = BundledCharmDescription(
            'wiki', data, series='precise')
        charm = resolve_charm_from_description(self.db, charm_description)
        self.assertIsNotNone(charm)
        self.assertEqual('wikipedia', charm['name'])
        self.assertEqual(3, charm['store_data']['revision'])
        self.assertEqual(3, charm['revision'])

    def test_resolve_by_branch_with_revision(self):
        data = {
            'branch': 'lp:charms/precise/wikipedia@2'
        }
        charm_description = BundledCharmDescription('wiki', data)
        charm = resolve_charm_from_description(self.db, charm_description)
        self.assertIsNotNone(charm)
        self.assertEqual('wikipedia', charm['name'])
        self.assertEqual(2, charm['store_data']['revision'])
        self.assertEqual(2, charm['revision'])

    def test_not_enough_info(self):
        data = {
            'name': 'wikipedia'
        }
        charm_description = BundledCharmDescription('wiki', data)
        charm = resolve_charm_from_description(self.db, charm_description)
        self.assertIsNone(charm)


class TestFeaturedSource(MongoTestBase):

    def test_set_and_get_work(self):
        # FeaturedSource instances manage a set of featured charms in a
        # MongoDB collection.
        charm_data_1 = factory.get_charm_json(
            name='name1', owner='owner1', series='series1', promulgated=True)
        charm_data_2 = factory.get_charm_json(
            name='name2', owner='owner2', series='series2', promulgated=True)
        charm_data_3 = factory.get_charm_json(
            name='name3', owner='owner3', series='series3', promulgated=True)
        bundle_data_1 = factory.get_bundle_data(
            name='name4', owner='owner4', basket_with_rev='basket1/1')
        featured_source = FeaturedSource(
            self.db.create_collection(factory.random_string(10)))
        featured_source.set_featured(charm_data_1, CHARM)
        featured_source.set_featured(charm_data_2, CHARM)
        featured_source.set_featured(bundle_data_1, BUNDLE)
        # The first two charms and bundle are featured, but the third is not.
        self.assertTrue(featured_source.is_featured(charm_data_1, CHARM))
        self.assertTrue(featured_source.is_featured(charm_data_2, CHARM))
        self.assertTrue(featured_source.is_featured(bundle_data_1, BUNDLE))
        self.assertFalse(featured_source.is_featured(charm_data_3, CHARM))

    def test_can_create_from_database(self):
        # A convienience constructor is proviede that accepts a database.
        charm_data = factory.get_charm_json(
            name='name1', owner='owner1', series='series1')
        featured_source = FeaturedSource.from_db(self.db)
        featured_source.set_featured(charm_data, 'charm')
        self.assertTrue(featured_source.is_featured(charm_data, 'charm'))

    def test_clearing_featured(self):
        # Once featured, an item can be removed from the featured set.
        charm_data = factory.get_charm_json(
            name='name1', owner='owner1', series='series1')
        featured_source = FeaturedSource.from_db(self.db)
        featured_source.set_featured(charm_data, 'charm')
        self.assertTrue(featured_source.is_featured(charm_data, 'charm'))
        # Now we remove the item's featured status and verify that it worked.
        featured_source.clear_featured(charm_data, 'charm')
        self.assertFalse(featured_source.is_featured(charm_data, 'charm'))
        # If the item is cleared again, nothing happens.
        featured_source.clear_featured(charm_data, 'charm')
        self.assertFalse(featured_source.is_featured(charm_data, 'charm'))

    def test_multiple_sets(self):
        # Setting an item as featured more than once will not include it more
        # than once in the set of featured items.
        charm_data = factory.get_charm_json(
            name='name1', owner='owner1', series='series1')
        featured_source = FeaturedSource.from_db(self.db)
        featured_source.set_featured(charm_data, 'charm')
        featured_source.set_featured(charm_data, 'charm')
        featured_source.set_featured(charm_data, 'charm')
        self.assertEqual(len(list(featured_source.collection.find())), 1)

    def test_get_featured_typed_ids_all(self):
        # Retrieve the unversioned ids of the featured items as used by
        # ElasticSearch.
        charm_data_1 = factory.get_charm_json(
            name='name1', owner='owner1', series='series1', promulgated=True)
        charm_data_2 = factory.get_charm_json(
            name='name2', owner='owner2', series='series2', promulgated=True)
        bundle_data_1 = factory.get_bundle_data(
            name='name3', owner='owner3', basket_with_rev='basket1/1')
        featured_source = FeaturedSource.from_db(self.db)
        featured_source.set_featured(charm_data_1, CHARM)
        featured_source.set_featured(charm_data_2, CHARM)
        featured_source.set_featured(bundle_data_1, BUNDLE)
        typed_ids = featured_source.get_featured_typed_ids(doctype=None)
        expected = [
            {'_id': '~owner1/series1/name1', '_type': CHARM},
            {'_id': '~owner2/series2/name2', '_type': CHARM},
            {'_id': '~owner3/basket1/name3', '_type': BUNDLE},
        ]
        self.assertEqual(expected, typed_ids)

    def test_get_featured_typed_ids_for_doctype(self):
        # Retrieve the unversioned ids of the featured doctype as used by
        # ElasticSearch.
        charm_data_1 = factory.get_charm_json(
            name='name1', owner='owner1', series='series1', promulgated=True)
        bundle_data_1 = factory.get_bundle_data(
            name='name3', owner='owner3', basket_with_rev='basket1/1')
        featured_source = FeaturedSource.from_db(self.db)
        featured_source.set_featured(charm_data_1, CHARM)
        featured_source.set_featured(bundle_data_1, BUNDLE)
        typed_ids = featured_source.get_featured_typed_ids(doctype=CHARM)
        expected = [{'_id': '~owner1/series1/name1', '_type': CHARM}]
        self.assertEqual(expected, typed_ids)
