# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type


from datetime import datetime
from textwrap import dedent
from time import sleep

from pyelasticsearch import ElasticSearch
from pyelasticsearch.exceptions import (
    ElasticHttpError,
    ElasticHttpNotFoundError,
)
import yaml

from charmworld.charmstore import (
    get_address,
    make_store_url,
)
from charmworld.models import (
    Bundle,
    Charm,
    store_bundles,
)
from charmworld.search import (
    BUNDLE,
    CHARM,
    ElasticSearchClient,
    charm_exact_fields,
    charm_free_text_fields,
    IncompatibleMapping,
    IndexMissing,
    IndexNotReady,
    InvalidCharmType,
    update,
    NegativeLimit,
    reindex,
    SearchServiceNotAvailable,
)
from charmworld.testing import (
    factory,
    temp_index_client,
    TestCase,
)
from charmworld.utils import get_ini


class TestTempIndexClient(TestCase):

    def test_temp_index_client(self):
        with temp_index_client() as client:
            self.assertNotEqual('charms', client.index_name)
            client._client.update_settings(client.index_name, {})
        with self.assertRaises(ElasticHttpNotFoundError):
            client._client.update_settings(client.index_name, {})


class TestElasticSearchClient(TestCase):

    def setUp(self):
        super(TestElasticSearchClient, self).setUp()
        self.use_index_client()

    def exists_in_index(self, id_, kind=CHARM):
        client = self.index_client
        if kind == CHARM:
            return client.get(id_) is not None
        try:
            result = client._client.get(client.index_name, kind, id_)
        except ElasticHttpNotFoundError:
            return False
        return result['exists']

    def test_index_charm(self):
        foo_charm = factory.get_charm_json()
        self.assertFalse(self.exists_in_index(foo_charm['_id']))
        self.index_client.index_charm(foo_charm)
        self.assertTrue(self.exists_in_index(foo_charm['_id']))

    def test_index_charm_max_retries_exceeded(self):
        # The Gopher port should be inactive.
        client = ElasticSearchClient(
            ElasticSearch(['http://localhost:70']), 'foo')
        with self.assertRaises(SearchServiceNotAvailable):
            client.index_charm({'_id': 'id', 'owner': 'foo', 'series': 'bar',
                                'name': 'baz'})

    def test_index_charms(self):
        foo_charm = factory.get_charm_json()
        bar_charm = factory.get_charm_json()
        self.assertFalse(self.exists_in_index(foo_charm['_id']))
        self.assertFalse(self.exists_in_index(bar_charm['_id']))
        self.index_client.index_charms([foo_charm, bar_charm])
        self.assertTrue(self.exists_in_index(foo_charm['_id']))
        self.assertTrue(self.exists_in_index(bar_charm['_id']))

    def test_index_charms_no_charms(self):
        # Assert that no exception is raised.
        with self.assertRaises(AssertionError):
            with self.assertRaises(BaseException):
                self.index_client.index_charms([])

    def test_delete_charm(self):
        charm_data = factory.get_charm_json()
        self.index_client.index_charm(charm_data)
        self.assertTrue(self.exists_in_index(charm_data['_id']))
        self.index_client.delete_charm(charm_data)
        self.assertFalse(self.exists_in_index(charm_data['_id']))

    def test_update_aliased_None(self):
        alias = ElasticSearchClient.from_settings(get_ini(), 'foo')
        alias.update_aliased(self.index_client.index_name, [])
        alias.update_aliased(None, [])
        self.assertEqual([self.index_client.index_name], alias.get_aliased())
        alias.update_aliased(None, [self.index_client.index_name])
        self.assertEqual([], alias.get_aliased())

    def makeCharm(self, *args, **kwargs):
        charm = factory.get_charm_json(*args, **kwargs)
        self.index_client.index_charm(charm)
        return charm

    def makeBundle(self, *args, **kwargs):
        bundle = factory.get_bundle_data(*args, **kwargs)
        bundle['_id'] = Bundle(bundle).id
        self.index_client.index_bundle(bundle)
        return bundle

    def test_search(self):
        result_keys = ['results', 'search_time', 'result_total', 'matches',
                       'matches_human_readable_estimate']
        client = self.index_client
        foo_charm = Charm(self.makeCharm())
        results = client.search('foo')
        self.assertItemsEqual(result_keys, results.keys())
        self.assertEqual([], results['results']['charm'])
        results = client.search(foo_charm.name)['results']['charm']
        self.assertEqual(1, len(results))
        self.assertEqual(foo_charm, results[0]['data'])
        # The absolute value of weight is not defined, so we just ensure
        # it's present.
        self.assertItemsEqual(['data', 'weight'], results[0].keys())

    def test_search_bundle(self):
        bundle = Bundle(self.makeBundle())
        client = self.index_client
        results = client.search(bundle.name)['results']
        self.assertEqual(1, len(results[BUNDLE]))
        self.assertEqual(bundle, results[BUNDLE][0]['data'])

    def _setup_store_bundles(self):
        self.basket_name = 'inigo'
        _id = '~abentley/{}/5/wordpress-stage'.format(self.basket_name)
        deployer_config = dedent("""\
            wordpress-stage:
                series: precise
                services:
                    blog:
                        charm: cs:precise/wordpress
                        constraints: mem=2
                        options:
                            tuning: optimized
                            engine: apache
                    db:
                        charm: cs:precise/mysql
        """)
        parsed = yaml.safe_load(deployer_config)
        store_bundles(
            None, parsed, 'abentley', '{}/5'.format(self.basket_name),
            None, None, None,
            None, index_client=self.index_client)
        return _id

    def test_store_bundles_basket_name_indexed(self):
        # Ensure that bundles stored into the index via the `store_bundles`
        # function are searchable via the index.
        _id = self._setup_store_bundles()

        # Basket name is indexed.
        results = self.index_client.search(
            self.basket_name)['results'][BUNDLE]
        self.assertEqual(1, len(results))
        self.assertEqual(_id, results[0]['data'].id)

    def test_store_bundles_bundle_name_indexed(self):
        # Ensure that bundles stored into the index via the `store_bundles`
        # function are searchable via the index.
        _id = self._setup_store_bundles()

        # Bundle name is indexed.
        results = self.index_client.search(
            'wordpress-stage')['results'][BUNDLE]
        self.assertEqual(1, len(results))
        self.assertEqual(_id, results[0]['data'].id)

    def test_store_bundles_series_indexed(self):
        _id = self._setup_store_bundles()
        # Series is indexed.
        results = self.index_client.search('precise')['results'][BUNDLE]
        self.assertEqual(1, len(results))
        self.assertEqual(_id, results[0]['data'].id)

    def test_store_bundles_owner_indexed(self):
        _id = self._setup_store_bundles()
        # Owner is indexed.
        results = self.index_client.search('abentley')['results'][BUNDLE]
        self.assertEqual(1, len(results))
        self.assertEqual(_id, results[0]['data'].id)

    def test_store_bundles_charm_names_indexed(self):
        _id = self._setup_store_bundles()
        # Charm names within services are index.
        results = self.index_client.search('mysql')['results'][BUNDLE]
        self.assertEqual(1, len(results))
        self.assertEqual(_id, results[0]['data'].id)

    def test_search_charms_and_bundles_same_name(self):
        charm = Charm(self.makeCharm(name='mozilla'))
        bundle = Bundle(self.makeBundle(name='mozilla'))
        client = self.index_client
        search_results = client.search('mozilla')
        results = search_results['results']
        self.assertEqual(2, len(results))
        self.assertEqual(2, search_results['result_total'])
        self.assertEqual(2, search_results['matches'])
        self.assertEqual(1, len(results['charm']))
        self.assertEqual(1, len(results['bundle']))
        self.assertEqual(charm, results['charm'][0]['data'])
        self.assertEqual(bundle, results['bundle'][0]['data'])

    def test_search_charms_and_bundles_description(self):
        charm = Charm(self.makeCharm(description='a mozilla charm'))
        bundle = Bundle(self.makeBundle(description='a mozilla bundle'))
        client = self.index_client
        search_results = client.search('mozilla')
        results = search_results['results']
        self.assertEqual(2, len(results))
        self.assertEqual(2, search_results['result_total'])
        self.assertEqual(2, search_results['matches'])
        self.assertEqual(1, len(results['charm']))
        self.assertEqual(1, len(results['bundle']))
        self.assertEqual(charm, results['charm'][0]['data'])
        self.assertEqual(bundle, results['bundle'][0]['data'])

    def test_search_not_ready(self):

        class FakeElasticSearch:

            @staticmethod
            def status(index):
                return {'indices': {'foo': {}}}

            @staticmethod
            def aliases():
                return {}

        client = ElasticSearchClient(FakeElasticSearch, 'foo')
        with self.assertRaises(IndexNotReady):
            client.search('bar')

    def test_max_retries_exceeded_search(self):
        # The Gopher port should be inactive.
        client = ElasticSearchClient(
            ElasticSearch(['http://localhost:70']), 'foo')
        with self.assertRaises(SearchServiceNotAvailable):
            client.search('bar')

    def search_ids(self, terms):
        client = self.index_client
        hits = client.search(terms)['results']['charm']
        return [hit['data']._id for hit in hits]

    def test_search_matches_on_search_terms(self):
        fields = charm_free_text_fields.keys() + charm_exact_fields
        fields = [field.split('.')[0] for field in fields]
        fields.extend(['requires', 'provides'])
        charm = factory.get_charm_json(categories=['unknown'])
        # Ensure all fields have unique values.
        charm['address'] = 'cs:bogus-address'
        charm['owner'] = 'steve'
        charm['series'] = 'grumpy'
        charm['name'] = 'arthur'
        charm['label'] = 'organic'
        charm['bname'] = 'my-dev-branch'
        charm['i_provides'] = ['provisions']
        charm['provides']['website']['interface'] = 'gopher'
        charm['requires']['database']['interface'] = 'reiserfs'
        charm['_id'] = 'asdf'
        # Best-effort ensure there's no overlap between search terms
        items = charm.items()
        for num, (key, value) in enumerate(items):
            if not isinstance(value, basestring):
                continue
            for key2, value2 in items[num + 1:]:
                if not isinstance(value2, basestring):
                    continue
                self.assertNotIn(value, value2, '%s in %s' % (key, key2))
                self.assertNotIn(value2, value, '%s in %s' % (key2, key))
        self.index_client.index_charm(charm)
        for key, query in items:
            # Skip ints and lists.
            if key in ('revision', 'downloads_in_past_30_days', 'downloads'):
                continue
            elif key in ('hooks', 'i_requires', 'i_provides', 'categories'):
                query = query[0]
            elif key == 'changes':
                query = query[0]['message']
            elif key in ('provides', 'requires'):
                query = query.values()[0].values()[0]
            elif key == 'config':
                query = query.values()[0][0]['description']
            elif key == 'store_data':
                query = query['digest']
            elif key == 'files':
                query = query['readme']['filename']
            elif key in ('first_change', 'last_change'):
                query = query['message']
            elif key == 'proof':
                query = query['w'][0]
            self.assertIsInstance(query, (basestring, bool))
            if isinstance(query, basestring):
                query = '"%s"' % self.index_client.escape_query(query)
            else:
                self.assertIsInstance(query, bool)
            ids = self.search_ids(query)
            if key in fields:
                self.assertEqual(1, len(ids), 'No hit for %s' % key)
            else:
                self.assertEqual(0, len(ids), 'Undesired hit on %s' % key)

    def test_official_charms_boost(self):
        """An official version of a charm rates 10x higher."""
        charm = factory.get_charm_json(promulgated=True, owner='foo')
        official_id = charm['_id']
        unofficial_id = official_id + '-1'
        client = self.index_client
        client.index_charm(charm)
        charm['owner'] = 'jrandom'
        charm['_id'] = unofficial_id
        charm['store_url'] = get_address(charm, short=False)
        client.index_charm(charm)
        results = client.search('terminal')['results']['charm']
        by_id = dict((result['data']._id, result) for result in results)
        self.assertAlmostEqual(by_id[official_id]['weight'],
                               by_id[unofficial_id]['weight'] * 10)

    def test_unreviewed_charmers_not_boosted(self):
        """Unpromulgated charms are not boosted even if owned by charmers."""
        charm = factory.get_charm_json(promulgated=False)
        charmers_id = charm['_id']
        jrandom_id = charmers_id + '-1'
        client = self.index_client
        client.index_charm(charm)
        charm['owner'] = 'jrandom'
        charm['_id'] = jrandom_id
        client.index_charm(charm)
        results = client.search('terminal')['results']['charm']
        by_id = dict((result['data']._id, result) for result in results)
        self.assertAlmostEqual(by_id[charmers_id]['weight'],
                               by_id[jrandom_id]['weight'])

    def get_charm_bundle_test_data(self):
        one = self.makeCharm(name='name1', promulgated=True)
        maintainer2 = 'jrandom@example.com (J. Random Hacker)'
        provides2 = {'cookie-factory': {
            'interface': 'imap',
        }}
        requires2 = {'postal-service': {
            'interface': 'envelope',
        }}
        options2 = {
            'foobar': {
                'type': 'str',
                'default': 'baz',
                'description': 'quxx',
            },
        }
        files2 = {
            'readme': {
                'filename': 'README.md',
                'subdir': '',
            },
            'config-changed': {
                'filename': 'config-changed',
                'subdir': 'hooks',
            },
        }
        two = self.makeCharm(owner='jrandom', revision=3,
                             series='warty', summary='Charm two',
                             name='name2', revno=13,
                             maintainer=maintainer2,
                             commit_message='message2',
                             provides=provides2,
                             requires=requires2, options=options2,
                             files=files2, subordinate=True)
        bundle = self.makeBundle(name='name3')
        return one, two, bundle

    def test_api_search_all(self):
        # With no parameters, all charms and bundles are returned.
        one, two, bundle = self.get_charm_bundle_test_data()
        result = self.get_charms_and_bundles(doctype=None)
        self.assertEqual(3, len(result))
        one_url = make_store_url(1, get_address(one, short=True))
        self.assertEqual(one, result[0])
        two_url = make_store_url(1, get_address(two, short=False))
        self.assertEqual(two, result[1])
        self.assertNotEqual(one['_id'], two['_id'])
        self.assertNotEqual(one_url, two_url)
        self.assertEqual(bundle, result[2])

    def test_api_search_charms_only(self):
        """With no parameters, all charms are returned."""
        one, two, bundle = self.get_charm_bundle_test_data()
        result = self.get_charms_and_bundles(doctype=CHARM)
        self.assertEqual(2, len(result))
        self.assertEqual(one, result[0])
        self.assertEqual(two, result[1])
        self.assertNotEqual(one['_id'], two['_id'])

    def test_api_search_bundles_only(self):
        """Passing doctype of 'bundles' retrieves only bundles."""
        one, two, bundle = self.get_charm_bundle_test_data()
        result = self.get_charms_and_bundles(doctype=BUNDLE)
        self.assertEqual(1, len(result))
        self.assertEqual(bundle, result[0])

    def test_autocomplete_matches_on_names(self):
        matching_charm = self.makeCharm(name='foo')
        differing_charm = self.makeCharm(name='bar')
        for field in charm_free_text_fields:
            if '.' in field or field == 'name':
                continue
            differing_charm[field] = 'foo'
        for field in charm_exact_fields:
            differing_charm[field] = 'foo'
        self.index_client.index_charm(differing_charm)
        result = self.index_client.api_search('foo', autocomplete=True)
        ids = [r['data']['_id'] for r in result]
        self.assertEqual([matching_charm['_id']], ids)

    def test_autocomplete_matches_on_prefix(self):
        charm1 = self.makeCharm(name='foo')
        charm2 = self.makeCharm(name='foobar')
        self.makeCharm(name='barfoo')
        result = self.index_client.api_search('foo', autocomplete=True)
        ids = [r['data']['_id'] for r in result]
        self.assertItemsEqual([charm1['_id'], charm2['_id']], ids)

    def get_charms_and_bundles(self, text='', type_=None, doctype=CHARM,
                               *args, **kwargs):
        filters = dict((key, [value]) for key, value in kwargs.items())
        if type_ is not None and isinstance(type_, basestring):
            type_ = [type_]
        result = self.index_client.api_search(
            text, filters, type_, doctype=doctype)
        charms_and_bundles = [r['data'] for r in result]
        return sorted(charms_and_bundles, key=lambda x: x['name'])

    def _test_filter(self, filter_str, make_charm=None):
        if make_charm is None:
            def make_charm(value, warty=False):
                kwargs = {filter_str: value}
                if filter_str != 'series' and warty:
                    kwargs['series'] = 'warty'
                return self.makeCharm(**kwargs)
        one = make_charm('common')
        self.index_client.index_charm(one)
        two = make_charm('common', warty=True)
        self.index_client.index_charm(two)
        three = make_charm('different')
        self.index_client.index_charm(three)
        four = make_charm('different-again')
        self.index_client.index_charm(four)
        result = self.get_charms_and_bundles(**{filter_str: 'common'})
        self.assertEqual(2, len(result))
        self.assertEqual(set([one['store_url'], two['store_url']]),
                         set(charm['store_url'] for charm in result))
        result = self.get_charms_and_bundles(**{filter_str: 'different'})
        self.assertEqual(set([three['store_url']]),
                         set(charm['store_url'] for charm in result))
        self.assertEqual(1, len(result))
        result = self.get_charms_and_bundles(
            **{filter_str: ['different', 'common']})
        self.assertEqual(3, len(result))

    def test_charms_name_filter(self):
        """Charms can be filtered by name."""
        self._test_filter('name')

    def test_charms_series_filter(self):
        """Charms can be filtered by series."""
        self._test_filter('series')

    def test_charms_owner_filter(self):
        """Charms can be filtered by owner."""
        self._test_filter('owner')

    def test_charms_provides_filter(self):
        """Charms can be filtered by provided interfaces."""
        def make_charm(value, warty=False):
            provides = {'foo': {'interface': value}}
            return self.makeCharm(provides=provides)
        self._test_filter('i_provides', make_charm)

    def test_charms_requires_filter(self):
        """Charms can be filtered by required interfaces."""
        def make_charm(value, warty=False):
            requires = {'foo': {'interface': value}}
            return self.makeCharm(requires=requires)
        self._test_filter('i_requires', make_charm)

    def test_charms_type_filter(self):
        """Filtering by type works as well as possible.

        Charms are considered 'approved' if they are promulgated,
        'community' otherwise.  'environment' causes no charms
        to be returned.
        """
        self.makeCharm(owner='alice', name='ant', promulgated=True)
        self.makeCharm(owner='steve', name='bat')
        result = self.get_charms_and_bundles(type_='community')
        self.assertEqual(['steve'], [charm['owner'] for charm in result])
        result = self.get_charms_and_bundles(type_='approved')
        self.assertEqual(['alice'], [charm['owner'] for charm in result])
        result = self.get_charms_and_bundles(type_='environment')
        self.assertEqual([], [charm['owner'] for charm in result])
        with self.assertRaises(InvalidCharmType):
            self.get_charms_and_bundles(type_='foo')
        result = self.get_charms_and_bundles(type_=['approved', 'community'])
        self.assertEqual(['alice', 'steve'],
                         [charm['owner'] for charm in result])

    def test_type_search_considers_promulgated(self):
        reviewed = self.makeCharm(promulgated=True)['_id']
        self.makeCharm(promulgated=False)
        results = self.get_charms_and_bundles(type_=['approved'])
        self.assertEqual([reviewed], [result['_id'] for result in results])

    def test_charms_text_search(self):
        abc_charm = self.makeCharm(summary='abc def')
        self.makeCharm(summary='ghi jkl')
        charms = self.get_charms_and_bundles(text='def')
        self.assertEqual(1, len(charms))
        self.assertEqual(abc_charm['_id'], charms[0]['_id'])

    def test_charms_text_search_respects_filters(self):
        official = self.makeCharm(summary='abc def', promulgated=True)
        self.makeCharm(promulgated=True)
        self.makeCharm(summary='abc def', owner='steve')
        self.makeCharm(owner='steve')
        result = self.get_charms_and_bundles(type_=['approved'], text='def')
        self.assertEqual([official['_id']],
                         [charm['_id'] for charm in result])

    def test_charms_limit(self):
        """Limit provides a maximum for charm numbers."""
        result = self.index_client.api_search('', {}, None, 6)
        self.assertEqual(0, len(result))
        self.makeCharm()
        self.makeCharm(owner='jrandom')
        result = self.index_client.api_search('', {}, None, 6)
        self.assertEqual(2, len(result))
        result = self.index_client.api_search('', {}, None, 1)
        self.assertEqual(1, len(result))
        result = self.index_client.api_search('', {}, None, 0)
        self.assertEqual(0, len(result))
        with self.assertRaises(NegativeLimit):
            result = self.index_client.api_search('', {}, None, -1)

    def test_api_search_with_store_error(self):
        """api_search() does not return charms withs errors by default."""
        self.makeCharm(name='evil-charm', charm_error=True)
        self.makeCharm(name='good-charm')
        self.makeBundle(name='great-bundle')
        result = self.index_client.api_search('', {}, None, doctype=None)
        self.assertEqual(2, len(result))
        names = set(charm['data']['name'] for charm in result)
        self.assertEqual(set(('good-charm', 'great-bundle')), names)
        # Charms with error are included in the search result
        # if the parameter valid_only is set to False.  Bundles are
        # always included.
        result = self.index_client.api_search(
            '', {}, None, valid_only=False, doctype=None)
        self.assertEqual(3, len(result))
        names = set(charm['data']['name'] for charm in result)
        self.assertEqual(
            set(('good-charm', 'evil-charm', 'great-bundle')), names)

    def test_search_with_store_error(self):
        """search() does not return charms withs errors by default."""
        self.makeCharm(name='evil-charm', owner='foo', charm_error=True)
        self.makeCharm(name='good-charm', owner='foo')
        self.makeBundle(name='great-bundle', owner='foo')
        result = self.index_client.search('foo')
        self.assertEqual(2, result['matches'])
        data = result['results']['charm'][0]['data']
        self.assertEqual('good-charm', data.name)
        data = result['results']['bundle'][0]['data']
        self.assertEqual('great-bundle', data.name)
        # Charms with error are included in the search result
        # if the parameter valid_only is set to False.
        result = self.index_client.search('foo', valid_only=False)
        self.assertEqual(3, result['matches'])
        names = set(e['data'].name for e in result['results']['charm'])
        names.update(set(e['data'].name for e in result['results']['bundle']))
        self.assertEqual(
            set(('good-charm', 'evil-charm', 'great-bundle')), names)

    def test_search_charm_with_error(self):
        """searches include charms with processing errors."""
        charm = Charm(self.makeCharm())
        result = self.index_client.search(charm.name)
        self.assertEqual(charm, result['results']['charm'][0]['data'])
        charm._representation['error'] = {'error': 'error text'}
        self.index_client.index_charm(charm._representation)
        result = self.index_client.search(charm.name)
        self.assertEqual(charm, result['results']['charm'][0]['data'])

    def test_api_search_charm_with_error(self):
        """Search results include charms with processing errors."""
        charm = self.makeCharm()
        charm['error'] = {'error': 'error text'}
        self.index_client.index_charm(charm)
        result = self.index_client.api_search(charm['name'])
        self.assertEqual([charm], [r['data'] for r in result])

    def test_search_charm_sort_most_downloaded(self):
        """Specifying 'downloaded' sorts results appropriately."""
        self.makeCharm(name='popular1', downloads=3)
        self.makeCharm(name='popular2', downloads=1)
        self.makeCharm(name='popular3', downloads=2)
        order132 = ['popular1', 'popular3', 'popular2']
        result = self.index_client.api_search()
        self.assertNotEqual(order132, [r['data']['name'] for r in result])
        result = self.index_client.api_search(sort='downloaded')
        self.assertEqual(order132, [r['data']['name'] for r in result])

    def test_search_charm_sort_newest(self):
        """Specifying 'new' sorts results appropriately."""
        self.makeCharm(name='new1', date_created=datetime(2013, 3, 1),
                       downloads=50)
        self.makeCharm(name='new2', date_created=datetime(2013, 1, 2),
                       downloads=20)
        self.makeCharm(name='new3', date_created=datetime(2013, 1, 1),
                       downloads=100)
        order123 = ['new1', 'new2', 'new3']
        result = self.index_client.api_search(sort='downloaded')
        self.assertNotEqual(order123, [r['data']['name'] for r in result])
        result = self.index_client.api_search(sort='new')
        self.assertEqual(order123, [r['data']['name'] for r in result])

    def test_search_unknown_sort(self):
        """Specifying an unknown sort raises an exception."""
        with self.assertRaises(ValueError) as e:
            self.index_client.api_search(sort='foo')
        self.assertEqual(str(e.exception), 'Invalid sort: foo.')

    def test_update_alias(self):
        alias = ElasticSearchClient(self.index_client._client,
                                    'temp_index_alias')
        self.assertEqual([], alias.get_aliased())
        alias.update_aliased(self.index_client.index_name, [])
        self.assertEqual([self.index_client.index_name], alias.get_aliased())
        self.use_context(temp_index_client('foo'))
        alias.update_aliased('foo', [])
        self.assertItemsEqual([self.index_client.index_name, 'foo'],
                              alias.get_aliased())
        self.use_context(temp_index_client('bar'))
        alias.update_aliased('bar', alias.get_aliased())
        self.assertEqual(['bar'], alias.get_aliased())

    def test_create_replacement_charm_indexed_in_place(self):
        charm_data = {'_id': 'a', 'name': 'foo', 'owner': 'bar',
                      'series': 'baz'}
        self.index_client.index_charm(charm_data)
        copy = self.index_client.create_replacement('index-copy')
        self.addCleanup(copy.delete_index)
        mapping = copy.get_mapping()[CHARM]
        self.assertIn('series', mapping['properties']['data']['properties'])
        self.assertEqual(charm_data, copy.get('a'))

    def test_create_replacement_bundle_indexed_in_place(self):
        bundle_data = {'_id': 'a', 'name': 'foo', 'owner': 'bar',
                       'basket_name': 'basket', 'data': {'series': 'baz'}}
        self.index_client.index_bundle(bundle_data)
        copy = self.index_client.create_replacement('index-copy')
        self.addCleanup(copy.delete_index)
        mapping = copy.get_mapping(doctype=BUNDLE)[BUNDLE]
        self.assertIn('series', mapping['properties']['data']['properties'])
        self.assertEqual(
            {u'_id': u'~bar/basket/foo', u'data': bundle_data},
            copy.get(Bundle(bundle_data).id, BUNDLE))

    def test_create_replacement_charms_replacing_old(self):
        charm_data = {'_id': 'a', 'name': 'foo', 'owner': 'bar',
                      'series': 'baz'}
        charm_data2 = {'_id': 'a', 'name': 'bar', 'owner': 'bar',
                       'series': 'baz'}
        self.index_client.index_charm(charm_data)
        copy = self.index_client.create_replacement(
            charms=[charm_data2])
        self.addCleanup(copy.delete_index)
        mapping = copy.get_mapping()[CHARM]
        self.assertIn('series', mapping['properties']['data']['properties'])
        self.assertEqual(charm_data2, copy.get('a'))

    def test_create_replacement_bundles_replacing_old(self):
        bundle_data = {'_id': 'a', 'name': 'foo', 'owner': 'bar',
                       'series': 'first', 'basket_name': 'basket'}
        bundle_data2 = {'_id': 'a', 'name': 'foo', 'owner': 'bar',
                        'series': 'second', 'basket_name': 'basket'}
        self.index_client.index_bundle(bundle_data)
        copy = self.index_client.create_replacement(
            bundles=[bundle_data2])
        self.addCleanup(copy.delete_index)
        mapping = copy.get_mapping(doctype=BUNDLE)[BUNDLE]
        self.assertIn('series', mapping['properties']['data']['properties'])
        self.assertEqual(
            {u'_id': u'~bar/basket/foo', u'data': bundle_data2},
            copy.get(Bundle(bundle_data).id, BUNDLE))

    def test_create_replacement_missing(self):
        client = ElasticSearchClient.from_settings(get_ini(), 'temp-index')
        # This should not raise an exception, even though the index has not
        # been created.
        copy = client.create_replacement()
        self.addCleanup(copy.delete_index)

    def test_put_mapping_incompatible_mapping_error(self):
        # The error we get from an incompatible mapping is IncompatibleMapping
        context = temp_index_client(name='foo', put_mapping=False)
        client = self.use_context(context)
        put_incompatible_mapping(client)
        with self.assertRaises(IncompatibleMapping):
            client.put_mapping()

    def test_put_mapping_missing_index(self):
        # The error we get from a missing index is IndexMissing
        client = ElasticSearchClient.from_settings(get_ini(), 'temp-index')
        client.stable_status_timeout = '0.5s'
        with self.assertRaises(IndexMissing):
            client.put_mapping()

    def test_get_mapping_default(self):
        # Calling get_mapping with no doctype returns charms.
        client = self.index_client
        client.put_mapping()
        mapping = client.get_mapping()
        self.assertEqual([CHARM], mapping.keys())

    def test_get_mapping_charms(self):
        # Calling get_mapping with CHARM returns charms.
        client = self.index_client
        client.put_mapping()
        mapping = client.get_mapping(CHARM)
        self.assertEqual([CHARM], mapping.keys())

    def test_get_mapping_bundles(self):
        # Calling get_mapping with BUNDLE returns bundles.
        client = self.index_client
        client.put_mapping()
        mapping = client.get_mapping(BUNDLE)
        self.assertEqual([BUNDLE], mapping.keys())

    def test_get_mapping_all(self):
        # Calling get_mapping with None returns full mapping.
        client = self.index_client
        client.put_mapping()
        mapping = client.get_mapping(None)
        self.assertEqual(['temp_index'], mapping.keys())
        self.assertEqual(
            [BUNDLE, CHARM],
            sorted(mapping['temp_index'].keys()))

    def test_get_items(self):
        # The method returns the typed documents that match the typed_ids.
        charm_data = {'name': 'charm', 'owner': 'owner1', 'series': 'series'}
        self.index_client.index_charm(charm_data)
        bundle_data = {
            'name': BUNDLE, 'owner': 'owner2', 'basket_name': 'basket'}
        self.index_client.index_bundle(bundle_data)
        typed_ids = [
            {'_id': '~owner1/series/charm', '_type': CHARM},
            {'_id': '~owner2/basket/bundle', '_type': BUNDLE}
        ]
        docs = self.index_client.get_items(typed_ids)
        expected = [
            {'data': charm_data, 'doctype': CHARM},
            {'data': bundle_data, 'doctype': BUNDLE}
        ]
        self.assertEqual(expected, docs)

    def test_get_items_with_empty_list(self):
        # When test_ids is an empty list, the list of docs is also empty.
        self.assertEqual([], self.index_client.get_items([]))

    def test_related_charms(self):
        client = self.index_client
        charm = factory.get_charm_json(provides={'abc': {'interface': 'def'}})
        client.index_charm(charm)
        requires, provides = client.related_charms(
            set(charm['i_provides']), set(charm['i_requires']))
        self.assertEqual({}, provides)
        self.assertEqual({}, requires)
        requires, provides = client.related_charms(set(['def']), set(['def']))
        self.assertEqual(['def'], provides.keys())
        self.assertEqual(charm, provides['def'][0]['data'])
        self.assertIs(float, type(provides['def'][0]['weight']))
        self.assertEqual(2, len(provides['def'][0]))
        self.assertEqual({}, requires)
        charm2 = factory.get_charm_json(
            provides={
                'abc': {'interface': 'def'},
                'mno': {'interface': 'pqr'}},
            requires={'ghi': {'interface': 'jkl'}})
        client.index_charm(charm2)
        requires, provides = client.related_charms(set(['jkl']), set())
        self.assertEqual({}, provides)
        self.assertEqual(['jkl'], requires.keys())
        self.assertEqual(charm2, requires['jkl'][0]['data'])
        requires, provides = client.related_charms(set(['jkl']), set(['def']))
        self.assertEqual(['def'], provides.keys())
        self.assertItemsEqual([charm, charm2],
                              [payload['data'] for payload in provides['def']])
        self.assertEqual([charm2],
                         [payload['data'] for payload in requires['jkl']])
        requires, provides = client.related_charms(set(),
                                                   set(['def', 'pqr', 'xyz']))
        self.assertItemsEqual(['def', 'pqr'], provides.keys())
        self.assertEqual([charm2], [payload['data']
                         for payload in provides['pqr']])
        self.assertItemsEqual([charm, charm2],
                              [payload['data'] for payload in provides['def']])

    def test_related_charms_official_boost(self):
        """An official version of a charm rates 10x higher."""
        charm = factory.get_charm_json(
            provides={'foo': {'interface': 'telnet'}}, promulgated=True)
        official_id = charm['_id']
        unofficial_id = official_id + '-1'
        client = self.index_client
        client.index_charm(charm)
        charm['owner'] = 'jrandom'
        charm['_id'] = unofficial_id
        charm['store_url'] = get_address(charm, short=False)
        client.index_charm(charm)
        requires, provides = client.related_charms(set(), set(['telnet']))
        by_id = dict((result['data']['_id'], result['weight'])
                     for result in provides['telnet'])
        self.assertAlmostEqual(by_id[official_id], by_id[unofficial_id] * 10)

    def test_related_charms_series_filter(self):
        def filter_data(result):
            return dict((key, [entry['data'] for entry in value])
                        for key, value in result.items())
        alpha_charm = factory.get_charm_json(
            series='alpha', provides={'asdf': {'interface': 'foo'}})
        beta_charm = factory.get_charm_json(
            series='beta', requires={'asdf': {'interface': 'bar'}})
        self.index_client.index_charms([alpha_charm, beta_charm])
        requires, provides = self.index_client.related_charms({'bar'}, {'foo'})
        self.assertEqual({'bar': [beta_charm]}, filter_data(requires))
        self.assertEqual({'foo': [alpha_charm]}, filter_data(provides))
        requires, provides = self.index_client.related_charms(
            {'bar'}, {'foo'}, series='alpha')
        self.assertEqual({}, filter_data(requires))
        self.assertEqual({'foo': [alpha_charm]}, filter_data(provides))
        requires, provides = self.index_client.related_charms(
            {'bar'}, {'foo'}, series='beta')
        self.assertEqual({'bar': [beta_charm]}, filter_data(requires))
        self.assertEqual({}, filter_data(provides))
        requires, provides = self.index_client.related_charms(
            {'bar'}, {'foo'}, series='gamma')
        self.assertEqual({}, filter_data(requires))
        self.assertEqual({}, filter_data(provides))

    def test_related_charms_exclude_name(self):
        charm = factory.get_charm_json(
            requires={'asdf': {'interface': 'bar'}})
        self.index_client.index_charm(charm)
        requires, provides = self.index_client.related_charms({'bar'}, set())
        self.assertEqual(charm, requires['bar'][0]['data'])
        requires, provides = self.index_client.related_charms(
            {'bar'}, set(), exclude_name=charm['name'])
        self.assertEqual({}, requires)

    def test_special_characters_in_full_text_term(self):
        # Some characters have a special meaning for the text parser
        # used by ElasticSearch/Lucene. Since these characters can
        # cause exceptions when used unexpectedly, like unbalanced
        # parentheses, they are replaced with spaces.
        for char in range(0, 128):
            char = chr(char)
            try:
                text = u'foo %s bar' % char
                self.index_client.search(text)
            except ElasticHttpError:
                self.fail(
                    "Character '%s' inside a search term causes "
                    "ElasticHttpError" % char)
            try:
                text = 'foo%s' % char
                self.index_client.search(text)
            except ElasticHttpError:
                self.fail(
                    "Character '%s' at the end of a search term causes "
                    "ElasticHttpError" % char)

    def test_no_mapping_changes_during_indexing(self):
        # The mapping does not change when charm is indexed.
        before = self.index_client.get_mapping()
        charm = factory.get_charm_json()
        self.index_client.index_charm(charm)
        after = self.index_client.get_mapping()
        self.assertEqual(before, after)


class TestIndexingBundles(TestCase):

    def setUp(self):
        super(TestIndexingBundles, self).setUp()
        self.use_index_client()

    def exists_in_index(self, bundle_id):
        try:
            result = self.index_client._client.get(
                self.index_client.index_name, 'bundle', bundle_id)
        except ElasticHttpNotFoundError:
            return False
        return result['exists']

    def test_index_bundles(self):
        bundle_data = factory.get_bundle_data()
        search_id = Bundle(bundle_data).search_id
        self.assertFalse(self.exists_in_index(search_id))
        self.index_client.index_bundles([bundle_data])
        self.assertTrue(self.exists_in_index(search_id))

    def test_index_bundles_no_bundles(self):
        # Assert that no exception is raised.
        with self.assertRaises(AssertionError):
            with self.assertRaises(BaseException):
                self.index_client.index_bundles([])

    def test_only_one_revision_per_bundle(self):
        bundle_data = factory.get_bundle_data(
            owner='a', name='b', basket_with_rev='c/1')
        bundle_data_2 = factory.get_bundle_data(
            owner='a', name='b', basket_with_rev='c/2')
        self.index_client.index_bundle(bundle_data)
        self.index_client.index_bundle(bundle_data_2)
        self.assertEqual(['~a/c/2/b'], [b['data']['_id'] for b in
                         self.index_client.api_search(doctype=None)])


def put_mapping(client, properties, dynamic=True):
    client._client.put_mapping(
        client.index_name, CHARM, {
            CHARM: {
                'dynamic': dynamic,
                'properties': {'data': {'properties': properties}}
            }
        }
    )
    client.wait_for_green_status()


def put_incompatible_mapping(index_client):
    put_mapping(index_client, {
        'name': {'type': 'string', 'index': 'analyzed'},
    })


class TestUpdate(TestCase):

    def test_update_noop(self):
        temp_client = self.use_index_client()
        update(temp_client)

    def test_compatible_update(self):
        index_client = self.use_index_client(put_mapping=False)
        put_mapping(index_client, {
            'box': {'type': 'string', 'index': 'not_analyzed'},
        })
        mapping = index_client.get_mapping()[CHARM]
        self.assertNotIn('name', mapping['properties']['data']['properties'])
        update(index_client)
        mapping = index_client.get_mapping()[CHARM]['properties']['data']
        self.assertIn('box', mapping['properties'])
        self.assertIn('name', mapping['properties'])

    def test_incompatible_update(self):
        index_client = self.use_index_client(put_mapping=False)
        put_incompatible_mapping(index_client)
        update(index_client)
        mapping = index_client.get_mapping()[CHARM]['properties']['data']
        self.assertEqual('not_analyzed',
                         mapping['properties']['name']['index'])

    def test_update_no_index(self):
        index_client = ElasticSearchClient.from_settings(
            get_ini(), 'temp-index')
        index_client.stable_status_timeout = '0.5s'
        actual_client = update(index_client)
        self.addCleanup(actual_client.delete_index)
        self.assertEqual([actual_client.index_name],
                         index_client.get_aliased())
        mapping = index_client.get_mapping()[CHARM]['properties']['data']
        self.assertEqual('not_analyzed',
                         mapping['properties']['name']['index'])

    def update_to_static_mapping(self, force_reindex):
        index_client = self.use_index_client(put_mapping=False)
        put_mapping(
            index_client,
            {'box': {'type': 'string', 'index': 'not_analyzed'}},
            dynamic=True)
        update(index_client, force_reindex)
        updated_mapping = index_client.get_mapping()
        # A property 'files' is not defined in the current mapping.
        self.assertNotIn(
            'files',
            updated_mapping['charm']['properties']['data']['properties'])
        index_client.index_charm(factory.get_charm_json())
        return index_client

    @staticmethod
    def get_charm_props_from_mapping(index_client):
        mapping = index_client.get_mapping()
        return mapping['charm']['properties']['data']['properties']

    def test_simple_change_dynamic_to_static_mapping(self):
        # If an existing mapping is dynamic (the default for ElasticSearch)
        # and if a new mapping is specified as static, the two mappings
        # are considered compatible, but the resulting mapping is
        # still dynamic.
        index_client = self.update_to_static_mapping(force_reindex=False)
        # Avoid spurious failures. The new mapping is not immediately
        # available after the index_charm() call. Calling
        # index_client.wait_for_green_status() does not help here.
        retry = 0
        while retry < 10:
            properties = self.get_charm_props_from_mapping(index_client)
            if 'files' not in properties:
                sleep(0.1)
                retry += 1
            else:
                break
        # charm['files'] is not supposed to be included in the new
        # mapping, but it still exists if force_reindex is not used.
        self.assertIn('files', properties)

    def test_dynamic_to_static_mapping_forced_reindex(self):
        # If an existing mapping is dynamic (the defult) and if a
        # new mapping is specified as static, the two mappings
        # are considered compatible, but the resulting mapping is
        # still dynamic.
        index_client = self.update_to_static_mapping(force_reindex=True)
        updated_mapping = index_client.get_mapping()
        # The mapping is indeed static; charm['files'] is not indexed.
        self.assertNotIn(
            'files',
            updated_mapping['charm']['properties']['data']['properties'])


class TestReindex(TestCase):

    def test_reindex_aliased(self):
        index_client = ElasticSearchClient.from_settings(
            get_ini(), 'temp_aliased')
        alias = ElasticSearchClient(index_client._client, 'alias')
        index_client.create_index()
        try:
            alias.update_aliased(index_client.index_name, [])
        except:
            index_client.delete_index()
            raise
        self.addCleanup(alias.delete_index)
        put_incompatible_mapping(index_client)
        self.assertEqual(['temp_aliased'], alias.get_aliased())
        new_aliased = reindex(alias)
        self.assertEqual([new_aliased.index_name], alias.get_aliased())
        mapping = alias.get_mapping()[CHARM]['properties']['data']
        self.assertEqual('not_analyzed',
                         mapping['properties']['name']['index'])

    def test_reindexed_no_client_charms(self):
        client = ElasticSearchClient.from_settings(get_ini())
        # This should not raise an exception, even though the index does not
        # exist.
        new_client = reindex(client, charms=[])
        new_client.delete_index()
