# Taken and Modified from trac / GPL

__metaclass__ = type

import collections
from ConfigParser import ConfigParser
from contextlib import contextmanager
import copy
import datetime
from dateutil.parser import parse
import logging.config
from os import environ
from os import path
import time
from urllib import unquote
import urlparse
# Required for get_session_url on Precise, but not on Quantal.
urlparse.uses_query.append('mongodb')


_units = (
    (3600 * 24 * 365, '%d year'),
    (3600 * 24 * 30, '%d month'),
    (3600 * 24 * 7, '%d week'),
    (3600 * 24, '%d day'),
    (3600, '%d hour'),
    (60, '%d minute')
)


def prettify_timedelta(diff):
    """Return a prettified time delta, suitable for pretty printing."""
    age_s = int(diff.days * 86400 + diff.seconds)

    for u, template in _units:
        r = float(age_s) / float(u)
        if r >= 1.9:
            r = int(round(r))
            v = template % r
            if r > 1:
                v += "s"
            return v
    return ''


def pretty_timedelta(time1, time2=None):
    """Calculate time delta between two `datetime` objects.
    (the result is somewhat imprecise, only use for prettyprinting).
    """
    if isinstance(time1, (str, unicode)):
        time1 = parse(time1)
    if time2 is None:
        time2 = datetime.datetime.utcnow()

    if time1 > time2:
        raise ValueError('The start date is before the end date.')

    diff = time2 - time1
    return prettify_timedelta(diff)


def configure_logging():
    logging.config.fileConfig(get_ini_path())


def get_ini_path():
    """Determine the absolute path of the selected ini file.

    The default ini file is the 'charmworld.ini', however it can be overridden
    by providing a INI=test.ini ENV variable.  Filenames are interpreted
    relative to the charmworld tree root regardless of the current working
    directory.
    """
    ini_file = environ.get('INI', 'charmworld.ini')
    ini_path = path.join(path.dirname(path.dirname(__file__)), ini_file)
    return ini_path


def get_ini():
    """Fetch the ini file in use and provide a dict of values back.

    This is to mimic the access provided to the Pyramid app to outside code.
    It provides a dict from the ConfigParser instance in order to mimic access
    of request.registry.settings. It only provides the main app section of
    config data.

    The default ini file is the 'charmworld.ini', however it can be overridden
    by providing a INI=test.ini ENV variable.   See get_ini_path.

    """
    ini = ConfigParser()
    ini.readfp(open(get_ini_path()))
    here = path.abspath(path.join(path.dirname(__file__), '../../'))
    ini.set('app:main', 'here', here)
    return dict(ini.items("app:main"))


def get_session_url(settings):
    parsed = urlparse.urlparse(settings['mongo.url'])
    path = settings['mongo.database'] + '.sessions'
    return urlparse.urlunparse((parsed[0], parsed[1], path, parsed[3],
                                parsed[4], parsed[5]))


def quote_key(key):
    """Replace the symbols '.', '$', '%' with escape sequences.

    PyMongo version 2.4.2 for example does not allow dots in keys, and
    while older versions allowed to store dictionaries like {'a.b.c': 'd'},
    a query like collection.find({'a.b.c': 'd'}) would not find this
    record because the query searches for a record like
    {'a': {'b': {'c': 'd'}}}.

    Moreover, the '$' is reserved for query keywords so it should be
    avoided too.
    """
    for char, replacement in (('%', '%25'), ('$', '%24'), ('.', '%2E')):
        key = key.replace(char, replacement)
    return key


def _process_dict_keys(data, key_mapper):
    """Apply the key_mapper function to keys in the data dictionary.

    The dictionary `data` is modified in place.
    If the dictionary contains any dicts or lists, the function is called
    recursively on those inner structures with no limits.
    """
    if isinstance(data, collections.Mapping):
        tmp = {}
        for key, value in data.items():
            if isinstance(key, str) or isinstance(key, unicode):
                new_key = key_mapper(key)
                if new_key != key:
                    tmp[new_key] = value
                    del data[key]
            _process_dict_keys(value, key_mapper)
        data.update(tmp)
    elif (isinstance(data, collections.Sequence)
          and not isinstance(data, str) and not isinstance(data, unicode)):
        for item in data:
            _process_dict_keys(item, key_mapper)


def quote_yaml(orig_data):
    """Traverse the given object and its sub-objects; if the object
    is a dictionary, replace string keys containing at least one of
    '.', '$', '%' with values that can safely be stored in Mongo.
    """
    data = copy.deepcopy(orig_data)
    _process_dict_keys(data, quote_key)
    return data


def unquote_yaml(orig_data):
    """Unquote dictionary keys that were quoted by quote_yaml().
    """
    data = copy.deepcopy(orig_data)
    _process_dict_keys(data, unquote)
    return data


def timestamp(date):
    """Return an RFC 3339 timestamp representing the supplied date(time).

    datetimes are expected to be naive, and represent UTC time.
    """
    return date.isoformat() + 'Z'


class Timer:
    """Context manager for timing a block of code."""

    def __init__(self):
        self.start = None
        self.end = None

    def __enter__(self):
        self.start = time.time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end = time.time()

    def duration(self):
        return self.end - self.start


@contextmanager
def write_locked(lockable):
    lockable.lock_write()
    try:
        yield lockable
    finally:
        lockable.unlock()


@contextmanager
def read_locked(lockable):
    lockable.lock_read()
    try:
        yield lockable
    finally:
        lockable.unlock()
