from deployer.relation import (
    Endpoint,
    EndpointPair,
)
from charmworld.lib.cdeployer import (
    parse_deployer,
    CharmworldDeployment,
)
from charmworld.lib.proof import (
    CharmProof,
    get_list_of_shared_interfaces,
    ProofError,
    verify_points_exist_and_same_type,
)
from charmworld.models import (
    BundledCharmDescription,
    Charm,
    resolve_charm_from_description,
)
from charmworld.views.utils import json_response


def _build_proof_response(response_data):
    # Build up a root messages list of errors that can be used to help the
    # client go through them all easily.
    error_messages = []
    if response_data['errors']:
        for bundle in response_data['errors']:
            prefix = bundle['name']
            for service_name, service in bundle['services'].items():
                for err in service:
                    msg = '%s: %s' % (prefix, err['message'])
                    error_messages.append(msg)

            for relation_error in bundle['relations']:
                msg = '%s: %s' % (prefix, relation_error['message'])
                error_messages.append(msg)

    if error_messages:
        response_data['error_messages'].extend(error_messages)
    return response_data


def _proof_bundles_charms_exist(db, bundle_info, services, series=None):
    """Verify the charms the bundle requests to deploy exist in the db."""
    found_charms = {}

    # Verify we can find the charms at all.
    for name, charm in services.items():
        try:
            charm_description = BundledCharmDescription(
                'proof', charm, series)
            charm_found = resolve_charm_from_description(
                db, charm_description)
            if not charm_found:
                raise ProofError(
                    charm_description.get_process(),
                    'Could not find charm: ' + name,
                )
            else:
                found_charms[name] = Charm(charm_found)
        except ProofError, exc:
            # We could not find the charm. No other verification can
            # happen.
            bundle_info['services'][name] = []
            bundle_info['services'][name].append({
                'message': exc.msg,
                'debug_info': exc.debug_info,
            })
    return found_charms, bundle_info


def _proof_bundles_relations(bundle_info, bundle_config, charms):
    """Given a set of relations, check they're valid for the charms found.

    - Verify that the charms specifed in the relation are valid for the
      bundle.
    - Verify that the interfaces between are compatible
    - Verify that if no specific interfaces is defined, we can find at least
      one common interface between the two services.

    """
    deployment = CharmworldDeployment(bundle_config)
    relations = deployment.get_relations()

    if not relations:
        return bundle_info

    # Begin the actual interface checks.
    for e_a, e_b in relations:
        # Use the deployer's idea of Endpoints to help resolve things down to
        # comparable objects.
        endpoints = [Endpoint(e_a), Endpoint(e_b)]

        # Check that the service for both endpoints exist.
        found = True
        for ep in endpoints:
            if not ep.service in charms:
                bundle_info['relations'].append({
                    'message': ('Invalid relation to charm: %s not found.'
                                % ep.service),
                    'debug_info': []
                })
                found = False

        if found:
            # Then both services in the endpoint were indeed found and we want
            # to verify that the interfaces exist and are good to go.
            requires = charms[endpoints[0].service].requires
            provides = charms[endpoints[1].service].provides

            check_requires = endpoints[0].name
            check_provides = endpoints[1].name

            if check_requires and check_provides:
                # Both names are provided. They must exist in each service and
                # be of the same interface type.
                try:
                    verify_points_exist_and_same_type(
                        check_requires, check_provides, requires, provides)
                except ProofError, exc:
                    bundle_info['relations'].append({
                        'message': exc.msg,
                        'debug_info': exc.debug_info,
                    })

            elif check_requires or check_provides:
                common_interfaces = get_list_of_shared_interfaces(
                    charms[endpoints[0].service],
                    charms[endpoints[1].service]
                )

                # We need to verify that the requested interface exists on the
                # service and that we have something compatible between them.
                found = False
                common = None
                interface_name = None
                if check_requires and check_requires in requires:
                    found = True
                    common = requires[check_requires]['interface']
                    interface_name = check_requires

                elif check_provides and check_provides in provides:
                    found = True
                    common = provides[check_provides]['interface']
                    interface_name = check_provides

                if not found or common not in common_interfaces:
                    if not found and not common:
                        msg = "Invalid relation requested: {}:{}".format(
                            check_requires if check_requires else "",
                            check_provides if check_provides else ""
                        )
                    else:
                        msg = "Invalid relation on interface: {}".format(
                            interface_name
                        )

                    bundle_info['relations'].append({
                        'message': msg,
                        'debug_info': [repr(EndpointPair(e_a, e_b))]
                    })

            else:
                # No interfaces were specified, just the two services. Make
                # sure we can find at least one common interface between them.
                common_interfaces = get_list_of_shared_interfaces(
                    charms[endpoints[0].service],
                    charms[endpoints[1].service]
                )

                if len(common_interfaces) == 0:
                    msg = "The two services %s:%s share no common interfaces."
                    bundle_info['relations'].append({
                        'message': msg % (
                            endpoints[0].service,
                            endpoints[1].service),
                        'debug_info': [repr(EndpointPair(e_a, e_b))]
                    })

    return bundle_info


def _proof_charm_config(bundle_info, bundle_config, charm_name, charm):
    """Given a charm, proof the config is valid for the charm found.

    If the deployer attempts to set illegal or non-existant config values to
    the charm treat these as proof errors.
    """
    check_config = bundle_config['services'][charm_name].get('options')
    if not check_config:
        # There's no config specified to validate.
        return bundle_info
    for test_key, test_value in check_config.iteritems():
        try:
            CharmProof.check_config(charm, test_key, test_value)
        except ProofError, exc:
            if charm_name not in bundle_info['services']:
                bundle_info['services'][charm_name] = []

            bundle_info['services'][charm_name].append({
                'message': exc.msg,
                'debug_info': exc.debug_info,
            })
    return bundle_info


def proof_deployer(request):
    """Check the deployer for proof errors."""
    # The top level is for general deployer file errors.
    # The errors here may be an error about the deployer file itself, or an
    # object that describes the issues found with a bundle within the deployer
    # file. errro_messages is added as a simple aggregate of all of the error
    # messages found to make life easier for the proof cli tool.
    response_data = {
        'errors': [],
        'error_messages': [],
        'debug_info': [],
    }
    deployer_string = request.params.get('deployer_file')
    deployer_format = request.params.get('deployer_format', 'yaml')

    # Verify we have a valid deployer file that we can find and parse.
    if not deployer_string:
        response_data['error_messages'].append(
            'No deployer file data received.')
        return json_response(400, response_data)

    try:
        deployer_data = parse_deployer(deployer_string, format=deployer_format)
    except ProofError, exc:
        # If we cannot parse the config file there's nothing more to do.
        # Return immediately.
        response_data['error_messages'].append(exc.msg)
        response_data['debug_info'].append(exc.debug_info)
        return json_response(400, response_data)

    # Proof each bundle found in the deployer file.
    for bundle_name, bundle_config in deployer_data.items():
        # Track all of the issues found in this info dict.
        bundle_info = {
            'name': bundle_name,
            'services': {},
            'relations': [],
        }

        # First, verify that we can locate the charms specified in this bundle
        # in the charmworld database.
        found_charms, bundle_info = _proof_bundles_charms_exist(
            request.db,
            bundle_info,
            bundle_config.get('services'),
            bundle_config.get('series'))

        # For the charms we did find in the system, verify that their
        # config is valid for the values allowed by the charm we found.
        for name, charm in found_charms.iteritems():
            bundle_info = _proof_charm_config(
                bundle_info, bundle_config, name, charm)

        # Verify that the relations specified in the bundle are valid for the
        # interfaces exposed on the charms that were found.
        bundle_info = _proof_bundles_relations(
            bundle_info, bundle_config, found_charms)

        # If there are errors in this bundles services or relation, make sure
        # we append the info to the response data that gets sent back to the
        # user.
        if bundle_info['services'].keys() or bundle_info['relations']:
            response_data['errors'].append(bundle_info)

    # After proofing each deployer file return with any failure info
    # found.
    return json_response(200, _build_proof_response(response_data))
