# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import os.path

from charmworld import cached_view_config
from charmworld.models import Bundle
from charmworld.views.api import json_response
from charmworld.views.helpers import (
    find_bundle,
    found,
    format_change,
)


ICON_FN = 'icon.svg'

# Note that RST is allowed but currently not rendered richly.  It will be
# displayed as text.
SUPPORTED_README_EXTS = ('', '.txt', '.md', '.markdown', '.mkd', '.rst')


class BundleDetail(Bundle):

    service_keys = (
        'branch',
        'charm',
        'constraints',
        'num_units',
        'options',
    )

    def __init__(self, bundle_data):
        super(BundleDetail, self).__init__(bundle_data)
        data = self._augmented = {}
        data['services_list'] = svclist = []

        # Flesh out the services and provide a default if there is none.
        services = self.data['services'] or {}
        for k, v in services.items():
            d = dict(name=k)
            for key in BundleDetail.service_keys:
                if key != 'options':
                    d[key] = v.get(key, u'')

            # Flatten options.
            options = v.get('options', {})
            opt_list = []
            for opt_key, opt_val in options.items():
                opt_list.append(
                    dict(name=opt_key,
                         value=opt_val))
            d['options'] = opt_list
            svclist.append(d)

        data['relations_list'] = rellist = []

        # Convert relations to a dict with 'left' and 'right' elements,
        # ensuring the right is a list.
        relations = self.data['relations'] or []
        for rel in relations:
            right = rel[1]
            if isinstance(right, basestring):
                right = [right]
            d = dict(left=rel[0],
                     right=', '.join(right))
            rellist.append(d)

    def __getattr__(self, name):
        try:
            return super(BundleDetail, self).__getattr__(name)
        except AttributeError:
            if name in self._augmented:
                return self._augmented[name]
            else:
                raise

    def icon_path(self, request):
        api_prefix = request.route_path('api_3', endpoint='bundle')
        path = '/'.join([api_prefix, self.id, 'file', ICON_FN])
        return path

    def readme_info(self, db):
        readme = readme_ext = None
        for fn in sorted(self.get_file_list(db)):
            namepart, ext = os.path.splitext(fn)
            if namepart.lower() == 'readme' and ext in SUPPORTED_README_EXTS:
                readme_handle = self.get_file(db, fn)
                readme = unicode(readme_handle.read(), 'utf8')
                readme_ext = ext[1:]
            break
        return readme, readme_ext


def _bundle_view(request, bundle):
    bundle = BundleDetail(dict(bundle))
    try:
        is_owner = bool((request.user is not None) and
                        (request.user.nickname == bundle.owner))
    except AttributeError:
        is_owner = False
    readme, readme_ext = bundle.readme_info(request.db)
    return dict(
        bundle=bundle,
        format_change=format_change,
        icon_path=bundle.icon_path(request),
        is_owner=is_owner,
        readme=readme,
        readme_ext=readme_ext,
    )


@cached_view_config(
    route_name="personal-bundle-revision",
    renderer="charmworld:templates/bundle.pt")
@cached_view_config(
    route_name="personal-bundle",
    renderer="charmworld:templates/bundle.pt")
def personal_bundle(request):
    bundle = find_bundle(request)
    return _bundle_view(request, found(bundle))


@cached_view_config(
    route_name="personal-bundle-json-revision")
@cached_view_config(
    route_name="personal-bundle-json")
def personal_bundle_json(request):
    bundle = find_bundle(request)
    if bundle is None:
        status = 404
        id_ = Bundle.construct_id(
            request.matchdict['owner'],
            request.matchdict['basket'],
            request.matchdict['bundle'],
            request.matchdict.get('rev'),
        )
        value = {'type': 'no_such_bundle', 'bundle_id': id_}
    else:
        status = 200
        value = {bundle.name: bundle.data}
    return json_response(status, value)


@cached_view_config(
    route_name="official-bundle-revision",
    renderer="charmworld:templates/bundle.pt")
def official_bundle(request):
    bundle = find_bundle(request, promulgated=True)
    return _bundle_view(request, found(bundle))


@cached_view_config(
    route_name="official-bundle-json")
def official_bundle_json(request):
    bundle = find_bundle(request, promulgated=True)
    if bundle is None:
        status = 404
        path = '/'.join([request.matchdict['basket'],
                         request.matchdict['bundle']])
        value = {'type': 'no_such_bundle', 'bundle_id': path}
    else:
        status = 200
        value = {bundle.name: bundle.data}
    return json_response(status, value)
