# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from datetime import datetime
from pyramid.exceptions import NotFound
import pymongo

from charmworld.models import (
    Bundle,
    Charm,
)
from charmworld.models import find_charms

# View helpers used throughout various views


def format_change(change, branch_spec=None):
    """Format a bzr commit."""
    parts = []
    date = change["created"]
    # bug 1199790: We may have a branch without any commits.
    # In this case, Charm and Bundle return dummy data for its attributes
    # first_change and last_change.
    if date is not None:
        date = datetime.utcfromtimestamp(date)
        parts.append(unicode(date.strftime("%Y/%m/%d")))
    parts.append(name_filter(change['committer']))
    parts.append(change["message"][:100])
    if branch_spec is not None:
        parts.append(
            u'(<a href="'
            'https://bazaar.launchpad.net/%s/revision/%s">revno %s</a>)' % (
                branch_spec, change["revno"], change["revno"]))
    return u" ".join(parts)


def format_proof(charm):
    proof = charm.proof
    messages = []
    for e in proof.get('c', []):
        messages.append('crit: %s' % e)
    for e in proof.get('e', []):
        messages.append("err: %s" % e)
    for e in proof.get('w', []):
        messages.append('warn: %s' % e)
    return messages


def found(res):
    if res is None:
        raise NotFound()
    return res


def interfaces(db):
    interfaces = {}

    # This is all static between loads
    results = find_charms(
        db, {}, {"name": 1, "i_provides": 1, "i_requires": 1,
                 'label': 1, 'short_url': 1, 'owner': 1, 'store_url': True,
                 'promulgated': 1})

    for charm in results:
        charm = Charm(charm)
        for p in charm.i_provides:
            if not p in interfaces:
                interfaces[p] = i_info = {
                    "name": p, "provided_by": [], "required_by": []}
            else:
                i_info = interfaces[p]
            i_info["provided_by"].append(charm)

        for d in charm.i_requires:
            if not d in interfaces:
                interfaces[d] = i_info = {
                    "name": d, "provided_by": [], "required_by": []}
            else:
                i_info = interfaces[d]
            i_info["required_by"].append(charm)

    # sort them
    interfaces = interfaces.values()
    interfaces.sort(lambda x, y: cmp(x["name"], y["name"]))

    for i in interfaces:
        i["provided_by"] = unique_sort(i["provided_by"])
        i["required_by"] = unique_sort(i["required_by"])

    return interfaces


def name_filter(n):
    if isinstance(n, list):
        n = n[0]
    idx = n.find('<')
    if idx == -1:
        return n
    return n.split("<", 1)[0].strip()


def sub_filter(d, req):
    if req.GET.get('sub'):
        d['subordinate'] = True
        return d, True
    return d, False


def unique_sort(charms):
    names = []
    results = []
    for c in charms:
        if c.name in names:
            if not c.promulgated:
                continue
            else:
                idx = names.index(c.name)
                results.pop(idx)
                names.pop(idx)
        results.append(c)
        names.append(c.name)
    results.sort(lambda x, y: cmp(x.name, y.name))
    return results


def find_charm(request, promulgated=False):
    spec = {
        "name": request.matchdict['charm'],
        "series": request.matchdict['series'],
    }
    if promulgated:
        spec["promulgated"] = True
    else:
        spec["owner"] = request.matchdict['owner']
    charm_data = found(request.db.charms.find_one(
        spec, sort=[('store_data.revision', pymongo.DESCENDING)]))
    # We are phasing out is_featured from the charm data, so remove it if this
    # is an old charm record that still has it.
    charm_data.pop('is_featured', None)
    return charm_data


def find_bundle(request, promulgated=False):
    """Find the bundle.

    May return None, which must be handled by the caller.
    """
    spec = dict(
        name=request.matchdict['bundle'],
    )
    basket = request.matchdict.get('basket')
    rev = request.matchdict.get('rev')
    if basket is not None:
        spec['basket_name'] = basket
    if rev is not None:
        spec['basket_revision'] = int(rev)
    if promulgated:
        spec["promulgated"] = True
    else:
        spec["owner"] = request.matchdict['owner']
    return Bundle.from_query(spec, request.db)
