# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from StringIO import StringIO

from charmworld.models import Bundle
from charmworld.testing import (
    factory,
    ViewTestBase,
    WebTestBase,
)
from charmworld.views.bundles import (
    BundleDetail,
    find_bundle,
    official_bundle_json,
    personal_bundle,
    personal_bundle_json,
)


class TestBundleDetail(ViewTestBase):

    @staticmethod
    def make_service_list(name):
        keys = sorted(BundleDetail.service_keys)
        svcs = {name: dict(zip(keys, range(len(keys))))}
        svcs[name]['options'] = {}
        return svcs

    def test_empty_lists(self):
        data = factory.get_bundle_data()
        bd = BundleDetail(data)
        self.assertEqual({}, bd.data['services'])
        self.assertEqual({}, bd.data['relations'])

    def test_flattened_services(self):
        services = self.make_service_list('svc1')
        data = factory.get_bundle_data(services=services)
        bd = BundleDetail(data)
        self.assertEqual(services, bd.data['services'])
        expected = [dict(name='svc1',
                         branch=0,
                         charm=1,
                         constraints=2,
                         num_units=3,
                         options=[])]
        self.assertEqual(expected, bd.services_list)

    def test_flattened_services_with_options(self):
        services = self.make_service_list('svc1')
        services['svc1']['options'] = dict(a=1, b=2)
        data = factory.get_bundle_data(services=services)

        bd = BundleDetail(data)
        self.assertEqual(services, bd.data['services'])
        expected = [dict(name='svc1',
                         branch=0,
                         charm=1,
                         constraints=2,
                         num_units=3,
                         options=[
                             dict(name='a', value=1),
                             dict(name='b', value=2),
                         ])
                    ]
        self.assertEqual(expected, bd.services_list)

    def test_flattened_relations(self):
        rels = [['a', 'b'],
                ['c', ['d', 'e']]]
        data = factory.get_bundle_data(relations=rels)

        bd = BundleDetail(data)
        self.assertEqual(rels, bd.data['relations'])

        expected = [dict(left='a', right='b'),
                    dict(left='c', right='d, e'),
                    ]
        self.assertEqual(expected, bd.relations_list)

    def test_icon_path(self):
        self.enable_routes()
        data = factory.get_bundle_data(
            name='sol',
            owner='mary',
            basket_with_rev='camerones/1')
        request = self.getRequest()
        bundle = BundleDetail(data)
        bundle.get_file = lambda s, x: True
        expected = '/'.join(['/api/3/bundle', bundle.id, 'file', 'icon.svg'])
        self.assertEqual(expected, bundle.icon_path(request))

    def test_readme(self):
        self.enable_routes()
        data = factory.get_bundle_data()
        bundle = BundleDetail(data)
        bundle.get_file_list = lambda s: ['cows', 'README.txt']
        fake_contents = 'README, yo'
        bundle.get_file = lambda s, x: StringIO(fake_contents)
        expected = (fake_contents, 'txt')
        self.assertEqual(expected, bundle.readme_info('ignored'))

    def test_readme_missing(self):
        self.enable_routes()
        data = factory.get_bundle_data()
        bundle = BundleDetail(data)
        bundle.get_file_list = lambda s: ['cows']
        expected = (None, None)
        self.assertEqual(expected, bundle.readme_info('ignored'))

    def test_readme_unsupported_markup(self):
        self.enable_routes()
        data = factory.get_bundle_data()
        bundle = BundleDetail(data)
        bundle.get_file_list = lambda s: ['cows', 'README.novel_markup']
        expected = (None, None)
        self.assertEqual(expected, bundle.readme_info('ignored'))

    def test_readme_supported_markups(self):
        self.enable_routes()
        data = factory.get_bundle_data()
        bundle = BundleDetail(data)
        fake_contents = 'README, yo'
        bundle.get_file = lambda s, x: StringIO(fake_contents)
        for ext in ('txt', 'rst', 'md', 'markdown', 'mkd', ''):
            if ext == '':
                target = 'ReadMe'
            else:
                target = '.'.join(['ReadMe', ext])
            bundle.get_file_list = lambda s: [target]
            expected = (fake_contents, ext)
            self.assertEqual(expected, bundle.readme_info('ignored'))


class TestBundles(ViewTestBase):

    def test_found_bundle(self):
        bundle, bundle_data = factory.makeBundle(self.db)
        request = self.getRequest()
        request.matchdict = dict(bundle=bundle.name,
                                 basket=bundle.basket_name,
                                 owner=bundle.owner)
        found = find_bundle(request)
        self.assertEqual(bundle, found)

    def test_multiple_bundles_picks_newest(self):
        basket_rev1 = 'riak/1'
        name = 'xyz'
        owner = 'matt'
        bundle1, bundle1_data = factory.makeBundle(
            self.db, name=name, owner=owner, basket_with_rev=basket_rev1)
        basket_rev2 = 'riak/10'
        bundle2, bundle2_data = factory.makeBundle(
            self.db, name=name, owner=owner, basket_with_rev=basket_rev2)
        request = self.getRequest()
        request.matchdict = dict(bundle=name,
                                 basket='riak',
                                 owner=owner)
        found = find_bundle(request)
        self.assertEqual(bundle2, found)

    def test_found_bundle_promulgated(self):
        bundle, bundle_data = factory.makeBundle(self.db, promulgated=True)
        request = self.getRequest()
        request.matchdict = dict(bundle=bundle.name)
        found = find_bundle(request, promulgated=True)
        self.assertEqual(bundle, found)

    def test_not_found_bundle(self):
        bundle, bundle_data = factory.makeBundle(self.db)
        request = self.getRequest()
        request.matchdict = dict(bundle='nobody-knows',
                                 owner=bundle.owner)
        found = find_bundle(request)
        self.assertEqual(None, found)

    def test_personal_bundle(self):
        self.enable_routes()
        basket = 'ball'
        rev = '4'
        bundle, bundle_data = factory.makeBundle(
            self.db, name='gigis', owner='ladonna',
            basket_with_rev='/'.join([basket, rev]),
            series='raring', promulgated=False,
        )
        request = self.getRequest()
        request.matchdict = dict(
            owner=bundle.owner,
            basket=basket,
            bundle=bundle.name,
        )
        expected = dict(
            series='raring',
            relations={},
            services={},
        )
        response = personal_bundle(request)
        self.assertEqual(expected, response['bundle'].data)

    def test_personal_bundle_json(self):
        basket = 'ball'
        rev = '4'
        bundle, bundle_data = factory.makeBundle(
            self.db, name='gigis', owner='ladonna',
            basket_with_rev='/'.join([basket, rev]),
            series='raring', promulgated=False,
        )
        request = self.getRequest()
        request.matchdict = dict(
            owner=bundle.owner,
            basket=basket,
            rev=rev,
            bundle=bundle.name,
        )
        expected = {'gigis': dict(
            series='raring',
            relations={},
            services={},
        )}
        response = personal_bundle_json(request)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(expected, response.json_body)

    def test_personal_bundle_json_not_found(self):
        owner = 'no-one'
        basket = 'no-basket'
        rev = '7'
        bundle = 'no-name'
        request = self.getRequest()
        request.matchdict = dict(
            owner=owner,
            basket=basket,
            rev=rev,
            bundle=bundle,
        )
        response = personal_bundle_json(request)
        self.assertEqual(response.status_code, 404)
        id_ = Bundle.construct_id(owner, basket, bundle, rev)
        expected = {'type': 'no_such_bundle', 'bundle_id': id_}
        self.assertEqual(expected, response.json_body)

    def test_official_bundle_json(self):
        bundle, bundle_data = factory.makeBundle(
            self.db, name='lantern', owner='andrea', basket_with_rev='eggs/1',
            series='raring', promulgated=True,
        )
        request = self.getRequest()
        request.matchdict = dict(
            basket='eggs',
            bundle='lantern',
        )
        expected = {'lantern': dict(
            series='raring',
            relations={},
            services={},
        )}
        response = official_bundle_json(request)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(expected, response.json_body)

    def test_official_bundle_json_not_found(self):
        basket = 'no-basket'
        bundle = 'no-name'
        request = self.getRequest()
        request.matchdict = dict(
            basket=basket,
            bundle=bundle,
        )
        response = official_bundle_json(request)
        self.assertEqual(response.status_code, 404)
        path = '/'.join([basket, bundle])
        expected = {'type': 'no_such_bundle', 'bundle_id': path}
        self.assertEqual(expected, response.json_body)


class BundleRoutingTests(WebTestBase):
    """Ensure the routing works."""

    def test_personal_bundle_route(self):
        owner = 'ladonna'
        basket = 'treme'
        bundle_name = 'gigis'
        rev = 1
        basket_with_rev = '%s/%d' % (basket, rev)
        bundle, bundle_data = factory.makeBundle(
            self.db, name=bundle_name, owner=owner,
            basket_with_rev=basket_with_rev, series='raring',
        )
        # Test that the route works.
        path = '/~{owner}/bundle/{basket}/{bundle}'.format(
            owner=owner, basket=basket, bundle=bundle_name)
        resp = self.app.get(path, status=200)
        self.assertIn(bundle.owner, resp.body)
        self.assertIn(bundle.data['series'], resp.body)
        self.assertIn(bundle.basket_name, resp.body)
        self.assertIn(bundle.owner, resp.body)
        # Ensure the short url to the branch on Launchpad is in the body.
        lp_path = '~{owner}/charms/bundles/{basket}/bundle'.format(
            owner=owner, basket=basket)
        lp_short_url = 'lp:{path}'.format(path=lp_path)
        self.assertIn(lp_short_url, resp.body)
        # Ensure the long url to the branch on Launchpad is in the body.
        lp_long_url = 'https://code.launchpad.net/{path}'.format(path=lp_path)
        self.assertIn(lp_long_url, resp.body)

    def test_personal_bundle_route_with_revision(self):
        bundle, bundle_data = factory.makeBundle(
            self.db, name='gigis', owner='ladonna',
            basket_with_rev='basket/1', series='raring',
        )
        path = '/~ladonna/bundle/basket/1/gigis'
        response = self.app.get(path, status=200)
        self.assertIn(
            'lp:~ladonna/charms/bundles/basket/bundle', response.body)

    def test_personal_bundle_route_404(self):
        bundle, bundle_data = factory.makeBundle(
            self.db, name='gigis', owner='ladonna',
            basket_with_rev='basket/1', series='raring',
        )
        path = '/~ladonna/bundle/basket/1/XXX'
        self.app.get(path, status=404)

    def test_personal_bundle_with_revision_json_route(self):
        # A JSON-encoded bundle can be requested in which the ID includes both
        # an owner and a revision.
        bundle, bundle_data = factory.makeBundle(
            self.db, name='gigis', owner='ladonna',
            basket_with_rev='basket/1', series='raring',
        )
        path = '/~ladonna/bundle/basket/1/gigis/json'
        resp = self.app.get(path, status=200)
        self.assertIn(bundle.data['series'], resp.body)
        self.assertIn(str(bundle.data['relations']), resp.body)
        self.assertIn(str(bundle.data['services']), resp.body)

    def test_personal_bundle_without_revision_json_route(self):
        # A JSON-encoded bundle can be requested in which the ID includes an
        # owner but no revision.
        bundle, bundle_data = factory.makeBundle(
            self.db, name='gigis', owner='lafawnda',
            basket_with_rev='basket/1', series='raring',
        )
        path = '/~lafawnda/bundle/basket/gigis/json'
        resp = self.app.get(path, status=200)
        self.assertIn(bundle.data['series'], resp.body)
        self.assertIn(str(bundle.data['relations']), resp.body)
        self.assertIn(str(bundle.data['services']), resp.body)

    def test_personal_bundle_json_route_404(self):
        bundle, bundle_data = factory.makeBundle(
            self.db, name='gigis', owner='ladonna',
            basket_with_rev='basket/1', series='raring',
        )
        path = '/~ladonna/bundle/basket/1/XXX/json'
        self.app.get(path, status=404)

    def test_official_bundle_route(self):
        bundle, bundle_data = factory.makeBundle(
            self.db, name='wiki', owner='charmers',
            basket_with_rev='wiki-basket/1', series='raring', promulgated=True,
        )
        path = '/bundle/wiki-basket/1/wiki/json'
        resp = self.app.get(path, status=200)
        self.assertIn(bundle.data['series'], resp.body)
        self.assertIn(str(bundle.data['relations']), resp.body)
        self.assertIn(str(bundle.data['services']), resp.body)

    def test_official_bundle_with_revision_route(self):
        bundle, bundle_data = factory.makeBundle(
            self.db, name='wiki', owner='charmers',
            basket_with_rev='wiki-basket/1', series='raring', promulgated=True,
        )
        path = '/bundle/wiki-basket/1/wiki'
        response = self.app.get(path, status=200)
        self.assertIn(
            'lp:~charmers/charms/bundles/wiki-basket/bundle', response.body)

    def test_official_bundle_route_404(self):
        bundle, bundle_data = factory.makeBundle(
            self.db, name='gigis', owner='ladonna',
            basket_with_rev='basket/1', series='raring', promulgated=True,
        )
        path = '/bundle/basket/1/XXX/json'
        self.app.get(path, status=404)
