# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from charmworld.models import (
    CharmSource,
    FeaturedSource,
)
from charmworld.testing import (
    factory,
    login_request,
    ViewTestBase,
    WebTestBase,
)
from charmworld.views.featured import (
    bundle_featured_edit,
    charm_featured_edit,
    featured_edit,
)


class TestFeaturedForm(ViewTestBase):

    def test_form_renders(self):
        """Sanity check that the form renders without errors."""
        cid, charm = factory.makeCharm(self.db, promulgated=True, owner='foo')
        request = self.getRequest()
        request.matchdict = {
            'charm': charm['name'],
            'series': charm['series'],
        }
        response = featured_edit(request, charm, 'charm', '')

        self.assertIn('form', response)
        self.assertIn('<label', response['form'])

    def test_form_includes_current_featured_value(self):
        """GET requests result in the current featured value being returned."""
        cid, charm = factory.makeCharm(self.db, promulgated=True, owner='foo')
        request = self.getRequest()
        request.matchdict = {
            'charm': charm['name'],
            'series': charm['series'],
        }
        response = featured_edit(request, charm, 'charm', '')
        self.assertFalse(response['values']['is_featured'])

    def test_checking_checkbox_sets_featured_status(self):
        """Verify checking the checkbox causes the charm to be featured."""
        charm = factory.get_charm_json(name='name1', promulgated=True)
        self.use_index_client()
        self.enable_routes()
        charm_source = CharmSource.from_request(self)
        charm_source.save(charm)
        # Add the submit button to trigger the right path through the view.
        post_data = {
            '_charset_': 'UTF-8',
            '__formid__': 'deform',
            'submit': 'submit',
            'is_featured': 'true',
        }
        request = self.getRequest(post=post_data)
        request.matchdict = {
            'charm': charm['name'],
            'series': charm['series'],
        }
        # Before processing the request, the charm is not featured.
        featured_source = FeaturedSource.from_db(self.db)
        self.assertFalse(featured_source.is_featured(charm, 'charm'))
        featured_edit(request, charm, 'charm', '')
        for found in charm_source._get_all(charm['_id']):
            self.assertTrue(featured_source.is_featured(found, 'charm'))

    def test_unchecking_checkbox_clears_featured_status(self):
        """Verify unchecking the checkbox causes the charm to be unfeatured."""
        charm = factory.get_charm_json(name='name1', promulgated=True)
        self.use_index_client()
        self.enable_routes()
        charm_source = CharmSource.from_request(self)
        charm_source.save(charm)
        # Add the submit button to trigger the right path through the view.
        post_data = {
            '_charset_': 'UTF-8',
            '__formid__': 'deform',
            'submit': 'submit',
        }
        request = self.getRequest(post=post_data)
        request.matchdict = {
            'charm': charm['name'],
            'series': charm['series'],
        }
        # Before processing the request, the charm is featured.
        featured_source = FeaturedSource.from_db(self.db)
        featured_source.set_featured(charm, 'charm')
        self.assertTrue(featured_source.is_featured(charm, 'charm'))
        featured_edit(request, charm, 'charm', '')
        for found in charm_source._get_all(charm['_id']):
            self.assertFalse(featured_source.is_featured(found, 'charm'))

    def test_redirect_on_submit(self):
        """After a submit, an HTTP redirect is raised with the given URL."""
        charm = factory.get_charm_json(name='name1', promulgated=True)
        self.use_index_client()
        self.enable_routes()
        charm_source = CharmSource.from_request(self)
        charm_source.save(charm)
        # Add the submit button to trigger the right path through the view.
        post_data = {
            '_charset_': 'UTF-8',
            '__formid__': 'deform',
            'submit': 'submit',
            'is_featured': 'true',
        }
        request = self.getRequest(post=post_data)
        request.matchdict = {
            'charm': charm['name'],
            'series': charm['series'],
        }
        # Before processing the request, the charm is not featured.
        featured_source = FeaturedSource.from_db(self.db)
        self.assertFalse(featured_source.is_featured(charm, 'charm'))
        response = featured_edit(request, charm, 'charm', 'URL')
        self.assertEqual(302, response.status_code)
        self.assertEqual(response.headers['Location'], 'URL')


class TestFeaturedCharmForm(ViewTestBase, WebTestBase):

    def setUp(self):
        super(TestFeaturedCharmForm, self).setUp()
        # Register the routes so URL construction will work in the tests.
        self.enable_routes()

    def test_form_renders(self):
        """Sanity check that the form renders without errors."""
        cid, charm = factory.makeCharm(self.db, promulgated=True, owner='foo')
        request = self.getRequest()
        request.matchdict = {
            'charm': charm['name'],
            'series': charm['series'],
        }
        response = charm_featured_edit(request)

        self.assertIn('form', response)
        self.assertIn('<label', response['form'])

    def test_charm_feature_form_route(self):
        """The charm feature form is wired up to a route."""
        charm = factory.makeCharm(self.db, promulgated=True, owner='foo')[1]
        url = '/charms/{series}/{name}/featured/edit'.format(**charm)
        with login_request(self.app, groups=['charmers']):
            # Get the URL and assert the given status code.
            self.app.get(url, status=200)

    def test_404_if_charm_is_not_promulgated(self):
        # If the charm indicated by the URL does not exist, a 404 is generated.
        charm = factory.makeCharm(self.db, promulgated=False, owner='foo')[1]
        url = '/charms/{series}/{name}/featured/edit'.format(**charm)
        with login_request(self.app, groups=['charmers']):
            # Get the URL and assert the given status code.
            self.app.get(url, status=404)


class TestFeaturedBundleForm(ViewTestBase, WebTestBase):

    def setUp(self):
        super(TestFeaturedBundleForm, self).setUp()
        # Register the routes so URL construction will work in the tests.
        self.enable_routes()

    def test_form_renders(self):
        """Sanity check that the form renders without errors."""
        bundle, bundle_data = factory.makeBundle(self.db, promulgated=True)
        request = self.getRequest()
        request.matchdict = {
            'bundle': bundle_data['name'],
            'basket': bundle_data['basket_name'],
        }
        response = bundle_featured_edit(request)

        self.assertIn('form', response)
        self.assertIn('<label', response['form'])

    def test_bundle_feature_form_route(self):
        """The bundle feature form is wired up to a route."""
        bundle, bundle_data = factory.makeBundle(self.db, promulgated=True)
        url = '/bundle/{basket_name}/{name}/featured/edit'.format(
            **bundle_data)
        with login_request(self.app, groups=['charmers']):
            # Get the URL and assert the given status code.
            self.app.get(url, status=200)

    def test_404_generated_when_bundle_is_not_promulgated(self):
        # If the bundle indicated by the URL does not exist, a 404 is
        # generated.
        bundle, bundle_data = factory.makeBundle(self.db, promulgated=False)
        url = '/bundle/{basket_name}/{name}/featured/edit'.format(
            **bundle_data)
        with login_request(self.app, groups=['charmers']):
            # Get the URL and assert the given status code.
            self.app.get(url, status=404)


class TestFeaturedFormAuthorization(WebTestBase):

    def test_form_403s_for_non_charmers(self):
        """The form is not permitted if you're not a charmer."""
        with login_request(self.app, groups=[]) as resp:
            cid, charm = factory.makeCharm(self.db)
            # The status force the check against the status code.
            resp = resp.goto(
                'charms/%s/%s/featured/edit' %
                (charm['series'], charm['name']), status=403)
            self.assertEqual(403, resp.status_code)

    def test_form_secured_from_users(self):
        """The form is only allowed to be edited/submitted by charmers."""
        with login_request(self.app, groups=['charmers']) as resp:
            cid, charm = factory.makeCharm(self.db, promulgated=True)
            resp = resp.goto(
                'charms/%s/%s/featured/edit' %
                (charm['series'], charm['name']),
                status=200)
            self.assertEqual(resp.status_code, 200)
