# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import os
import re
import unittest

from mock import patch

from paste.deploy import loadapp
from paste.deploy.config import PrefixMiddleware

from charmworld.models import FeaturedSource
from charmworld.testing import (
    factory,
    ViewTestBase,
    WebTestBase,
)
from charmworld.tests.test_health import TestCheckMixin
from charmworld.views.misc import heartbeat


class TestAppConfig(unittest.TestCase):

    def test_proxy_filter(self):
        ini_file = os.environ.get('INI', 'charmworld.ini')
        ini_path = os.path.normpath(os.path.join(
            os.path.dirname(__file__), '..', '..', '..', ini_file))
        wsgi_app = loadapp('config:%s' % ini_path)
        self.assertTrue(isinstance(wsgi_app, PrefixMiddleware))
        with patch.object(wsgi_app, 'app') as mock_method:
            environ = {
                'METHOD': 'GET',
                'SERVER_NAME': 'example.com',
                'SERVER_PORT': '80',
                'SERVER_PROTOCOL': 'HTTP/1.0',
                'PATH_INFO': '/',
                'HTTP_X_FORWARDED_PROTO': 'https'}
            wsgi_app(environ, None)
            passed_environ = mock_method.call_args[0][0]
        self.assertFalse('HTTP_X_FORWARDED_PROTO' in passed_environ)
        self.assertEqual('https', passed_environ['wsgi.url_scheme'])


class TestBaseViews(WebTestBase):

    def test_homeview(self):
        """Make sure the home view renders correctly."""
        resp = self.app.get(
            '/',
            status=200)

        body = resp.body
        self.assertIn("Juju Charms", body)

        # Make sure our cached_view_config is setting the Cache-control
        # correctly.

        for head, value in resp.headers.iteritems():
            if head == 'Cache-Control':
                self.assertEqual('max-age=450', value)
                break
        else:
            self.fail('Cache-Control header not found.')

    def test_main_template_scheme(self):
        """Image and stylesheet URLs respect the WSGI scheme."""
        src_pattern = re.compile('src="([a-z]+)://')
        css_pattern = re.compile('href="([a-z]+)://.*\.css.*"')
        # Https enabled.
        response = self.app.get(
            '/', extra_environ={'wsgi.url_scheme': 'https'})
        https_count = 0
        for match in src_pattern.finditer(response.ubody):
            https_count += 1
            self.assertEqual('https', match.group(1))
        for match in css_pattern.finditer(response.ubody):
            https_count += 1
            self.assertEqual('https', match.group(1))
        self.assertTrue(https_count >= 2)
        # Http default.
        response = self.app.get('/')
        http_count = 0
        for match in src_pattern.finditer(response.ubody):
            http_count += 1
            self.assertEqual('http', match.group(1))
        for match in css_pattern.finditer(response.ubody):
            http_count += 1
            self.assertEqual('http', match.group(1))
        self.assertTrue(https_count == http_count)


class HeartbeatWebTestCase(WebTestBase):

    def test_heartbeat_cache_control(self):
        # Never cache heartbeats.
        response = self.app.get('/heartbeat', status=200)
        body = response.body
        # No data was setup, so heartbeat knows the site is not operational.
        self.assertIn("Fail", body)
        self.assertIn(
            ('Cache-Control',
             'max-age=0, must-revalidate, no-cache, no-store, public'),
            response.headerlist)


class HeartbeatViewTestCase(ViewTestBase, TestCheckMixin):

    def test_all_checks_pass(self):
        # All checks pass when the services are available and data is returned.
        charm_data = factory.makeCharm(self.db, downloads=5)[1]
        factory.makeBundle(self.db)
        self.use_index_client()
        self.index_client.index_charm(charm_data)
        FeaturedSource.from_db(self.db).set_featured(charm_data, 'charm')

        def make_all_collection_names():
            def all_collection_names():
                return [
                    'fs.chunks', 'system.indexes', 'basket-queue', 'baskets',
                    'bundles', 'charm-queue', 'charms']
            return all_collection_names

        request = self.getRequest()
        with patch.object(request.db, 'collection_names',
                          new_callable=make_all_collection_names):
            response = heartbeat(request)

        checks = response['checks']
        self.assertEqual(7, len(checks))
        remark = '1 charms found'
        self.assertCheck(checks[0], 'charms collection', 'Pass', remark)
        remark = '1 bundles found'
        self.assertCheck(checks[1], 'bundles collection', 'Pass', remark)
        remark = '1 matches found'
        self.assertCheck(checks[2], 'charms index', 'Pass', remark)
        remark = 'Interesting has new, popular, and featured.'
        self.assertCheck(checks[3], 'API2 interesting', 'Pass', remark)
        self.assertCheck(checks[4], 'API3 interesting', 'Pass', remark)
        remark = 'All required collections exist.'
        self.assertCheck(checks[5], 'MongoDB collections', 'Pass', remark)
        remark = 'queued charms: 0, queued baskets: 0.'
        self.assertCheck(checks[6], 'Ingest queue sizes', 'Pass', remark)

    def test_checks_fail(self):
        # When services or data are not available, the checks fail.
        # No setup is performed, creating a case where services and data
        # are not available
        response = heartbeat(self.getRequest())
        checks = response['checks']
        self.assertEqual(7, len(checks))
        remark = 'There are no charms. Is ingest running?'
        self.assertCheck(checks[0], 'charms collection', 'Fail', remark)
        remark = 'There are no bundles. Is ingest running?'
        self.assertCheck(checks[1], 'bundles collection', 'Fail', remark)
        remark = 'The elasticsearch charm index is not available.'
        self.assertCheck(checks[2], 'charms index', 'Fail', remark)
        remark = 'The API2 is not available.'
        self.assertCheck(checks[3], 'API2 interesting', 'Fail', remark)
        remark = 'The API3 is not available.'
        self.assertCheck(checks[4], 'API3 interesting', 'Fail', remark)
        remark = (
            'Required collections do not (yet) exist: basket-queue, '
            'baskets, bundles, charm-queue, charms')
        self.assertCheck(checks[5], 'MongoDB collections', 'Fail', remark)
        remark = 'queued charms: 0, queued baskets: 0.'
        self.assertCheck(checks[6], 'Ingest queue sizes', 'Pass', remark)
