import yaml

from charmworld.models import Charm
from charmworld.testing import (
    factory,
    load_data_file,
)


class TestBundleProof:
    """Verify that a bundle can be proofed in charmworld."""

    def setup_relatable_charms(self, second_requires, second_provides):
        # The default charm provides website:https and requires
        # database:mongodb
        _id, charm_one = factory.makeCharm(
            self.db,
            description=''
        )

        _id, charm_two = factory.makeCharm(
            self.db,
            description='',
            requires=second_requires,
            provides=second_provides,
        )

        charm_one = Charm(charm_one)
        one_store_url = charm_one.store_url

        charm_two = Charm(charm_two)
        two_store_url = charm_two.store_url

        return one_store_url, two_store_url

    def test_bundle_proof_no_deployer_file(self):
        request = self.make_request('bundle', remainder='/proof')
        response = self.api_class(request)()
        self.assertEqual(400, response.status_code)
        self.assertEqual(
            'No deployer file data received.',
            response.json_body['error_messages'][0])

    def test_bundle_proof_invalid_deployer_file(self):
        request = self.make_request('bundle', remainder='/proof')
        request.params = {
            'deployer_file': '{'
        }
        response = self.api_class(request)()
        self.assertEqual(400, response.status_code)
        self.assertIn(
            'Could not parse',
            response.json_body['error_messages'][0])

    def test_bundle_proof_charm_exists(self):
        deployer_file = load_data_file('sample_bundle/bundles.yaml')

        request = self.make_request('bundle', remainder='/proof')
        request.params = {
            'deployer_file': deployer_file
        }

        response = self.api_class(request)()
        self.assertEqual(200, response.status_code)
        results = response.json_body
        self.assertEqual(1, len(results['errors']))
        self.assertEqual(4, len(results['errors'][0]['services'].keys()))
        self.assertIn(
            'Could not find ',
            results['errors'][0]['services']['db'][0]['message'])

        # And each failure has a list of 'process' information.
        self.assertIn(
            'debug_info',
            results['errors'][0]['services']['db'][0].keys())

    def test_bundle_proof_valid(self):
        # We have to create the charm so that we can verify it exists.
        _id, charm = factory.makeCharm(
            self.db,
            description=''
        )

        # and we'll cheat and just find it straight by the store url.
        charm = Charm(charm)
        store_url = charm.store_url

        bundle_config = {
            'wiki': {
                'services': {
                    'charm': {
                        'charm_url': str(store_url)
                    }
                }
            }
        }

        request = self.make_request('bundle', remainder='/proof')
        request.params = {
            'deployer_file': yaml.dump(bundle_config)
        }
        response = self.api_class(request)()
        self.assertEqual(200, response.status_code)
        results = response.json_body
        self.assertEqual(0, len(results['errors']))

    def test_bundle_proof_invalid_config_type(self):
        # We have to create the charm so that we can verify it exists.
        _id, charm = factory.makeCharm(
            self.db,
            description=''
        )

        # and we'll cheat and just find it straight by the store url.
        charm = Charm(charm)
        store_url = charm.store_url

        bundle_config = {
            'wiki': {
                'services': {
                    'charm': {
                        'charm_url': str(store_url),
                        'options': {
                            'script-interval': 'invalid'
                        }
                    }
                }
            }
        }

        request = self.make_request('bundle', remainder='/proof')
        request.params = {
            'deployer_file': yaml.dump(bundle_config)
        }
        response = self.api_class(request)()
        self.assertEqual(200, response.status_code)
        results = response.json_body
        self.assertEqual(1, len(results['errors']))
        self.assertEqual(1, len(results['error_messages']))

        # Note that the View will prefix the error message from the proof with
        # the bundle name to help identify where the error message came from.
        self.assertEqual(
            'wiki: script-interval is not of type int.',
            results['error_messages'][0]
        )

    def test_bundle_proof_supports_multiple_errors(self):
        _id, charm = factory.makeCharm(
            self.db,
            description=''
        )

        # and we'll cheat and just find it straight by the store url.
        charm = Charm(charm)
        store_url = charm.store_url

        bundle_config = {
            'wiki': {
                'services': {
                    'charm': {
                        'charm_url': str(store_url),
                        'options': {
                            'script-interval': 'invalid',
                            'no-exist': 'hah invalid',
                        }
                    },
                    'fail': {
                        'name': 'will-fail'
                    }
                }
            }
        }

        request = self.make_request('bundle', remainder='/proof')
        request.params = {
            'deployer_file': yaml.dump(bundle_config)
        }
        response = self.api_class(request)()
        self.assertEqual(200, response.status_code)
        results = response.json_body
        self.assertEqual(1, len(results['errors']))
        self.assertEqual(3, len(results['error_messages']))

        # One message from not finding a charm.
        self.assertEqual(
            'wiki: Could not find charm: fail',
            results['error_messages'][0]
        )
        # The other two from checking the config of the other charm.
        self.assertEqual(
            'wiki: script-interval is not of type int.',
            results['error_messages'][1]
        )
        self.assertEqual(
            'wiki: The charm has no option for: no-exist',
            results['error_messages'][2]
        )

    def test_bundle_proof_errors_on_invalid_relations(self):
        # The default charm provides website:https and requires
        # database:mongodb
        one_store_url, two_store_url = self.setup_relatable_charms(
            {
                'proxy': {
                    'interface': 'http'
                }
            },
            {
                'cache': {
                    'interface': 'memcache'
                },
                'db': {
                    'interface': 'mongodb'
                }
            }
        )

        bundle_base = """
wiki:
  series: precise
  promulgated: True
  services:
    wiki:
      charm: {}
    db:
      charm: {}
  relations:
    - ["wiki:proxy", db]""".format(one_store_url, two_store_url)

        request = self.make_request('bundle', remainder='/proof')
        request.params = {
            'deployer_file': bundle_base
        }
        response = self.api_class(request)()
        self.assertEqual(200, response.status_code)
        self.assertIn(
            'wiki: Invalid relation requested: proxy',
            response.json_body['error_messages'][0])

    def test_bundle_proof_errors_on_incompatible_relations(self):
        one_store_url, two_store_url = self.setup_relatable_charms(
            {
                'proxy': {
                    'interface': 'http'
                }
            }, {
                'cache': {
                    'interface': 'memcache'
                },
                'db': {
                    'interface': 'mongodb'
                }
            }
        )

        bundle_base = """
wiki:
  series: precise
  promulgated: True
  services:
    wiki:
      charm: {}
    db:
      charm: {}
  relations:
    - ["wiki:db", "db:proxy"]""".format(one_store_url, two_store_url)

        request = self.make_request('bundle', remainder='/proof')
        request.params = {
            'deployer_file': bundle_base
        }
        response = self.api_class(request)()
        self.assertEqual(200, response.status_code)
        self.assertIn(
            'wiki: The requested relation db to proxy is incompatible',
            response.json_body['error_messages'][0])

    def test_bundle_proof_passes_with_valid_relations(self):
        one_store_url, two_store_url = self.setup_relatable_charms(
            {
                'proxy': {
                    'interface': 'https'
                }
            }, {
                'cache': {
                    'interface': 'memcache'
                },
                'db': {
                    'interface': 'mongodb'
                }
            }
        )

        bundle_base = """
wiki:
  series: precise
  promulgated: True
  services:
    wiki:
      charm: {}
    db:
      charm: {}
  relations:
    - ["wiki:database", "db:db"]""".format(one_store_url, two_store_url)

        request = self.make_request('bundle', remainder='/proof')
        request.params = {
            'deployer_file': bundle_base
        }
        response = self.api_class(request)()
        self.assertEqual(200, response.status_code)
        self.assertEqual(0, len(response.json_body['error_messages']))

    def test_bundle_proof_passes_with_valid_vague_as_can_be(self):
        one_store_url, two_store_url = self.setup_relatable_charms(
            {
                'proxy': {
                    'interface': 'https'
                }
            }, {
                'cache': {
                    'interface': 'memcache'
                },
                'db': {
                    'interface': 'mongodb'
                }
            }
        )

        bundle_base = """
wiki:
  series: precise
  promulgated: True
  services:
    wiki:
      charm: {}
    db:
      charm: {}
  relations:
    - ["wiki", "db"]""".format(one_store_url, two_store_url)

        request = self.make_request('bundle', remainder='/proof')
        request.params = {
            'deployer_file': bundle_base
        }
        response = self.api_class(request)()
        self.assertEqual(200, response.status_code)
        self.assertEqual(0, len(response.json_body['error_messages']))

    def test_bundle_proof_errors_with_missing_service(self):
        one_store_url, two_store_url = self.setup_relatable_charms(
            {
                'proxy': {
                    'interface': 'https'
                }
            }, {
                'cache': {
                    'interface': 'memcache'
                },
                'db': {
                    'interface': 'mongodb'
                }
            }
        )

        bundle_base = """
wiki:
  series: precise
  promulgated: True
  services:
    wiki:
      charm: {}
    db:
      charm: {}
  relations:
    - ["wiki", "party"]""".format(one_store_url, two_store_url)

        request = self.make_request('bundle', remainder='/proof')
        request.params = {
            'deployer_file': bundle_base
        }
        response = self.api_class(request)()
        self.assertEqual(200, response.status_code)
        self.assertEqual(
            'wiki: Invalid relation to charm: party not found.',
            response.json_body['error_messages'][0])
