# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
import datetime
import itertools

import pymongo
from pyramid.view import view_config

from charmworld import cached_view_config, utils
from charmworld.models import (
    CharmSource,
    QADataSource,
)
from charmworld.views.helpers import format_proof


# 3m
REPORT_CACHE = 180

TYPES_WITH_DATES = ['task', 'proposal', 'question', 'pull request']


def get_missing_qa_items(source, db):
    qa_source = QADataSource.from_db(db)
    qaed_charms = [(a['name'], a['series']) for a in
                   qa_source.answers.find()]
    return [charm for charm in source.find_promulgated_charms() if
                     (charm.name, charm.series) not in qaed_charms]


def get_qa_entry_data(item):
    return {
        'summary': 'Missing QA data on %s' % item.name,
        'item': item.short_url + "/qa/edit",
    }


def get_proof_entry_data(item):
    return {
        'summary': 'Proof error on %s' % item.name,
        'item': item.short_url + "#proof",
    }


def get_store_entry_data(item):
    return {
        'summary': 'Store error on %s' % item.name,
        'item': 'http://code.launchpad.net/%s' % item.branch_spec,
    }


def format_items_for_queue(items, item_type):
    item_queue = []
    for item in items:
        entry = {'type': item_type, 'status': 'N\A'}
        if item_type == 'qa':
            entry.update(get_qa_entry_data(item))
        if item_type == 'proof':
            entry.update(get_proof_entry_data(item))
        if item_type == 'store':
            entry.update(get_store_entry_data(item))
        item_queue.append(entry)
    return item_queue


def build_review_queue(request):
    source = CharmSource.from_request(request)
    proof_items = source.find_proof_error_charms(False)
    store_items = source.find_mia_charms()
    qa_items = get_missing_qa_items(source, request.db)

    review_items = list(
        request.db.review_queue.find(
            sort=[('date_modified', pymongo.ASCENDING)]))

    askubuntu_items = list(
        request.db.answer_queue.find(
            sort=[('date_modified', pymongo.ASCENDING)]))

    github_pull_requests = list(
        request.db.github_pull_requests.find(
            sort=[('date_modified', pymongo.ASCENDING)]))

    return list(itertools.chain(
        review_items,
        format_items_for_queue(proof_items, 'proof'),
        format_items_for_queue(store_items, 'store'),
        format_items_for_queue(qa_items, 'qa'),
        askubuntu_items,
        github_pull_requests
    ))


def get_sparklines(stats):
    mins = []
    maxes = []
    means = []
    for s in stats:
        mins.append(str(s['min']))
        means.append(str(s['mean']))
        maxes.append(str(s['max']))

    return {
        'mins': ','.join(mins),
        'means': ','.join(means),
        'maxes': ','.join(maxes)}


@cached_view_config(
    route_name="tools",
    renderer="charmworld:templates/tools.pt")
def tools(request):
    return {}


@view_config(
    route_name="tools-review",
    http_cache=REPORT_CACHE,
    renderer="charmworld:templates/review.pt")
def charm_review(request):
    review_queue = build_review_queue(request)

    stats = list(request.db.review_queue_latency.find(
        date={"$gte": datetime.datetime.utcnow() - datetime.timedelta(180)},
        sort=[('date_modified', pymongo.ASCENDING)]))
    sparklines = get_sparklines(stats)
    if stats:
        current_stats = stats[0]
        max_latency = utils.prettify_timedelta(
            datetime.timedelta(seconds=current_stats['max']))
        min_latency = utils.prettify_timedelta(
            datetime.timedelta(seconds=current_stats['min']))
        mean_latency = utils.prettify_timedelta(
            datetime.timedelta(seconds=current_stats['mean']))
    else:
        max_latency = min_latency = mean_latency = 'N/A'

    return {
        'queue': review_queue,
        'TYPES_WITH_DATES': TYPES_WITH_DATES,
        'age_format': utils.pretty_timedelta,
        'max': max_latency,
        'mean': mean_latency,
        'min': min_latency,
        'sparklines': sparklines}


# TODO After the store errors in review queue are approved as the way to deal
# with reviewing charm store errors, this view can go away.
@view_config(
    route_name="tools-store-missing",
    http_cache=REPORT_CACHE,
    renderer="charmworld:templates/store-missing.pt")
def charm_store_missing(request):
    queue = list(CharmSource.from_request(request).find_mia_charms())
    return {'queue': queue, 'age_format': utils.pretty_timedelta}


# TODO After the proofs in review queue are approved as the way to deal with
# reviewing charm proof errors, this view can go away.
@view_config(
    route_name="tools-proof-errors",
    http_cache=REPORT_CACHE,
    renderer="charmworld:templates/charm-proof-errors.pt")
def proof_errors(request):
    sub_only = bool(request.GET.get('sub'))
    source = CharmSource.from_request(request)
    charms = list(source.find_proof_error_charms(sub_only))

    return {
        'charms': charms,
        'format_proof': format_proof,
        'sub_only': sub_only,
    }
